/* ----------------------------------------------------------------------------
  ex10.C
  mbwall 10apr95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   Sample program that illustrates how to use a distance function to do 
speciation.
---------------------------------------------------------------------------- */
#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>
#include <math.h>

float objectiveFunction(GAGenome &);
float distanceFunction(GAGenome & a, GAGenome & b);

int
main(int argc, char **argv)
{
  cout << "This program uses sharing to do speciation.  The objective\n";
  cout << "function has more than one optimum, so different genomes\n";
  cout << "may have equally high scores.  Speciation keeps the population\n";
  cout << "from clustering at one optimum.  The function is\n";
  cout << "  y = sin(x*2*pi)\n";
  cout << "where x is constrained to\n";
  cout << "  0 <= x <= 5\n";
  cout << "Notice that speciation will retard convergence, but if you run\n";
  cout << "the GA long enough it will converge regardless of the\n";
  cout << "speciation.  You can control this with the shape of the sharing\n";
  cout << "function (see Goldberg for details).\n";
  cout << "\n\n";
  cout.flush();

// Declare variables for the GA parameters and set them to some default values.

  int popsize  = 50;
  int ngen     = 50;
  float pmut   = 0.0;
  float pcross = 0.6;
  int speciate = 1;
  int dofunc   = 0;
  char filename[128] = "";

// Parse the command line to get the flag that tells us whether or not to do
// speciation.  By default we do it.

  int i;
  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	ngen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("no", argv[i]) == 0){
      speciate = 0;
      continue;
    }
    else if(strcmp("func", argv[i]) == 0){
      dofunc = 1;
      continue;
    }
    else if(strcmp("file", argv[i]) == 0 || strcmp("f", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the file option needs a filename.\n";
        exit(1);
      }
      else{
	sprintf(filename, argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  no\tdo not do speciation\n";
      cerr << "  func\tjust calculate function\n";
      cerr << "  file\tdump function to specified file\n";
      cerr << "\n";
      exit(1);
    }
  }

// If a filename was specified, then dump the x,y pairs of the function into
// that file (this is so you can look at the data).  Then we exit.

  if(dofunc && strcmp(filename,"") != 0){
    ofstream datastream(filename, ios :: out);
    if(!datastream){
      cerr << "Cannot open " << filename << " for output.\n";
      exit(1);
    }
    for(float x=0.0; x<5.0; x+=0.05)
      datastream << x << "\t" << sin(x*2*M_PI) << "\n";
    datastream.close();
    cout << "the function has been written to '" << filename << "'\n";
    exit(0);
  }

// Create a phenotype for two variables.  The number of bits you can use to
// represent any number is limited by the type of computer you are using.  In
// this case, we use 16 bits to represent a floating point number whose value
// can range from -5 to 5, inclusive.  The bounds on x1 and x2 can be applied
// here and/or in the objective function.

  GABin2DecPhenotype map;
  map.add(31, 0, 5);

// Create the template genome using the phenotype map we just made.

  GABin2DecGenome chrom(map, objectiveFunction);

// Now create the GA using the genome and set the distance function.  We also
// set the scaling to sigma truncation so that we can handle negative objective
// scores.

  GASteadyStateGA ga(chrom);
  ga.pReplacement(0.4);
  GASigmaTruncationScaling trunc;
  ga.scalingScheme(trunc);
  GASharing sharing(distanceFunction);
  if(speciate) ga.scalingScheme(sharing);

// Set the other GA parameters then run the GA.

  ga.populationSize(popsize);
  ga.nGenerations(ngen);
  ga.pMutation(pmut);
  ga.pCrossover(pcross);
  ga.evolve();

// Dump the results of the GA to the screen.

  chrom = ga.bestOfAll();
  cout << "the ga found an optimum at the point " << chrom.phenotype(0) <<"\n";
  cout << "\nhere are all of the members of the population:\n";
  cout.width(9);
  cout << "x";
  cout.width(9);
  cout << "\ty(x)\n--------------------------\n";
  for(i=0; i<ga.population().size(); i++){
    cout.width(9);
    cout<<((GABin2DecGenome&)(ga.population().individual(i))).phenotype(0);
    cout << "\t";
    cout.width(9);
    cout << ga.population().individual(i).score() << "\n";
  }
  cout << "\n\n";
  cout.flush();

// Dump the best of generation scores to a file.

  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";

  if(strcmp(filename,"") != 0){
    ofstream datastream(filename, ios :: out);
    if(!datastream){
      cerr << "Cannot open " << filename << " for output.\n";
      exit(1);
    }
    for(i=0; i<ga.population().size(); i++){
      datastream << ((GABin2DecGenome&)
		     (ga.population().individual(i))).phenotype(0);
      datastream << "\t" << ga.population().individual(i).score() << "\n";
    }
    datastream.close();
    cout << "the output has been written to '" << filename << "'\n";
    exit(0);
  }

  return 0;
}
 




// This objective function is simply the sine of the decimal value in the
// genome.  Notice that this function can return negative values, so we
// must use sigma truncation scaling (or some other negative-safe scaling).
float
objectiveFunction(GAGenome & c){
  GABin2DecGenome & chrom = (GABin2DecGenome &)c;
  return sin(chrom.phenotype(0)*2*M_PI);
}





// Here are a couple of possible distance functions for this problem.  One of
// them uses the genes to determine the same-ness, the other uses the
// phenotypes to determine same-ness.  If the genomes are the same, then
// we return a 1.  If they are completely different then we return a 0.
//   In either case, you should be sure that the distance function will return
// values only between 0 and 1 inclusive.  If your function returns values
// outside these limits, the GA will produce bogus results and it WILL NOT warn
// you that your distance function is brain-dead!

// This distance function uses the genes to determine same-ness.  All we do 
// is check to see if the bit strings are identical.
/*
float
distanceFunction(GAGenome & c1, GAGenome & c2){
  GABin2DecGenome & a = (GABin2DecGenome &)c1;
  GABin2DecGenome & b = (GABin2DecGenome &)c2;

  float x=0;
  for(int i=a.length()-1; i>=0; i--)
    x += (a[i] == b[i] ? 1 : 0);

  return x/a.length();
}
*/


// This distance function looks at the phenotypes rather than the genes of the
// genome.  This distance function will try to drive them to extremes.

float
distanceFunction(GAGenome & c1, GAGenome & c2){
  GABin2DecGenome & a = (GABin2DecGenome &)c1;
  GABin2DecGenome & b = (GABin2DecGenome &)c2;

  return 1.0 - fabs(a.phenotype(0) - b.phenotype(0)) / 5;
}

