/* ----------------------------------------------------------------------------
  ex11.C
  mbwall 13apr95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   This example shows how to use an order-based list genome.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>


#if defined(__GNUG__)
#define NO_AUTO_INST
#endif


// This is how we auto-include the list operators.  This will work only if we
// are using only one type of list - if you are going to instantiate lists of
// more than one type then you'll have to copy the code from list.op.C 
// explicitly and replace 'LIST_TYPE' with the types you are going to use
// then change the function names so they don't conflict.
#define LIST_TYPE int
#include <ga/list.op.h>


// If your compiler does not do automatic instantiation (e.g. g++ 2.6.8),
// then define the NO_AUTO_INST directive.
#ifdef NO_AUTO_INST
#include <ga/listtmpl.C>
#include <ga/list.ch.C>
#include <ga/list.xs.C>
template class GAList<int>;
template class GAListGenome<int>;
template class GAList1PtCrossSite<int>;
template class GAListOrderedCrossSite<int>;
#endif



// The objective function tells how good a genome is.  The Initializer defines
// how the lists should be initialized.
float objective(GAGenome &);
void ListInitializer(GAGenome &);


int
main(int argc, char *argv[])
{
  cout << "This program illustrates the use of order-based lists.  The\n";
  cout << "list in this problem contains 25 numbers, 0 to 24.  It tries\n";
  cout << "to put them in descending order from 24 to 0.\n\n";
  cout.flush();

// Set the default values of the parameters and declare the params variable.

  int i;
  GAParameters params(30,	// population size
		      0.6,	// probability of crossover
		      0.01,	// probability of mutation
		      1000);	// number of generations

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

// Now create the GA and run it.  We first create a genome with the
// operators we want.  Since we're using a template genome, we must assign
// all three operators.  We use the order-based crossover site when we assign
// the crossover operator.

  GAListGenome<int> chrom(objective);
  GAListOrderedCrossSite<int> crossSite;
  chrom.initializationOperator(ListInitializer);
  chrom.mutationOperator(GAListSwapMutator);
  chrom.crossover(GAListPartialMatchCrossover, crossSite);

// Now that we have our genome, we create the GA (it clones the genome to 
// make all of the individuals for its populations).  Set the parameters on 
// the GA then let it evolve.

  GASteadyStateGA ga(chrom);
  ga.parameters(params);
  ga.pReplacement(0.25);
  ga.evolve();

// Assign the best that the GA found to our genome then print out the results.

  chrom = ga.bestOfAll();
  cout << "the ga generated the following list (objective score is ";
  cout << chrom.score() << "):\n";
  cout << chrom << "\n";

  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";

  return 0;
}
 


/* ----------------------------------------------------------------------------
Objective function
  The objective function gives one point for every number in the correct
position.  We're trying to get a sequence of numbers from n to 0 in descending
order.
---------------------------------------------------------------------------- */
float
objective(GAGenome & c)
{
  GAListGenome<int> & chrom = (GAListGenome<int> &)c;

  float score = (*chrom.head() == chrom.size()-1); // move to head of list
  for(int i=chrom.size()-2; i>0; i--)
    score += (*chrom.next() == i); // cycle through list and look at nodes
  return score;
}



/* ----------------------------------------------------------------------------
List Genome Operators
-------------------------------------------------------------------------------
The initializer creates a list with n elements in it and puts a unique digit
in each one.
---------------------------------------------------------------------------- */
void
ListInitializer(GAGenome & c)
{
  GAListGenome<int> &child=(GAListGenome<int> &)c;
  while(child.head()) child.destroy(); // destroy any pre-existing list

  int n=25;
  child.insert(0,GAListBASE::HEAD); // the head node contains a '0'
  for(int i=1; i<n; i++)
    child.insert(i);		// each subsequent node contains a number
}




/* ----------------------------------------------------------------------------
List genome method overrides
---------------------------------------------------------------------------- */
//   Here we override the _write method for the List class.  This lets us see
// exactly what we want (the default _write method dumps out pointers to the
// data rather than the data contents).
//   This routine prints out the contents of each element of the list, 
// separated by a space.  It does not put a newline at the end of the list.
//   Notice that you can override ANY function of a template class.  This is
// called "specialization" in C++ and it lets you tailor the behaviour of a 
// template class to better fit the type.
void
GAListGenome<LIST_TYPE>::_write(ostream & os) const
{
  LIST_TYPE *cur, *head;
  GAListIter<LIST_TYPE> iter(*this);
  if(head=iter.head()) os << *head << " ";
  for(cur=iter.next(); cur && cur != head; cur=iter.next())
    os << *cur << " ";
}
