/* ----------------------------------------------------------------------------
  ex12.C
  mbwall 13apr95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   This example shows how to use an order-based string genome.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>

// This is the declaration of the objective function that we will use in this
// example.
float objective(GAGenome &);

// Every genome must have an initializer.  Here we declare our own initializer
// that will be used in this example.
void AlphabetInitializer(GAGenome &);


int
main(int argc, char *argv[])
{
  cout << "This program illustrates the use of order-based strings.  The\n";
  cout << "string in this problem contains 26 letters, a to z.  It tries\n";
  cout << "to put them in alphabetic order.\n\n";
  cout.flush();

// Set the default values of the parameters.

  int i;
  GAParameters params(25,	// population size
		      0.9,	// probability of crossover 
		      0.01,	// probability of mutation
		      4000);	// number of generations

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

// Now create the GA and run it.  We first create a genome with the
// operators we want.  Since we're using a template genome, we must assign
// all three operators.  We use the order-based crossover site when we assign
// the crossover operator.

  GAStringAlleleSet alleles;
  for(i=0; i<26; i++)
    alleles.add('a'+i);

  GAStringGenome chrom(26, alleles, objective);
  GAStringOrderedCrossSite crossSite;
  chrom.initializationOperator(AlphabetInitializer);
  chrom.mutationOperator(GAStringSwapMutator);
  chrom.crossover(GAStringPartialMatchCrossover, crossSite);

  GASteadyStateGA ga(chrom);
  ga.parameters(params);
  ga.evolve();

  chrom = ga.bestOfAll();
  cout << "the ga generated the following string (objective score is ";
  cout << chrom.score() << "):\n";
  cout << chrom << "\n";

  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";

  return 0;
}
 


/* ----------------------------------------------------------------------------
Objective function
  The objective function gives one point for every number in the correct
position.  We're trying to get a sequence of numbers from n to 0 in descending
order.
---------------------------------------------------------------------------- */
float
objective(GAGenome & c)
{
  GAStringGenome & chrom = (GAStringGenome &)c;

  float score = 0;
  for(int i=0; i<chrom.size(); i++)
    score += (chrom[i] == 'a' + i);
  return score;
}



/* ----------------------------------------------------------------------------
AlphabetInitializer
  This initializer creates a string genome with the letters a-z as its 
elements.  Once we have assigned all the values, we randomize the string.
---------------------------------------------------------------------------- */
void
AlphabetInitializer(GAGenome & c)
{
  GAStringGenome &chrom=(GAStringGenome &)c;
  int i;
  for(i=chrom.size()-1; i>=0; i--)
    chrom[25-i] = 'a'+i;
//  chrom.gene(25-i, 'a'+i);    // this is another way to set the genes

  for(i=chrom.size()-1; i>=0; i--)
    if(GARandomBit()) chrom.swap(i, GARandomInt(0, chrom.size()-1));
}
