/* ----------------------------------------------------------------------------
  ex13.C
  mbwall 29apr95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  This example illustrates the use of a GA-within-a-GA.  It uses a steady
state GA to find the smiley face, then it uses another steady state GA to match
a sequence of random numbers.  It doesn't try to do the random numbers until
it has gotten 90% of the way to the smiley face.
  This is very similar in function to the composite genome example, but 
here we evolve the two genomes separately rather than as a single entity.
  This kind of application can be useful for situations where the computational
cost of calculating feasibility is rather high and the feasible space is 
sparse.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>


typedef struct _UserData {
  int width, height;
  short **picture_target;
  float *numbers_target;
  GA2DBinaryStringGenome *picture_chrom;
  GABin2DecGenome *numbers_chrom;
} UserData;

float pictureObjective(GAGenome &);
float numbersObjective(GAGenome &);

int
main(int argc, char *argv[])
{
  cout << "This program runs a GA-within-GA.  The outer level GA tries to\n";
  cout << "match the pattern read in from a file.  The inner level GA is\n";
  cout << "run only when the outer GA reaches a threshhold objective score\n";
  cout << "then it tries to match a sequence of numbers that were generated\n";
  cout << "randomly at the beginning of the program's execution.\n\n";

// Set the default values of the parameters and declare the params variable.

  int i,j;
  GAParameters params(30, 0.8, 0.005, 200);
  char filename[128] = "smiley.dat";

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("file", argv[i]) == 0 || strcmp("f", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the file option needs a filename.\n";
        exit(1);
      }
      else{
	sprintf(filename, argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  f\tfilename from which to read (" << filename << ")\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

// Create a user data object.  We'll keep all of the information for this 
// program in this object.
  UserData data;

// Read in the pattern from the specified file.  File format is pretty simple:
// two integers that give the height then width of the matrix, then the matrix
// of 1's and 0's (with whitespace inbetween).

  ifstream inStream(filename, ios :: in);
  if(!inStream){
    cerr << "Cannot open " << filename << " for input.\n";
    exit(1);
  }
  inStream >> data.height;
  inStream >> data.width;
  data.picture_target = new short*[data.width];
  for(i=0; i<data.width; i++)
    data.picture_target[i] = new short[data.height];
  for(j=0; j<data.height; j++)
    for(i=0; i<data.width; i++)
      inStream >> data.picture_target[i][j];
  inStream.close();

// Print out the pattern to be sure we got the right one.

  cout << "input pattern:\n";
  for(j=0; j<data.height; j++){
    for(i=0; i<data.width; i++){
      cout << (data.picture_target[i][j] == 1 ? '*' : ' ') << " ";
    }
    cout << "\n";
  }
  cout << "\n"; cout.flush();

// Generate the random sequence of numbers.

  int n=7;
  float min[] = {0, 0,   3, -5, 100,    0.001, 0};
  float max[] = {1, 100, 3, -2, 100000, 0.010, 7};
  GARandomSeed();
  data.numbers_target = new float[n];
  for(i=0; i<n; i++)
    data.numbers_target[i] = GARandomFloat(min[i], max[i]);
  data.numbers_target[6] = GARandomInt((int)min[6], (int)max[6]);

// Print out the sequence so we'll know what we have to match.

  cout << "input sequence:\n";
  for(i=0; i<n; i++){
    cout.width(10);
    cout << data.numbers_target[i] << " ";
  }
  cout << "\n"; cout.flush();


// Create a phenotype for the numbers to map them to a bin2dec genome.

  GABin2DecPhenotype map;
  for(i=0; i<n; i++)
    map.add(8, min[i], max[i]);

// Create a couple of genomes for keeping track in our user data.
  data.picture_chrom = new GA2DBinaryStringGenome(data.width, data.height);
  data.numbers_chrom = new GABin2DecGenome(map);

// Now create the GA and run it.

  GA2DBinaryStringGenome
    picture_chrom(data.width, data.height, pictureObjective, (void *)&data);
  GABin2DecGenome numbers_chrom(map);
//  GASteadyStateGA ga(picture_chrom);
  GAReplacementGA ga(picture_chrom);
  ga.parameters(params);
  ga.populationSize(150);
  ga.evolve();

// Now that we have evolved the best solution, put the best into our genomes
// then print them out.

  picture_chrom = ga.bestOfAll();
  cout << "the ga generated:\n";
  for(j=0; j<data.height; j++){
    for(i=0; i<data.width; i++){
      cout << (picture_chrom.gene(i,j) == 1 ? '*' : ' ') << " ";
    }
    cout << "\n";
  }
  cout << "\n"; cout.flush();

  numbers_chrom = *data.numbers_chrom;
  for(i=0; i<map.nPhenotypes(); i++){
    cout.width(10);
    cout << numbers_chrom.phenotype(i) << " ";
  }
  cout << "\n\n"; cout.flush();

  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";


// free up all of the space we were using.

  for(i=0; i<data.height; i++)
    delete data.picture_target[i];
  delete [] data.picture_target;
  delete [] data.numbers_target;

  delete data.picture_chrom;
  delete data.numbers_chrom;

  return 0;
}
 




// This is the primary objective function.  If it gets a genome whose score
// is sufficiently high, then it runs another GA on the number sequence and
// returns a composite score.
float
pictureObjective(GAGenome & c)
{
  GA2DBinaryStringGenome & chrom = (GA2DBinaryStringGenome &)c;
  UserData * data = (UserData *)c.userData();

  float value=0.0;
  for(int i=0; i<chrom.width(); i++)
    for(int j=0; j<chrom.height(); j++)
      value += (float)(chrom.gene(i,j) == data->picture_target[i][j]);

  if(value/(chrom.width()*chrom.height()) < 0.95) return(value);

  *data->picture_chrom = chrom;
  data->picture_chrom->objectiveFunction(numbersObjective);
//  GASteadyStateGA ga(*(data->numbers_chrom));
  GAReplacementGA ga(*(data->numbers_chrom));
  ga.populationSize(150);
  ga.nGenerations(400);
  ga.pMutation(0.01);
  ga.pCrossover(0.9);
  ga.evolve();
  *data->numbers_chrom = ga.bestOfAll();
  value += 10 * ga.bestOfAll().score();

  return(value);
}


// This is the objective function for matching the sequence of numbers.
float
numbersObjective(GAGenome & c)
{
  GABin2DecGenome & chrom = (GABin2DecGenome &)c;
  UserData * data = (UserData *)c.userData();

  float value=chrom.nPhenotypes();
  for(int i=0; i<chrom.nPhenotypes(); i++)
    value += 1.0 / (1.0 + fabs(chrom.phenotype(i) - data->numbers_target[i]));
  return(value);
}
