/* ----------------------------------------------------------------------------
  ex14.C
  mbwall 29apr95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   Example program for a composite genome derived from the GAGenome and
containing a list of lists.
---------------------------------------------------------------------------- */
#define _ex14_C_

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>
#include "ex14.h"


#if defined(__GNUG__)
#define NO_AUTO_INST
#endif


// If your compiler does not do automatic instantiation (e.g. g++ 2.6.8),
// then define the NO_AUTO_INST directive.
#ifdef NO_AUTO_INST
#include <ga/listtmpl.C>
#include <ga/list.ch.C>
#include <ga/list.xs.C>
template class GAList<int>;
template class GAListGenome<int>;
template class GAList1PtCrossSite<int>;
template class GAListOrderedCrossSite<int>;
#endif


// Here we specify how big the lists will be and how many lists will be in each
// composite genome.  These are the default values - you can change them from
// the command line.  Beware that this program will break if nrobots is bigger
// than the size of the lists.
int listsize = 10;
int nrobots = 6;


float objective(GAGenome &);



int
main(int argc, char *argv[])
{
  cout << "This example shows how to create a genome that contains\n";
  cout << "a list of lists.  We create a composite genome that has\n";
  cout << "lists in it.  Each list has some nodes, only one of which\n";
  cout << "contains the number 0.  The objective is to move the node with\n";
  cout << "number 0 in it to the nth position where n is the number of the\n";
  cout << "list within the composite genome.\n\n";
  cout.flush();

// Set the default values of the parameters and declare the params variable.

  int i;
  GAParameters params(30,	// population size
		      0.6,	// probability of crossover
		      0.02,	// probability of mutation
		      300);	// number of generations

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("nr", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of robots needs a value.\n";
        exit(1);
      }
      else{
	nrobots = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("ls", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": list size needs a value.\n";
        exit(1);
      }
      else{
	listsize = atoi(argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      cerr << "  nr\tnumber of robots (" << nrobots << ")\n";
      cerr << "  ls\tsize of the list (" << listsize << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

// Create an instance of our custom genome.  Once we create an instance,
// we can then create a GA (the GA uses the 'clone' member of the genome
// to generate the population before it starts evolving).

  RobotPathGenome chrom(nrobots, objective);

// This creates then evolves the GA

  GASteadyStateGA ga(chrom);
  ga.parameters(params);
  ga.evolve();

// Now dump the results to the screen.

  chrom = ga.bestOfAll();
  cout << "the ga generated:\n";
  cout << chrom << "\n";

  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";

  return 0;
}
 

// The objective function should evaluate the genomes.  This one tries to
// put the node with value 0 into the nth position where n is the number of the
// list in the composite genome.  We're assuming that there are more nodes
// in the list than there are lists in the composite genome.
float
objective(GAGenome & c)
{
  RobotPathGenome & chrom = (RobotPathGenome &)c;
  float score=0;
  for(int i=0; i<chrom.npaths(); i++)
    if(*chrom.path(i).warp(i) == 0) score += 1;
  return score;
}


// This is the initialization operator for our lists.  We create a list that is
// n long and whose nodes contain numbers in sequence.
//   The first thing to do in the initializer is to clear out any old 
// contents - we do not want to build our new list on a previously existing
// one!  Notice that we have to cast the genome into the type of 
// genome we're using (in this case a list).  The GA always operates on
// generic genomes.
//   All of our lists must be the same length since we're going to use the
// ordered crossover operators.
void ListInitializer(GAGenome & c){
  GAListGenome<int> & list = (GAListGenome<int> &)c;

  while(list.head()) list.destroy(); // this will trash any existing list

// Insert the head of the list.  This node has a random number in it, but the
// number is in a range different than those in the rest of the list.  This
// way we'll be able to see how the lists get scrambled up.  Since we're using
// ordered crossover (see the header file) we should never end up with more 
// than one node in each list that has a given value.
  list.insert(0, GAListBASE::HEAD);

// Loop through n times and insert nodes onto the end of the list.
  int i;
  for(i=0; i<listsize-1; i++)
    list.insert(i+20, GAListBASE::AFTER);

// Now randomize the contents of the list.
  for(i=0; i<listsize; i++)
    if(GARandomBit()) list.swap(i, GARandomInt(0, listsize-1));
}



// Here is the override of the _write method for our lists.  The default 
// write method just prints out pointers the contents of the nodes (it has
// no way of knowing in advance how you'll want things printed).  Here we 
// do almost the same thing, but print out the contents of the nodes rather
// than the pointers to the contents.
void
GAListGenome<int>::_write(ostream & os) const
{
  int *cur, *head;
  GAListIter<int> iter(*this);
  if(head=iter.head()) os << *head << " ";
  for(cur=iter.next(); cur && cur != head; cur=iter.next())
    os << *cur << " ";
}
