/* ----------------------------------------------------------------------------
  ex14.h
  mbwall 29apr95
  Copyright 1995 Massachusetts Institute of Technology

DESCRIPTION:
  This header defines a composite genome that uses n different list
genomes to represent pseudo-indpendent sequences (this example was written
for a robot trajectory problem and stripped down for example purposes).
  The composite genome contains a user-specifiable number of lists.  Each
list behaves differently and is not affected by mutations, crossovers, or 
initializations of the other lists.
---------------------------------------------------------------------------- */
#ifndef _ex14_h_
#define _ex14_h_

#include <ga/xsite.h>
#include <ga/genome.h>
#include <ga/list.ch.h>
#include <ga/list.xs.h>


// This is the definition of the crossover site object.  It is used by the
// composite genome for crossovers.  
//   Notice that this crossover site does not need to store any special 
// information; the site associated with each genome inside the composite
// takes care of that.  All this site does is act as a broker to tell each of
// the contained sites when to do their thing.
class RobotPathCrossSite : public GACrossSite {
  GADefineIdentity("RobotPathCrossSite", 201): // define the name and id

public:
  RobotPathCrossSite(){}
  RobotPathCrossSite(const RobotPathCrossSite &){}
  RobotPathCrossSite & operator=(const GACrossSite &){return *this;}
  RobotPathCrossSite & operator=(const RobotPathCrossSite &){return *this;}
  virtual ~RobotPathCrossSite(){}
  virtual void copy(const GACrossSite &){}
  virtual GACrossSite * clone() const {return new RobotPathCrossSite(*this);}
  void pick(GAGenome &);
  void pick(GAGenome &, const GACrossSite &){}
  void pick(GAGenome &, const GAGenome &, const GAGenome &);
  void mirror(GAGenome &, const GAGenome &);
};




// We want to create only one instance of this crossover site.  It is used as
// the default crossover site for this class of genomes.  
//   We also include here the header file for the list genome operators.
// These are used when we define the default mutator and crossover for the 
// lists when we create a robot path genome.
//   The #ifdef _ex14_C_ makes sure that we instantiate the crossover sites 
// only once.
#ifdef _ex14_C_
#define LIST_TYPE int
#include <ga/list.op.h>
GAListOrderedCrossSite<int> listOrderedCrossSite;
RobotPathCrossSite robotPathCrossSite;
#else
#include <ga/list.op.h>
extern GAListOrderedCrossSite<int> listOrderedCrossSite;
extern RobotPathCrossSite robotPathCrossSite;
#endif



// We have to define a list initializer for each of the lists in the robot
// path genome.  Here we'll use the same one for all of the lists.  The
// declaration of the initializer is here - the definition is in the .C file.
void ListInitializer(GAGenome &);



// These are the default operators for this genome class.  All they do is
// call the respective initializer, mutator, and crossover methods on the
// lists.  We declare them here then define them later on.
inline void RobotPathInitializer(GAGenome &);
inline int  RobotPathMutator(GAGenome &, float);
inline void RobotPathCrossover(GAGenome &, const GAGenome &, const GAGenome &);




// This is the class definition.  We override the methods of the base class and
// define a few methods of our own to access the protected members.  The list
// genomes in the composite genome are assigned the 'List' operators
// by default (they can be overridden by using the 'Operator' members on the
// lists explicitly).  We MUST make these assignments because lists do not
// automatically get the operators (since they are template classes).
class RobotPathGenome : public GAGenome {
  GADefineIdentity("RobotPath", 202): // define the name and id

public:
  RobotPathGenome(int nrobots,
		  GAObjectiveFunction f=NULL,
		  void * u=NULL,
		  GA * g=NULL) :
  GAGenome(RobotPathInitializer, RobotPathMutator, RobotPathCrossover,
	   &robotPathCrossSite){
    of=f; ud=u; ga=g; n = nrobots;
    list = (n ? new GAListGenome<int> * [n] : (GAListGenome<int> **)NULL);
    for(int i=0; i<n; i++){
      list[i] = new GAListGenome<int>(of,ud,ga);
      list[i]->initializationOperator(ListInitializer);
      list[i]->mutationOperator(GAListSwapMutator);
      list[i]->crossover(GAListPartialMatchCrossover, listOrderedCrossSite);
    }
  }
  RobotPathGenome(const RobotPathGenome & orig) :
  GAGenome(RobotPathInitializer, RobotPathMutator, RobotPathCrossover,
	   &robotPathCrossSite)
    {n = 0; list = NULL; copy(orig);}
  RobotPathGenome operator=(const GAGenome & arg)
    {copy(arg); return *this;}
  RobotPathGenome operator=(const RobotPathGenome & arg)
    {copy(arg); return *this;}
  virtual ~RobotPathGenome(){
    for(int i=0; i<n; i++)
      delete list[i];
    delete [] list;
  }
  virtual GAGenome *clone(const GACloneMethod flag=gaCloneContents) const {
    RobotPathGenome * cpy = new RobotPathGenome(0);
    cpy->GAGenome::copy(*this);
    cpy->n = n;
    cpy->list = new GAListGenome<int> * [n];
    for(int i=0; i<n; i++)
      cpy->list[i] = (GAListGenome<int> *)list[i]->clone(flag);
    return cpy;
  }
  virtual void copy(const GAGenome & c){
    if(&c == this) return;
    GAGenome::copy(c);	// copy the base class part
    RobotPathGenome & rc = (RobotPathGenome &)c;
    if(n == rc.n){
      for(int i=0; i<n; i++)
	list[i]->copy(*rc.list[i]);
    }
    else{
      int i;
      for(i=0; i<n; i++)
	delete list[i];
      delete [] list;
      n = rc.n;
      list = new GAListGenome<int> * [n];
      for(i=0; i<n; i++)
	list[i] = (GAListGenome<int> *)rc.list[i]->clone();
    }
  }

  friend ostream & operator<<(ostream & os, RobotPathGenome & arg)
    {arg._write(os); return(os);}
  friend istream & operator>>(istream & is, RobotPathGenome & arg)
    {arg._read(is); return(is);}

  GAListGenome<int> & path(const int i){return *list[i];}
  int npaths() const {return n;}

protected:
  int n;
  GAListGenome<int> **list;
  void _read(istream & is){for(int i=0; i<n; i++) is >> *(list[i]);}
  void _write(ostream & os){
    for(int i=0; i<n; i++)
      os << "list " << i << ":\t" << *(list[i]) << "\n";
  }

private:
  RobotPathGenome(){}
};



// These are the definitions of the initialization, mutation, and crossover
// operators for the robot path genome.
inline void 
RobotPathInitializer(GAGenome & c){
  RobotPathGenome & child = (RobotPathGenome &)c;
  for(int i=0; i<child.npaths(); i++)
    child.path(i).initialize();
}
inline int 
RobotPathMutator(GAGenome & c, float pmut){
  RobotPathGenome & child = (RobotPathGenome &)c;
  int nMut = 0;
  for(int i=0; i<child.npaths(); i++)
    nMut += child.path(i).mutate(pmut);
  return(nMut);
}
inline void 
RobotPathCrossover(GAGenome & c, const GAGenome & mom, const GAGenome & dad){
  RobotPathGenome & child = (RobotPathGenome &)c;
  RobotPathGenome & rc1 = (RobotPathGenome &)mom;
  RobotPathGenome & rc2 = (RobotPathGenome &)dad;
  for(int i=0; i<child.npaths(); i++)
    child.path(i).cross(rc1.path(i), rc2.path(i)); 
}




// These are the pick and mirror methods for the crossover site.  All these do
// is call the pick and mirror methods for the crossover sites associated with
// each embedded genome.  
//   Normally these would go into a .C file - you'll get multiply defined
// errors if you try to include this header in more than one file.
void 
RobotPathCrossSite::pick(GAGenome &c){
  RobotPathGenome & child = (RobotPathGenome &)c;
  for(int i=0; i<child.npaths(); i++)
    child.path(i).crossoverSite().pick(child.path(i));
}
void 
RobotPathCrossSite::pick(GAGenome &c, const GAGenome &a, const GAGenome &b){
  RobotPathGenome &child = (RobotPathGenome &)c;
  RobotPathGenome &mom = (RobotPathGenome &)a;
  RobotPathGenome &dad = (RobotPathGenome &)b;
  for(int i=0; i<child.npaths(); i++)
    child.path(i).crossoverSite().pick(child.path(i),mom.path(i),dad.path(i));
}
void 
RobotPathCrossSite::mirror(GAGenome &c, const GAGenome &a){
  RobotPathGenome &child = (RobotPathGenome &)c;
  RobotPathGenome &sibling = (RobotPathGenome &)a;
  for(int i=0; i<child.npaths(); i++)
    child.path(i).crossoverSite().mirror(child.path(i), sibling.path(i));
}

#endif
