/* ----------------------------------------------------------------------------
  ex17.C
  mbwall 5may95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   Example program illustrating the use of a 2DArrayGenome with trinary
alleles (-1, 0, 1).  The objective function for this program tries to 
alternate 0 and 1 then put -1 in the corners.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>


#if defined(__GNUG__)
#define NO_AUTO_INST
#endif


// This is a quick-and-easy way to get the template operators into your code.
// If you plan to use operators other than those built in to the library then
// you should not include the array operators file like this.  You can still
// override the template methods if you want to (see the _write method at the
// end of this file for example).
//   You cannot use this method if you are going to instantiate more than one
// type of the template class.  If you try to do it this way, you'll get
// functions multiply defined.  Sorry!
#define ARRAY_TYPE int	       // we're using arrays of ints
#include <ga/array2.op.h>      // now include the file with the operators in it


// If your compiler does not do automatic instantiation (e.g. g++ 2.6.8),
// then define the NO_AUTO_INST directive.
#ifdef NO_AUTO_INST
#include <ga/array2.ch.C>
#include <ga/array2.xs.C>
template class GA2DArray1PtCrossSite<int>;
template class GA2DArrayGenome<int>;
#endif



float objective(GAGenome &);

int
main(int argc, char *argv[])
{
  cout << "This program illustrates the use of a 2DArrayGenome with\n";
  cout << "three alleles.  It tries to fill a 2D array with alternating\n";
  cout << "0s and 1s, and -1s at the corners.  You will have to run it\n";
  cout << "for something like 20000 generations (using single-child-per-\n";
  cout << "generation) to see the perfect convergence.\n\n";
  cout.flush();

// Set the default values of the parameters and declare the params variable.

  int i;
  GAParameters params(40, 0.6, 0.001, 200);

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

  int aset[] = {-1,0,1};
  GAAlleleSet<int> alleles(3, aset);
  GA2DArrayGenome<int> chrom(20, 20, alleles, objective);
  GA2DArray1PtCrossSite<int> crossSite;
  chrom.initializationOperator(GA2DArrayRandomInitializer);
  chrom.mutationOperator(GA2DArrayFlipMutator);
  chrom.crossover(GA2DArray1PtCrossover, crossSite);

  GASteadyStateGA ga(chrom);
  GASigmaTruncationScaling trunc;
  ga.scalingScheme(trunc);
  ga.pReplacement(0.25);
  ga.parameters(params);
  ga.evolve();

  cout << "the ga generated:\n" << ga.bestOfAll() << endl;
  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";

  return 0;
}
 


float
objective(GAGenome & c)
{
  GA2DArrayGenome<int> & chrom = (GA2DArrayGenome<int> &)c;

  float value=0.0;
  int count=0;
  for(int i=0; i<chrom.width(); i++){
    for(int j=0; j<chrom.height(); j++){
      if(chrom.gene(i,j) == 0 && count%2 == 0)
        value += 1.0;
      if(chrom.gene(i,j) == 1 && count%2 != 0)
        value += 1.0;
      count++;
    }
  }
  value -= 4;
  value += 10*(chrom.gene(0,0) == chrom.allele(0));
  value += 10*(chrom.gene(chrom.width()-1,0) == chrom.allele(0));
  value += 10*(chrom.gene(0,chrom.height()-1) == chrom.allele(0));
  value += 10*(chrom.gene(chrom.width()-1,chrom.height()-1)==chrom.allele(0));
  return(value);
}


// Here we override the built-in write method for the 2DArray genome so
// that we get better spacing.  The default just stacks the characters one
// after another.  Here we do fixed spacing so that the -1 and 1 don't screw
// each other up.
void
GA2DArrayGenome<int>::_write(ostream & os) const 
{
  for(int j=0; j<ny; j++){
    for(int i=0; i<nx; i++){
      os.width(3);
      os << gene(i,j);
    }
    os << "\n";
  }
}
