/* ----------------------------------------------------------------------------
  ex18.C
  mbwall 28jul94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   Example program for the SimpleGA class and 2DBinaryStringGenome class.
This program reads in a 2D pattern from a data file then tries to match the
pattern in a 2D binary string genome.  The type of GA can be specified at
the command line.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>

float objective(GAGenome &);
int count=0;

int
main(int argc, char *argv[])
{
  cout << "This program is designed to compare the GA types.  You can\n";
  cout << "specify steady-state, replacement, or simple GA and tweak any\n";
  cout << "of the parameters for each of these GA types.  The objective\n";
  cout << "function tries to match a pattern read in from a file.\n\n";
  cout.flush();

// Set the default values of the parameters and declare the params variable.

  const int SIMPLE=0, STEADY_STATE=1, REPLACEMENT=2;
  int whichGA = SIMPLE;
  int i,j;
  GAParameters params(30, 0.9, 0.001, 400);
  char filename[128] = "smiley.dat";

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("ga", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": which GA do you want? (simple, ss, or repl)\n";
        exit(1);
      }
      else{
	if(strcmp(argv[i],"simple") == 0)
	  whichGA = SIMPLE;
	else if(strcmp(argv[i],"ss") == 0)
	  whichGA = STEADY_STATE;
	else if(strcmp(argv[i],"repl") == 0)
	  whichGA = REPLACEMENT;
	else
	  cerr << argv[0] << ": ga must be one of: simple, ss, or repl.\n";
        continue;
      }
    }
    else if(strcmp("file", argv[i]) == 0 || strcmp("f", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the file option needs a filename.\n";
        exit(1);
      }
      else{
	sprintf(filename, argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  f\tfilename from which to read (" << filename << ")\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      cerr << "  ga simple|ss|repl\twhich GA to use (simple)\n";
      cerr << "\n";
      exit(1);
    }
  }

// Read in the pattern from the specified file.  File format is pretty simple:
// two integers that give the height then width of the matrix, then the matrix
// of 1's and 0's (with whitespace inbetween).

  ifstream inStream(filename, ios :: in);
  if(!inStream){
    cerr << "Cannot open " << filename << " for input.\n";
    exit(1);
  }

  int height, width;
  inStream >> height;
  inStream >> width;

  short **target = new short*[width];
  for(i=0; i<width; i++)
    target[i] = new short[height];

  for(j=0; j<height; j++)
    for(i=0; i<width; i++)
      inStream >> target[i][j];

  inStream.close();

// Print out the pattern to be sure we got the right one.

  cout << "input pattern:\n";
  for(j=0; j<height; j++){
    for(i=0; i<width; i++){
      cout << (target[i][j] == 1 ? '*' : ' ') << " ";
    }
    cout << "\n";
  }
  cout << "\n"; cout.flush();

// Now create the GA and run it.

  GA2DBinaryStringGenome chrom(width, height, objective, (void *)target);
  GAStatistics stats;

  switch(whichGA){
  case STEADY_STATE:
    {
      GASteadyStateGA ga(chrom);
      ga.parameters(params);
      ga.evolve();
      ga.scores("bog.dat");
      chrom = ga.bestOfAll();
      stats = ga.statistics();
    }
    break;
  case REPLACEMENT:
    {
      GAReplacementGA ga(chrom);
      ga.parameters(params);
      ga.evolve();
      ga.scores("bog.dat");
      chrom = ga.bestOfAll();
      stats = ga.statistics();
    }
    break;
  case SIMPLE:
  default:
    {
      GASimpleGA ga(chrom);
      ga.parameters(params);
      ga.evolve();
      ga.scores("bog.dat");
      chrom = ga.bestOfAll();
      stats = ga.statistics();
    }
    break;
  }

  cout << "the ga generated:\n";
  for(j=0; j<height; j++){
    for(i=0; i<width; i++){
      cout << (chrom.gene(i,j) == 1 ? '*' : ' ') << " ";
    }
    cout << "\n";
  }
  cout << "\nthe statistics for the run are:\n" << stats;
  cout << "\nthe objective function was called " << count << " times\n";
  cout << "\nbest of generation data are in 'bog.dat'\n";
  cout.flush();

  for(i=0; i<height; i++)
    delete target[i];
  delete [] target;

  return 0;
}
 


float
objective(GAGenome & c)
{
  GA2DBinaryStringGenome & chrom = (GA2DBinaryStringGenome &)c;
  short **pattern = (short **)c.userData();

  float value=0.0;
  for(int i=0; i<chrom.width(); i++)
    for(int j=0; j<chrom.height(); j++)
      value += (float)(chrom.gene(i,j) == pattern[i][j]);

  count++;

  return(value);
}
