/* ----------------------------------------------------------------------------
  ex20.C
  mbwall 5sep95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   This example runs the royal road problem.  See the comments near the 
objective functions for details about the function itself.
   Some of this was copied (at least partially) from the galopps genetic
algorithm library and from the pga package.  I used a bunch of globals in this
example - not good programming style, but it gets the job done.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <ga/ga.h>


// This is the objective function for computing Holland's 1993 ICGA version
// of the Royal Road problem.  It has been corrected per GAList volume 7
// number 23, 8/26/93.  No bonus points are awarded for a given level until 
// it has been achieved (this fixes Holland's coding error in GAList).
//   Holland posed this problem as a challenge to test the 
// performance of genetic algorithms.  He indicated that, with the parameter 
// settings of
//
//     schemata size = 8
//     bits between schemata = 7
//     m* = 4
//     U* = 1.0
//     u = 0.3
//     v = 0.02
//
// he could attain royal_road_level 3 most of the time within
// 10,000 function evaluations.  He challenged other GA users to match or beat
// that performance.  He indicated that he used a population size of 512 to
// obtain his solutions, and did NOT use a "simple genetic algorithm."
//   The genome for this problem is a single-dimension bit string with length
// defined by the block size and gap size as:
//
//     length = (blocksize+gapsize) * (2^K)
//
// where K= 1,2,3, or 4.  Holland used K = 4.

#define NBLOCKS 16		// this number is 2^K

const int BLOCKSIZE=8;	// block size - length of target schemata
const int GAPSIZE=7;	// gap size - number of bits between target schemata
const int MSTAR=4;	// Holland's m* - up to this many bits in low level
			// block gets reward
const float USTAR=1.0;     // Holland's U* - first block earns this
const float RR_U=0.3;      // Holland's u - increment for lowest level match
const float RR_V=0.02;     // Holland's v - reward/penalty per bit

int nbits = (BLOCKSIZE+GAPSIZE)*NBLOCKS;
int blockarray[NBLOCKS];
int highestLevel=0;

float
RoyalRoad(GAGenome & c){
  GA1DBinaryStringGenome & chrom = (GA1DBinaryStringGenome &)c;

  float score = 0.0;
  int total, i, j, index, n;

// do the lowest level blocks first

  n = 0;
  for(i=0; i<NBLOCKS; i++) {
    total = 0;
    for(j=i*(BLOCKSIZE + GAPSIZE); j<i*(BLOCKSIZE+GAPSIZE)+BLOCKSIZE; j++)
      if(chrom.gene(j) == 1) total++;  // count the bits in the block
    if(total > MSTAR && total < BLOCKSIZE)
      score -= (total-MSTAR)*RR_V;
    else if(total <= MSTAR)
      score += total * RR_V;
    if(total == BLOCKSIZE) {
      blockarray[i] = 1;
      n++;
    }
    else{
      blockarray[i] = 0;
    }
  }

// bonus for filled low-level blocks

  if(n > 0) score += USTAR + (n-1)*RR_U;

// now do the higher-level blocks

  n = NBLOCKS;		// n is now number of filled low level blocks
  int proceed = 1;		// should we look at the next higher level?
  int level = 0;
  while ((n > 1) && proceed) {
    proceed = 0;
    total = 0;
      /* there are n valid blocks in the blockarray each time */
      /* round, so n=2 is the last.                           */
    for(i=0,index=0; i<(n/2)*2; i+=2,index++) {
      if(blockarray[i] == 1 && blockarray[i+1] == 1) {
	total++;
	proceed = 1;
	blockarray[index] = 1;
      }
      else{
	blockarray[index] = 0;
      }
    }
    if(total > 0){
      score += USTAR + (total-1)*RR_U;
      level++;
    }
    n /= 2;
  }

  if(highestLevel < level) highestLevel = level;

  return(score);
}











// The rest of this is standard for the GAlib examples.
int
main(int argc, char *argv[])
{
  cout << "Running Holland's Royal Road test problem with a genome that is\n";
  cout << nbits << " bits long (" << NBLOCKS << " blocks).  The parameters ";
  cout << "are as follows: \n\n";
  cout << "\tblock size: " << BLOCKSIZE << "\n";
  cout << "\t  gap size: " << GAPSIZE << "\n";
  cout << "\t        m*: " << MSTAR << "\n";
  cout << "\t        u*: " << USTAR << "\n";
  cout << "\t         u: " << RR_U << "\n";
  cout << "\t         v: " << RR_V << "\n";
  cout << "\n\n";
  cout.flush();

// Set the default values of the parameters and declare the params variable.

  GAParameters params(512, 0.9, 0.001, 10000);

// Parse the command line for arguments.

  for(int i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

  GA1DBinaryStringGenome chrom(nbits, RoyalRoad);
  GASigmaTruncationScaling trunc;
  GASteadyStateGA ga(chrom);
  ga.parameters(params);
  ga.scalingScheme(trunc);
  ga.evolve();
  ga.scores("bog.dat");

  cout << "the ga generated:\n" << ga.bestOfAll() << "\n";
  cout << "the highest level achieved was " << highestLevel << "\n";
  cout << "\nthe statistics for the run are:\n" << ga.statistics();
  cout << "\nbest-of-generation data are in 'bog.dat'\n";
  cout.flush();

  return 0;
}
