/* ----------------------------------------------------------------------------
  ex5.C
  mbwall 10dec94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   Example program for a composite genome derived from the GAGenome and
containing a 2DBinStr genome and a Bin2Dec genome.  This program uses
a steady-state GA to match a 2D pattern in the 2DBinStr part of the genome
and a sequence of numbers in the Bin2Dec part.
---------------------------------------------------------------------------- */
#define _ex5_C_

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>
#include "ex5.h"

float objective(GAGenome &);

typedef struct _UserData {
  short ** str;
  float * b2d;
} UserData;



int
main(int argc, char *argv[])
{
  cout << "This program shows how to use a composite genome.  It reads\n";
  cout << "a matrix from a data file and a set of values to be matched in\n";
  cout << "a binary-to-decimal genome then uses a steady-state GA to\n";
  cout << "match the pattern and value set.\n\n";

// Set the default values of the parameters and declare the params variable.

  int i,j, n;
  GAParameters params(40, 0.8, 0.005, 400);
  char filename1[128] = "graph.dat";
  char filename2[128] = "values.dat";

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("file1", argv[i]) == 0 || strcmp("f1", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the file1 option needs a filename.\n";
        exit(1);
      }
      else{
	sprintf(filename1, argv[i]);
        continue;
      }
    }
    else if(strcmp("file2", argv[i]) == 0 || strcmp("f2", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the file2 option needs a filename.\n";
        exit(1);
      }
      else{
	sprintf(filename2, argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  f1\t1st filename from which to read (" << filename1 << ")\n";
      cerr << "  f2\t2nd filename from which to read (" << filename2 << ")\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

// First we read in the pattern for the 2DBinStr genome.
// File format is pretty simple:
//   two integers that give the height then width of the matrix,
//   then the matrix of 1's and 0's (with whitespace inbetween).

  ifstream inStream1(filename1, ios :: in);
  if(!inStream1){
    cerr << "Cannot open " << filename1 << " for input.\n";
    exit(1);
  }

  int height, width;
  inStream1 >> height;
  inStream1 >> width;

  short **target = new short*[width];
  for(i=0; i<width; i++)
    target[i] = new short[height];

  for(j=0; j<height; j++)
    for(i=0; i<width; i++)
      inStream1 >> target[i][j];

  inStream1.close();

// Print out the pattern to be sure we got the right one.

  cout << "input pattern:\n";
  for(j=0; j<height; j++){
    for(i=0; i<width; i++){
      cout << (target[i][j] == 1 ? '*' : ' ') << " ";
    }
    cout << "\n";
  }
  cout << "\n"; cout.flush();

// Now we read in a sequence of numbers that the Bin2Dec genome is supposed
// to match for its objective.  File format is
// pretty simple:  a single integer that tells how many numbers will follow,
// then the sequence of numbers.

  ifstream inStream2(filename2, ios :: in);
  if(!inStream2){
    cerr << "Cannot open " << filename2 << " for input.\n";
    exit(1);
  }
  inStream2 >> n;
  float *sequence = new float[n];
  for(i=0; i<n; i++)
      inStream2 >> sequence[i];
  inStream2.close();

// Print out the sequence to be sure we got the right one.

  cout << "input sequence:\n";
  for(i=0; i<n; i++)
    cout << sequence[i] << " ";
  cout << "\n"; cout.flush();

// Create a phenotype then fill it with the phenotypes we will need to map to
// the values we read from the file.  The arguments to the add() method of a
// Bin2Dec phenotype are (1) number of bits, (2) min value, and (3) max value.
// The phenotype maps a floating-point number onto the number of bits that
// you designate.  Here we just make everything use 8 bits and pick a min and
// max based upon the number we read in from the file.  You can experiment with
// the number of bits and max/min values in order to make the GA work better
// or worse.

  GABin2DecPhenotype map;
  for(i=0; i<n; i++)
    map.add(8, (float)floor(sequence[i]), (float)floor(sequence[i])+1.0);

// Create an instance of our user data structure and stuff it with the values
// that we read in from the files.

  UserData mydata;
  mydata.str = target;
  mydata.b2d = sequence;

// Now create the GA and run it.  First a genome, then the GA.
  CompositeGenome chrom(width, height, map, objective, (void *)&mydata);
  GASteadyStateGA ga(chrom);
  ga.parameters(params);
  ga.evolve();


  chrom = ga.bestOfAll();
  cout << "the ga generated:\n";
  cout << "2D string:\n";
  for(j=0; j<height; j++){
    for(i=0; i<width; i++){
      cout << (chrom.binstr().gene(i,j) == 1 ? '*' : ' ') << " ";
    }
    cout << "\n";
  }
  cout << "\n";

  cout << "decimal values:\n";
  for(i=0; i<map.nPhenotypes(); i++)
    cout << chrom.bin2dec().phenotype(i) << " ";
  cout << "\n\n"; cout.flush();

  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";

// Don't forget to free up the memory we allocated in this routine.

  for(i=0; i<height; i++)
    delete target[i];
  delete [] target;
  delete [] sequence;

  return 0;
}
 


// In this objective function we try to match the pattern in the 2D part of the
// genome and match the sequence of values in the binary-to-decimal part of the
// genome.  The overall score is the sum of both parts.
float
objective(GAGenome & c)
{
  CompositeGenome & bc = (CompositeGenome &)c;
  GA2DBinaryStringGenome & str = bc.binstr();
  GABin2DecGenome & b2d = bc.bin2dec();

  int i;

  short **pattern = ((UserData *)c.userData())->str;

  float val1=0.0;
  for(i=0; i<str.width(); i++)
    for(int j=0; j<str.height(); j++)
      val1 += (float)(str.gene(i,j) == pattern[i][j]);

  float *sequence = ((UserData *)c.userData())->b2d;

  float val2=b2d.nPhenotypes();
  for(i=0; i<b2d.nPhenotypes(); i++)
    val2 += 1.0 / (1.0 + fabs(b2d.phenotype(i) - sequence[i]));

  return(val1 + val2);
}
