/* ----------------------------------------------------------------------------
  ex5.C
  mbwall 10dec94
  Copyright 1995 Massachusetts Institute of Technology

DESCRIPTION:
  multi-genome genome example

  You can use the GA library to build genomes that contain other 
genomes.  By embedding multiple genomes into one you can avoid problems
of mixing allele sets.  You can also do some pretty cool parallel evolutions.

  This genome contains a Bin2Dec genome and a 2DBinStr genome.  We
also define a crossover site specifically for use with this genome.  
---------------------------------------------------------------------------- */
#ifndef _ex5_h_
#define _ex5_h_

#include <ga/xsite.h>
#include <ga/genome.h>
#include <ga/binstr2.ch.h>
#include <ga/binstr2.xs.h>
#include <ga/bin2dec.ch.h>


// This is the definition of the crossover site object.  It is used by the
// composite genome for crossovers.  All we do here is act as arbiter and
// messenger.  When this object is told to do a pick, it tells the cross sites
// of the member genomes' cross sites to do a pick.  We do not have to 
// allocate any memory or store any data.
class CompositeCrossSite : public GACrossSite {
  GADefineIdentity("CompositeCrossSite", 202):

public:
  CompositeCrossSite(){}
  CompositeCrossSite(const CompositeCrossSite &){}
  CompositeCrossSite & operator=(const GACrossSite &){return *this;}
  CompositeCrossSite & operator=(const CompositeCrossSite &){return *this;}
  virtual ~CompositeCrossSite(){}
  virtual void copy(const GACrossSite &){}
  virtual GACrossSite * clone() const {return new CompositeCrossSite(*this);}
  void pick(GAGenome &c);
  void pick(GAGenome &c, const GAGenome&a, const GAGenome&b);
  void mirror(GAGenome &c,const GAGenome &a);
};




// We want to create only one instance of this crossover site.  It is used as
// the default crossover site for this class of genomes.
#ifdef _ex5_C_
CompositeCrossSite _ccCrossSite;
#else
extern CompositeCrossSite _ccCrossSite;
#endif




// These are the default operators for this genome class.  The are declared
// here and defined below.  See the comments below for details about what these
// do and don't do.
inline void CCInitializer(GAGenome &);
inline int  CCMutator(GAGenome &, float);
inline void CCCrossover(GAGenome &,const GAGenome &, const GAGenome &);




// This is the class definition.  We override the methods of the base class and
// define a few methods of our own to access the protected members.  The things
// we define here are the bare minimum that a genome must have to operate
// correctly.
class CompositeGenome : public GAGenome {
  GADefineIdentity("CompositeGenome", 201):

public:
  CompositeGenome(int element, int bond, GABin2DecPhenotype & p,
		  GAObjectiveFunction f=NULL,
		  void * u=NULL,
		  GA * g=NULL) :
  GAGenome(CCInitializer, CCMutator, CCCrossover, &_ccCrossSite){
    of=f; ud=u; ga=g;
    str = new GA2DBinaryStringGenome(element, bond, f, u, g);
    b2d = new GABin2DecGenome(p, f, u, g);
  }
  CompositeGenome(const CompositeGenome & orig) :
  GAGenome(CCInitializer, CCMutator, CCCrossover, &_ccCrossSite){
    str = new GA2DBinaryStringGenome(orig.binstr());
    b2d = new GABin2DecGenome(orig.bin2dec());
    copy(orig);
  }
  CompositeGenome operator=(const GAGenome & arg)
    {copy(arg); return *this;}
  CompositeGenome operator=(const CompositeGenome & arg)
    {copy(arg); return *this;}
  virtual ~CompositeGenome(){delete str; delete b2d;}
  virtual GAGenome * clone(const GACloneMethod) const 
    {return new CompositeGenome(*this);}
  virtual void copy(const GAGenome & c){
    if(&c == this) return;
    CompositeGenome & bc = (CompositeGenome &)c;
    GAGenome::copy(c);
    str->copy(*(bc.str));
    b2d->copy(*(bc.b2d));
  }

  friend ostream & operator<<(ostream & os, CompositeGenome & arg)
    {arg._write(os); return(os);}
  friend istream & operator>>(istream & is, CompositeGenome & arg)
    {arg._read(is); return(is);}

  GA2DBinaryStringGenome & binstr() const {return *str;}
  GABin2DecGenome & bin2dec() const {return *b2d;}

protected:
  GA2DBinaryStringGenome *str;
  GABin2DecGenome *b2d;
  void _read(istream & is){is >> *str; is >> *b2d;}
  void _write(ostream & os){
    os << "binstr part:\n" << *str << "\n"; 
    os << "bin2dec part:\n" << *b2d << "\n";
  }

private:
  CompositeGenome(){}
};




// These are the default initialization, mutation, and crossover operators for
// this genome class.  There's not much to them, so we make them inline.
//   You can change the initializer, mutator, and crossover operators on the
// member genomes by using the access member functions on those 
// genomes, or build them into the creator for the composite genome.
inline void CCInitializer(GAGenome & c){
  CompositeGenome & child = (CompositeGenome &)c;
  child.binstr().initialize();
  child.bin2dec().initialize();
}

inline int CCMutator(GAGenome & c, float pmut){
  CompositeGenome & child = (CompositeGenome &)c;
  int nMut = child.binstr().mutate(pmut);
  nMut += child.bin2dec().mutate(pmut);
  return(nMut);
}

inline void CCCrossover(GAGenome & c, const GAGenome & m, const GAGenome & d){
  CompositeGenome & child = (CompositeGenome &)c;
  CompositeGenome & mom = (CompositeGenome &)m;
  CompositeGenome & dad = (CompositeGenome &)d;
  child.binstr().cross(mom.binstr(), dad.binstr()); 
  child.bin2dec().cross(mom.bin2dec(), dad.bin2dec()); 
}


// These are the pick and mirror methods for the crossover site.  All they do
// is call the pick and mirror methods of the crossover sites that belong to 
// the genomes contained in the composite genome.
void 
CompositeCrossSite::pick(GAGenome & c){
  CompositeGenome & child = (CompositeGenome &)c;
  child.binstr().crossoverSite().pick(child.binstr());
  child.bin2dec().crossoverSite().pick(child.bin2dec());
}
void 
CompositeCrossSite::pick(GAGenome & c, const GAGenome & a, const GAGenome & b){
  CompositeGenome & child = (CompositeGenome &)c;
  CompositeGenome & mom = (CompositeGenome &)a;
  CompositeGenome & dad = (CompositeGenome &)b;
  child.binstr().crossoverSite().pick(child, mom.binstr(), dad.binstr());
  child.bin2dec().crossoverSite().pick(child, mom.bin2dec(), dad.bin2dec());
}
void 
CompositeCrossSite::mirror(GAGenome & c, const GAGenome & a){
  CompositeGenome & child = (CompositeGenome &)c;
  CompositeGenome & sibling = (CompositeGenome &)a;
  child.binstr().crossoverSite().mirror(child, sibling.binstr());
  child.bin2dec().crossoverSite().mirror(child, sibling.bin2dec());
}

#endif
