/* ----------------------------------------------------------------------------
  ex6.C
  mbwall 14jan95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   Example program for the tree and list genomes.  This example contains
the code to run a tree genome.  The list genome is almost the same - 
just change tree to list and modify the initialization methods.
  This program illustrates how to override default operators.  For template
genomes you MUST supply initialization, mutation, and crossover operators.
For non-template genomes (such as 1DBinaryString or Bin2Dec) you MAY supply
these operators, but if you do not a default will be assigned and used.
  There are sample operators defined in the file that contains the template
class definitions.  You can copy these as a basis for your own operators.  Or
you can use the operators in the library by including the operators file for
this type of genome (see below).
  This program also illustrates how to override member functions of the 
template classes.  Here we override the default _write() method so that we get
the contents of the nodes rather than the pointers to the node contents.  You
can override ANY member function of a template class, but there are only a few
that MUST be overridden before they are used.  See the documentation about the
template genome you want to use for more information.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>


#if defined(__GNUG__)
#define NO_AUTO_INST
#endif


// This is how we auto-include the tree operators.  This will work only if we
// are using only one type of tree - if you are going to instantiate trees of
// more than one type then you'll have to copy the code from tree.op.C 
// explicitly and replace 'TREE_TYPE' with the types you are going to use
// then change the function names so they don't conflict.
#define TREE_TYPE int
#include <ga/tree.op.h>


// If your compiler does not do automatic instantiation (e.g. g++ 2.6.8),
// then define the NO_AUTO_INST directive.
#ifdef NO_AUTO_INST
#include <ga/treetmpl.C>
#include <ga/tree.ch.C>
#include <ga/tree.xs.C>
template class GATree<int>;
template class GATreeGenome<int>;
template class GATree1PtCrossSite<int>;
#endif


// The objective function is declared here and defined below.
float objective(GAGenome &);

// This is the declaration for the initialization operator for our trees.
void TreeInitializer(GAGenome &);

// This is a recursive function that will be used in the 'write' method for 
// our tree genomes.
void WriteNode(ostream & os, GANode<int> * n);


int
main(int argc, char *argv[])
{
  cout << "This example uses a SteadyState GA and Tree<int> genome.  It\n";
  cout << "tries to maximize the size of the tree genomes that it\n";
  cout << "contains.  The genomes contain ints in its nodes.\n\n";
  cout.flush();

// Set the default values of the parameters and declare the params variable.

  int i;
  GAParameters params(30,       // population size
                      0.7,      // probability of crossover
                      0.01,     // probability of mutation
                      100);     // number of generations

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

// Now create the GA and run it.  We first create a chromsome with the
// operators we want.  Since we're using a template genome, we must assign
// all three operators.  We use the single point crossover and cross site.

  GATreeGenome<int> chrom(objective);
  GATree1PtCrossSite<int> crossSite;
  chrom.initializationOperator(TreeInitializer);
  chrom.mutationOperator(GATreeSwapSubtreeMutator);
  chrom.crossover(GATree1PtCrossover, crossSite);

// Now that we have a genome, we use it to create our GA.  After creating the
// GA we set the parameters and tell the GA to keep the best-of-generation
// score every 10 generations.

  GASteadyStateGA ga(chrom);
  ga.parameters(params);
  ga.scoreFrequency(10);
  ga.evolve();

  chrom = ga.bestOfAll();
  cout << "the ga generated a tree with " << chrom.size();
  cout << " nodes, " << chrom.depth() << " levels deep.\n";
  cout << chrom << "\n";

  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";

  return 0;
}
 


/* ----------------------------------------------------------------------------
Objective function
  All we do in this objective function is try to maximize the size of the tree.
Just return the tree size.  This means that if you run this objective function
for many generations you'll run out of memory!  There is no limit to tree or
list sizes built-in to the GA library.
---------------------------------------------------------------------------- */
float
objective(GAGenome & c)
{
  GATreeGenome<int> & chrom = (GATreeGenome<int> &)c;
  return chrom.size();
}



/* ----------------------------------------------------------------------------
Here is the initializer for our genomes.  It builds a tree of n items of type
int.  Notice that we first destroy any tree that is already in the genome 
before we do our initialization.  This is so that the genomes can be re-used.
When you re-run a GA, it does not destroy the individuals in the population - 
it reuses them.  Thus, the initializer must make sure that the genome is 
cleaned up before it tries to initialize it.
---------------------------------------------------------------------------- */
void
TreeInitializer(GAGenome & c)
{
  GATreeGenome<int> &child=(GATreeGenome<int> &)c;

// destroy any pre-existing tree
  child.root();
  child.destroy();

// Create a new tree with depth of 'depth' and each eldest node containing
// 'n' children (the other siblings have none).
  int depth=2, n=3, count=0;
  child.insert(count++,GATreeBASE::ROOT);

  for(int i=0; i<depth; i++){
    child.eldest();
    child.insert(count++);
    for(int j=0; j<n; j++)
      child.insert(count++,GATreeBASE::AFTER);
  }
}






/* ----------------------------------------------------------------------------
Tree genome method overrides
-------------------------------------------------------------------------------
  Here we override the built-in methods for the tree class.  We can do this
because the tree class is template-ized - when the compiler looks for an
instance of the tree class methods, it finds these so it won't generate an
instance from the templates.  You can do this with ANY method of a template
class.  Here we do it only for the write method.
  The default write operator prints out pointers to the contents of each node.
Here we print out the actual contents of each node.  This assumes that the 
object in our node has the operator<< defined for it.
---------------------------------------------------------------------------- */
void 
WriteNode(ostream & os, GANode<int> * n)
{
  if(!n) return;
  GANodeBASE * node = (GANodeBASE *)n;

  os.width(10);
  os << ((GANode<int> *)node)->contents << " ";
  os.width(10); 
  if(node->parent) os << ((GANode<int> *)node->parent)->contents << " ";
  else os << "." << " ";
  os.width(10); 
  if(node->child) os << ((GANode<int> *)node->child)->contents << " ";
  else os << "." << " ";
  os.width(10);
  if(node->next) os << ((GANode<int> *)node->next)->contents << " ";
  else os << "." << " ";
  os.width(10);
  if(node->prev) os << ((GANode<int> *)node->prev)->contents << "\n";
  else os << ".\n";
  WriteNode(os, (GANode<int> *)node->child);

  for(GANodeBASE * tmp=node->next; tmp && tmp != node; tmp=tmp->next){
    os.width(10);
    os << ((GANode<int> *)tmp)->contents << " ";
    os.width(10);
    if(tmp->parent) os << ((GANode<int> *)tmp->parent)->contents << " ";
    else os << "." << " ";
    os.width(10); 
    if(tmp->child) os << ((GANode<int> *)tmp->child)->contents << " ";
    else os << "." << " ";
    os.width(10); 
    if(tmp->next) os << ((GANode<int> *)tmp->next)->contents << " ";
    else os << "." << " ";
    os.width(10); 
    if(tmp->prev) os << ((GANode<int> *)tmp->prev)->contents << "\n";
    else os << ".\n";
    WriteNode(os, (GANode<int> *)tmp->child);
  }
}

void
GATreeGenome<int>::_write(ostream & os) const
{
  os << "      node     parent      child       next       prev\n";
  WriteNode(os, (GANode<int> *)rt);
}
