/* ----------------------------------------------------------------------------
  ex7.C
  mbwall 19jan95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   This example functions the same way as example 3, but this example shows
how to use some of the other member functions in the GA library.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>

float objective(GAGenome &);

int
main(int argc, char *argv[])
{
  cout << "This program reads in a data file then runs a steady-state GA \n";
  cout << "whose objective function tries to match the pattern of bits that\n";
  cout << "are in the data file.\n\n";

// Set the default values of the parameters and declare the params variable.

  int i,j;
  GAParameters params(50,	// population size
		      0.8,	// probability of crossover
		      0.001,	// probability of mutation
		      200,	// number of generations
		      gaDefPConv, // percentage convergence
		      gaDefNConv, // number of generations to convergence
		      20);	  // score frequency
  char filename[128] = "smiley.dat";

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("file", argv[i]) == 0 || strcmp("f", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the file option needs a filename.\n";
        exit(1);
      }
      else{
	sprintf(filename, argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  f\tfilename from which to read (" << filename << ")\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

// Read in the pattern from the specified file.  File format is pretty simple:
// two integers that give the height then width of the matrix, then the matrix
// of 1's and 0's (with whitespace inbetween).
//   Here we use a binary string genome to store the desired pattern.  This
// shows how you can read in directly from a stream into a genome.  (This can 
// be useful in a population initializer when you want to bias your population)

  ifstream inStream(filename, ios :: in);
  if(!inStream){
    cerr << "Cannot open " << filename << " for input.\n";
    exit(1);
  }

  int height, width;
  inStream >> height;
  inStream >> width;
  GA2DBinaryStringGenome target(width, height);
  inStream >> target;
  inStream.close();

// Print out the pattern to be sure we got the right one.

  cout << "input pattern:\n";
  for(j=0; j<height; j++){
    for(i=0; i<width; i++){
      cout << (target.gene(i,j) == 1 ? '*' : ' ') << " ";
    }
    cout << "\n";
  }
  cout << "\n"; cout.flush();

// Now create the first genome and the GA.  When we create the genome, we give
// it not only the objective function but also 'user data'.  In this case, the
// user data is a pointer to our target pattern.

  GA2DBinaryStringGenome chrom(width, height, objective, (void *)&target);
  GASteadyStateGA ga(chrom);

// When you use a GA with overlapping populations, the default score
// frequency (how often the best of generation score is recorded) defaults
// to 100.  We use the parameters member function to change this value (along
// with all of the other parameters we set above).  You can also change the 
// score frequency using the scoreFrequency member function of the GA.  Each of
// the parameters can be set individually if you like.

  ga.parameters(params);

// The default selection method is RouletteWheel.  Here we set the selection
// method to TournamentSelection.

  ga.selectionFunction(GATournamentSelector);

// We can control the amount of overlap from generation to generation using the
// pReplacement member function.  If we specify a value of 1 (100%) then the
// entire population is replaced each generation.  Notice that the percentage
// must be high enough to have at least one individual produced in each 
// generation.  If not, the GA will post a warning message.

  ga.pReplacement(0.3);

// Often we use the number of generations as the criterion for terminating the
// GA run.  Here we override that and tell the GA to use convergence as a 
// termination criterion.  Note that you can pass ANY function as the stopping
// criterion (as long as it has the correct signature).
//   Notice that the values we set here for p- and n-convergence override those
// that we set in the parameters object.

  ga.completionFunction(GATerminateUponConvergence);
  ga.pConvergence(0.01);	// converge to within 1%
  ga.nConvergence(100);		// within the last 200 generations

// Evolve the GA 'by hand'.  When you use this method, be sure to initialize
// the GA before you start to evolve it.  You can print out the status of the
// current population by using the ga.population() member function.  This is
// also how you would print the status of the GA to disk along the way (rather
// than waiting to the end then printing the scores, for example).

  ga.initialize();
  while(!ga.done()){
    ++ga;
  }

// Now that the GA is finished, we set our default genome to equal the contents
// of the best genome that the GA found.  Then we print it out.

  chrom = ga.bestOfAll();
  cout << "the ga generated:\n";
  for(j=0; j<height; j++){
    for(i=0; i<width; i++)
      cout << (chrom.gene(i,j) == 1 ? '*' : ' ') << " ";
    cout << "\n";
  }
  cout << "\n"; cout.flush();

  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";

  return 0;
}
 



// This objective function just tries to match the genome to the pattern in the
// user data.
float
objective(GAGenome & c)
{
  GA2DBinaryStringGenome & chrom = (GA2DBinaryStringGenome &)c;
  GA2DBinaryStringGenome * pattern = 
    (GA2DBinaryStringGenome *)c.userData();

  float value=0.0;
  for(int i=0; i<chrom.width(); i++)
    for(int j=0; j<chrom.height(); j++)
      value += (float)(chrom.gene(i,j) == pattern->gene(i,j));
  return(value);
}
