/* ----------------------------------------------------------------------------
  ex8.C
  mbwall 14jan95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   Example program for the list genome.  This example contains
the code to run a genetic algorithm with a list genome.
  This program illustrates how to override default operators.  For template
genomes you MUST supply initialization, mutation, and crossover operators.
For non-template genomes (such as 1DBinaryString or Bin2Dec) you MAY supply
these operators, but if you do not a default will be assigned and used.
  There are sample operators defined in the file that contains the template
class definitions.  You can copy these as a basis for your own operators.  Or
you can use the operators in the library by including the operators file for
this type of genome (see below).
  This program also illustrates how to override member functions of the 
template classes.  Here we override the default _write() method so that we get
the contents of the nodes rather than the pointers to the node contents.  You
can override ANY member function of a template class, but there are only a few
that MUST be overridden before they are used.  See the documentation about the
template genome you want to use for more information.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>


#if defined(__GNUG__)
#define NO_AUTO_INST
#endif


// This is how we auto-include the list operators.  This will work only if we
// are using only one type of list - if you are going to instantiate lists of
// more than one type then you'll have to copy the code from list.op.C 
// explicitly and replace 'LIST_TYPE' with the types you are going to use
// then change the function names so they don't conflict.
#define LIST_TYPE int
#include <ga/list.op.h>


// If your compiler does not do automatic instantiation (e.g. g++ 2.6.8),
// then define the NO_AUTO_INST directive.
#ifdef NO_AUTO_INST
#include <ga/listtmpl.C>
#include <ga/list.ch.C>
#include <ga/list.xs.C>
template class GAList<int>;
template class GAListGenome<int>;
template class GAList1PtCrossSite<int>;
template class GAListOrderedCrossSite<int>;
#endif


// Objective function and initializer declarations.
float objective(GAGenome &);
void ListInitializer(GAGenome &);


int
main(int argc, char *argv[])
{
  cout << "This program runs a steady-state GA whose objective function\n";
  cout << "tries to maximize the size of the list and tries to make lists\n";
  cout << "that contain the number 101 in the nodes.  The lists contain\n";
  cout << "ints in the nodes.\n\n";
  cout.flush();

// Set the default values of the parameters and declare the params variable.

  int i;
  GAParameters params(40,	// population size
		      0.6,	// probability of crossover
		      0.05,	// probability of mutation
		      200);	// number of generations

// Parse the command line for arguments.

  for(i=1; i<argc; i++){
    if(strcmp("ngen", argv[i]) == 0 || strcmp("n", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": number of generations needs a value.\n";
        exit(1);
      }
      else{
	params.nGen = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pop", argv[i]) == 0 || strcmp("p", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": population size needs a value.\n";
        exit(1);
      }
      else{
	params.popSize = atoi(argv[i]);
        continue;
      }
    }
    else if(strcmp("pcross", argv[i]) == 0 || strcmp("c", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the crossover option needs a value.\n";
        exit(1);
      }
      else{
	params.pCross = (float)atof(argv[i]);
        continue;
      }
    }
    else if(strcmp("pmut", argv[i]) == 0 || strcmp("m", argv[i]) == 0){
      if(++i >= argc){
        cerr << argv[0] << ": the mutation option needs a value.\n";
        exit(1);
      }
      else{
	params.pMut = (float)atof(argv[i]);
        continue;
      }
    }
    else {
      cerr << argv[0] << ":  unrecognized arguement: " << argv[i] << "\n\n";
      cerr << "valid arguements include:\n";
      cerr << "  n\tnumber of generations (" << params.nGen << ")\n";
      cerr << "  p\tpopulation size (" << params.popSize << ")\n";
      cerr << "  c\tcrossover probability (" << params.pCross << ")\n";
      cerr << "  m\tmutation probability (" << params.pMut << ")\n";
      cerr << "\n";
      exit(1);
    }
  }

// Now create the GA and run it.  We first create a genome with the
// operators we want.  Since we're using a template genome, we must assign
// all three operators.  We use the single point cross site.

  GAListGenome<int> chrom(objective);
  GAList1PtCrossSite<int> crossSite;
  chrom.initializationOperator(ListInitializer);
  chrom.mutationOperator(GAListSwapMutator);
  chrom.crossover(GAList1PtCrossover, crossSite);

// Now that we have a genome, we use it to create our GA.  After creating the
// GA we set the parameters and tell the GA to use sigma truncation scaling
// rather than the default (linear scaling).

  GASteadyStateGA ga(chrom);
  ga.parameters(params);
  GASigmaTruncationScaling trunc;
  ga.scalingScheme(trunc);
  ga.evolve();

  chrom = ga.bestOfAll();
  cout << "the ga generated a list with " << chrom.size() << " nodes\n";
  cout << chrom << "\n";

  ga.scores("bog.dat");
  cout << "best of generation data are in 'bog.dat'\n";

  return 0;
}
 


/* ----------------------------------------------------------------------------
Objective function
  There is no limit to the size of a list (only the memory you have on your
computer).  This objective function tries to build a list that contains the
number 101 in all of its nodes.  If we don't put some control on this objective
then the list will grow without bound.  So we dampen it a bit with a penalty 
for large size.  However, this will make the score go negative, so we must use
a scaling object that allows negative objective scores.
  We could get lists with no contents, so we have to check for that.  Just make
sure that head has contents.  If it does, then the rest of the genome will.
  When we access the node contents we have to dereference the member functions.
For example, we do *chrom.head() not chrom.head().  This is because the member
functions return a pointer to the node's contents, not the actual contents.
---------------------------------------------------------------------------- */
float
objective(GAGenome & c)
{
  GAListGenome<int> & chrom = (GAListGenome<int> &)c;
  int count=0;
  if(!chrom.head()) return 0;
  count = (*chrom.head() == 101) ? 1 : 0; // move to head of the list
  for(int i=1; i<chrom.size(); i++)
    count += (*chrom.next() == 101) ? 1 : 0; // check each element of the list
  return 5*count - chrom.size();
}



/* ----------------------------------------------------------------------------
Here is the initializer for our genomes.  It builds a list of n items of type
int.  Notice that we first destroy any list that is already in the genome 
before we do our initialization.  This is so that the genomes can be re-used.
When you re-run a GA, it does not destroy the individuals in the population - 
it reuses them.  Thus, the initializer must make sure that the genome is 
cleaned up before it tries to initialize it.
---------------------------------------------------------------------------- */
void
ListInitializer(GAGenome & c)
{
  GAListGenome<int> &child=(GAListGenome<int> &)c;

  while(child.head()) child.destroy(); // destroy any pre-existing list

  int n=75;
  child.insert(100,GAListBASE::HEAD);
  for(int i=0; i<n; i++)
    child.insert(i+100);
}





/* ----------------------------------------------------------------------------
List genome method overrides
-------------------------------------------------------------------------------
  Here we override the built-in methods for the list class.  We can do this
because the list class is template-ized - when the compiler looks for an
instance of the list class methods, it finds these so it won't generate an
instance from the templates.  You can do this with ANY method of a template
class.  Here we do it only for the write method.
  The default write operator prints out pointers to the contents of each node.
Here we print out the actual contents of each node.  This assumes that the 
object in our node has the operator<< defined for it.
---------------------------------------------------------------------------- */
void
GAListGenome<int>::_write(ostream & os) const
{
  int *cur, *head;
  GAListIter<int> iter(*this);
  if((head=iter.head()) != NULL) os << *head << " ";
  for(cur=iter.next(); cur && cur != head; cur=iter.next())
    os << *cur << " ";
}
