/* ----------------------------------------------------------------------------
  bitstr.C
  mbwall 13sep95
  Copyright 1995 Massachusetts Institute of Technology

  This code can be freely distributed and modified under the terms of the GNU
  public license.   See the COPYING file for details.

 DESCRIPTION:
  Source file for the BitString genome (derived from the GNU BitString object).
---------------------------------------------------------------------------- */
#include <ga/random.h>
#include "bitstr.h"

// Set all the initial values to zero.  We do NOT call the initialize method at
// this point - initialization must be done explicitly by the user of the
// genome (eg when the population is created or reset).  If we called the
// initializer routine here then we could end up with multiple initializations
// and/or calls to dummy initializers (for example when the genome is 
// created with a dummy initializer and the initializer is assigned later on).
BitStringGenome::
BitStringGenome(const unsigned int l, 
		GAObjectiveFunction f,
		void * u,
		GA * g) :
GAGenome(BitStringUniformRandomInitializer, 
	 BitStringUniformRandomFlipMutator,
	 BitStringUniformCrossover,
	 (GACrossSite *)0),
BitString(){
  of=f; ud=u; ga=g;
  for(int i=0; i<l; i++)
    *this += 1;
}


// This is the class-specific copy method.  It will get called by the super
// class since the superclass operator= is set up to call ccopy (and that is
// what we define here - a virtual function).  We should check to be sure that
// both genomes are the same class.
void
BitStringGenome::copy(const GAGenome & orig){
  if(&orig == this) return;
  if(!sameClass(orig)){
    GAErr(className(), "copy", gaErrObjectTypeMismatch);
    return;
  }
  GAGenome::copy(orig);
  BitStringGenome &bsg = (BitStringGenome &)orig;
  BitString::operator=(bsg._substr(0,bsg.length()));
}


// The clone method basically does the same thing as the copy method, but here
// we return a pointer to a completely new genome (the copy method just fills
// the current genome with the copied contents).  It is the responsibility of 
// the caller to free the memory returned by this routine.
GAGenome *
BitStringGenome::clone(const GACloneMethod) const {
  return new BitStringGenome(*this);
}







// These are the operators used by the BitStringGenome.


// The random initializer sets the bits in the bit string randomly to 0 or 1.
// It uses the GARandomBit function to do this (GARandomBit is pretty efficient
// in terms of its calls to your system's random function).
void 
BitStringUniformRandomInitializer(GAGenome & c)
{
  BitStringGenome &genome=(BitStringGenome &)c;
  for(int i=genome.length()-1; i>=0; i--)
    genome.gene(i, GARandomBit());
}


// Randomly pick bits in the bit string then flip their values.  We try to be
// smart about the number of times we have to call any random functions.  If
// the requested likliehood is small enough (relative to the number of bits in
// the genome) then we must do a weighted coin toss on each bit in the genome.
// Otherwise, we just do the expected number of flips (note that this will not
// guarantee the requested mutation rate, but it will come close when the
// length of the bit string is long enough).
int 
BitStringUniformRandomFlipMutator(GAGenome & c, float pmut)
{
  BitStringGenome &genome=(BitStringGenome &)c;
  register int n, i;
  if(pmut <= 0.0) return(0);

  float nMut = pmut * (float)(genome.length());
  if(nMut < 1.0){		// we have to do a flip test on each bit
    nMut = 0;
    for(i=genome.length()-1; i>=0; i--){
      if(GAFlipCoin(pmut)){
	genome.gene(i, genome.gene(i) ? 0 : 1);
	nMut++;
      }
    }
  }
  else{				// only flip the number of bits we need to flip
    for(n=1; n<nMut; n++){
      i = GARandomInt(0, genome.length()-1); // the index of the bit to flip
      genome.gene(i, genome.gene(i) ? 0 : 1);
    }
  }
  return((int)nMut);
}


// Randomly take elements from each parent.  This implementation does not use
// a crossover site.  To do uniform crossover properly we should keep a bit
// mask so that each child will be properly mirrored during a dual child
// crossover operation.
void
BitStringUniformCrossover(GAGenome &c, const GAGenome &c1, const GAGenome &c2)
{
  BitStringGenome &child=(BitStringGenome &)c;
  BitStringGenome &mom=(BitStringGenome &)c1;
  BitStringGenome &dad=(BitStringGenome &)c2;

  for(int i=child.length()-1; i>=0; i--)
    child.gene(i, (GARandomBit() ? mom.gene(i) : dad.gene(i)));
}
