// $Header: /usr/people/mbwall/src/ga/ga/RCS/array1.ch.C,v 1.2 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  array1.ch.C
  mbwall 25feb95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Source file for the 1D array genome.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <string.h>
#include <ga/array1.ch.h>


/* ----------------------------------------------------------------------------
1DArrayGenome
---------------------------------------------------------------------------- */
// Set all the initial values to NULL or zero, then allocate the space we'll
// need (using the resize method).  We do NOT call the initialize method at
// this point - initialization must be done explicitly by the user of the
// genome (eg when the population is created or reset).  If we called the
// initializer routine here then we could end up with multiple initializations
// and/or calls to dummy initializers (for example when the genome is 
// created with a dummy initializer and the initializer is assigned later on).
template <class T> 
GA1DArrayGenome<T>::
GA1DArrayGenome(const unsigned int length, 
		GAAlleleSet<T> & s,
		GAObjectiveFunction f,
		void * u,
		GA * g) :
GAGenome(GANoInitializer, GANoMutator, GANoCrossover, (GACrossSite *)0),
GAArray<T>(length){
  of=f; ud=u; ga=g; as=&s;
  nx=minX=maxX=length;
}


// This is the copy initializer.  We set everything to the default values, then
// copy the original.  The ArrayGenome creator takes care of zeroing
// the data and sz members.
template <class T> 
GA1DArrayGenome<T>::
GA1DArrayGenome(const GA1DArrayGenome<T> & orig) : GAArray<T>(orig.sz){
  GA1DArrayGenome<T>::copy(orig);
}


// This is the class-specific copy method.  It will get called by the super
// class since the superclass operator= is set up to call ccopy (and that is
// what we define here - a virtual function).  We should check to be sure that
// both genomes are the same class and same dimension.  This function tries
// to be smart about they way it copies.  If we already have data, then we do
// a memcpy of the one we're supposed to copy.  If we don't or we're not the 
// same size as the one we're supposed to copy, then we adjust ourselves.
//   The ArrayGenome takes care of the resize in its copy method.
template <class T> void
GA1DArrayGenome<T>::copy(const GAGenome & orig){
  if(&orig == this) return;
  if(!sameClass(orig)){
    GAErr(className(), "copy", gaErrObjectTypeMismatch);
    return;
  }
  GA1DArrayGenome<T> & c = (GA1DArrayGenome<T> &)orig;
  GAGenome::copy(c);
  GAArray<T>::copy(c);
  as = c.as;
  nx = c.nx; minX = c.minX; maxX = c.maxX;
}


template <class T> GAGenome *
GA1DArrayGenome<T>::clone(const GACloneMethod flag) const {
  GA1DArrayGenome<T> *cpy = new GA1DArrayGenome<T>(nx,*as);
  if(flag == gaCloneContents){ cpy->copy(*this);}
  else{
    cpy->GAGenome::copy(*this); 
    cpy->as = as; cpy->maxX = maxX; cpy->minX = minX;
  }
  return cpy;
}


//   Resize the genome.
//   A negative value for the length means that we should randomly set the
// length of the genome (if the resize behaviour is resizeable).  If
// someone tries to randomly set the length and the resize behaviour is fixed
// length, then we don't do anything.
//   We pay attention to the values of minX and maxX - they determine what kind
// of resizing we are allowed to do.  If a resize is requested with a length
// less than the min length specified by the behaviour, we set the minimum 
// to the length.  If the length is longer than the max length specified by
// the behaviour, we set the max value to the length.
//   We return the total size (in bits) of the genome after resize. 
template <class T> int
GA1DArrayGenome<T>::resize(const int l)
{
  int length=l;
  if(length == nx) return nx;

  if(minX == maxX){		// resizing is not allowed for this one
    if(length < 0) return nx;	// do nothing
    minX = maxX = length;	// resize to specified length
  }
  else if(length >=0){
    if(length < minX) minX = length;
    if(length > maxX) maxX = length;
  }
  else length = GARandomInt(minX, maxX);

  nx = GAArray<T>::size(length);
  _evaluated = gaFalse; _nmut = 0;
  return sz;
}


// We read data from a stream as a series of objects separated by whitespace.
// We assume that each object has operator >> defined for it.
template <class T> void
GA1DArrayGenome<T>::_read(istream & is)
{
  int i=0;
  do{
    is >> gene(i++);
  } while(is && is.peek() != EOF && i < nx);

  if(is.peek() == EOF && i < nx)
    GAErr(className(), "_read", gaErrUnexpectedEOF);
}


// When we write the data to a stream we do it without any spaces.  Also, there
// is no newline at the end of the stream of digits.
template <class T> void
GA1DArrayGenome<T>::_write(ostream & os) const
{
  for(int i=0; i<nx; i++)
    os << gene(i);
}


//   Set the resize behaviour of the genome.  A genome can be fixed
// length, resizeable with a max and min limit, or resizeable with no limits
// (other than an implicit one that we use internally).
//   A value of 0 means no resize, a value less than zero mean unlimited 
// resize, and a positive value means resize with that value as the limit.
template <class T> int
GA1DArrayGenome<T>::
resizeBehaviour(const unsigned int lower, const unsigned int upper)
{
  if(upper < lower){
    GAErr(className(), "resizeBehaviour", gaErrBadResizeBehaviour);
    return resizeBehaviour();
  }
  minX = lower; maxX = upper;
  if(nx > upper) GA1DArrayGenome<T>::resize(upper);
  if(nx < lower) GA1DArrayGenome<T>::resize(lower);
  return resizeBehaviour();
}
