// $Header: /usr/people/mbwall/src/ga/ga/RCS/array1.op.C,v 1.2 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  array1.op.C
  mbwall 25feb95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Outline code for doing built-in operators for the 1D array genome.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <ga/config.h>
#include <ga/random.h>
#include <ga/array1.ch.h>
#include <ga/array1.xs.h>

#if USE_ORDER_CROSSOVER == 1
int GA1DArrayIsHole(GA1DArrayGenome<ARRAY_TYPE> &, 
		    const GA1DArrayGenome<ARRAY_TYPE> &, int, int, int);
#endif


/* ----------------------------------------------------------------------------
1D Array Genome
---------------------------------------------------------------------------- */
// The random initializer sets the elements of the array based on the alleles
// set.  We choose randomly the allele for each element.
#if USE_RANDOM_INITIALIZER == 1
void 
GA1DArrayRandomInitializer(GAGenome & c)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  child.resize(gaResizeAny);			// let chrom resize if it can
  for(int i=child.length()-1; i>=0; i--)
    child.gene(i, child.alleles().allele());
}
#endif


// Random initializer for order-based genome.  Loop through the genome
// and assign each element the next allele in the allele set.  Once each
// element has been initialized, scramble the contents by swapping elements.
#if USE_ORDERED_INITIALIZER == 1
void 
GA1DArrayOrderedInitializer(GAGenome & c)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  child.resize(gaResizeAny);			// let chrom resize if it can
  int length = child.length()-1;
  int n=0;
  int i;
  for(i=length; i>=0; i--){
    child.gene(i, child.alleles().allele(n++));
    if(n >= child.alleles().size()) n = 0;
  }
  for(i=length; i>=0; i--)
    child.swap(i, GARandomInt(0, length));
}
#endif


// Randomly pick elements in the array then set the element to any of the 
// alleles in the allele set for this genome.
#if USE_FLIP_MUTATOR == 1
int 
GA1DArrayFlipMutator(GAGenome & c, float pmut)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  register int n, i;
  if(pmut <= 0.0) return(0);

  float nMut = pmut * (float)(child.length());
  if(nMut < 1.0){		// we have to do a flip test on each bit
    nMut = 0;
    for(i=child.length()-1; i>=0; i--){
      if(GAFlipCoin(pmut)){
	child.gene(i, child.alleles().allele());
	nMut++;
      }
    }
  }
  else{				// only flip the number of bits we need to flip
    for(n=1; n<nMut; n++){
      i = GARandomInt(0, child.length()-1); // the index of the bit to flip
      child.gene(i, child.alleles().allele());
    }
  }
  return((int)nMut);
}
#endif


// Randomly swap elements in the array.
#if USE_SWAP_MUTATOR == 1
int 
GA1DArraySwapMutator(GAGenome & c, float pmut)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  register int n, i;
  if(pmut <= 0.0) return(0);

  float nMut = pmut * (float)(child.length());
  int length = child.length()-1;
  if(nMut < 1.0){		// we have to do a flip test on each bit
    nMut = 0;
    for(i=length; i>=0; i--){
      if(GAFlipCoin(pmut)){
	child.swap(i, GARandomInt(0, length));
	nMut++;
      }
    }
  }
  else{				// only flip the number of bits we need to flip
    for(n=1; n<nMut; n++)
      child.swap(GARandomInt(0, length), GARandomInt(0, length));
  }
  return((int)nMut);
}
#endif



// Randomly take elements from each parent.  No crossover site is needed for 
// this crossover operator.  (I'll put a mask in the next version.  For now
// we don't use one.)
#if USE_UNIFORM_CROSSOVER == 1
void
GA1DArrayUniformCrossover(GAGenome &c, const GAGenome &c1, const GAGenome &c2)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  GA1DArrayGenome<ARRAY_TYPE> &mom=(GA1DArrayGenome<ARRAY_TYPE> &)c1;
  GA1DArrayGenome<ARRAY_TYPE> &dad=(GA1DArrayGenome<ARRAY_TYPE> &)c2;

  if(child.resizeBehaviour() == gaNoResize){
    if(mom.length() != dad.length()){
      GAErr(c.className(), "GA1DArrayUniformCrossover", 
	    gaErrParentSizeMismatch);
      return;
    }
    if(child.length() != mom.length()){
      GAErr(c.className(), "GA1DArrayUniformCrossover",
	    gaErrChildSizeMismatch);
      return;
    }
  }
  else{
    GAErr(c.className(), "GA1DArrayUniformCrossover", gaErrFixedOnly);
    return;
  }
  
  for(int i=child.length()-1; i>=0; i--)
    child.gene(i, (GARandomBit() ? mom.gene(i) : dad.gene(i)));
}
#endif


// Single point crossover for 1D array genomes.  Pick a single point then
// copy genetic material from each parent.
#if USE_1PT_CROSSOVER == 1
void
GA1DArray1PtCrossover(GAGenome & c,
		       const GAGenome & c1, const GAGenome & c2)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  GA1DArrayGenome<ARRAY_TYPE> &mom=(GA1DArrayGenome<ARRAY_TYPE> &)c1;
  GA1DArrayGenome<ARRAY_TYPE> &dad=(GA1DArrayGenome<ARRAY_TYPE> &)c2;
  GA1DArray1PtCrossSite<ARRAY_TYPE> & xsite = 
    (GA1DArray1PtCrossSite<ARRAY_TYPE> &)child.crossoverSite();

  unsigned int x2;
  if(child.resizeBehaviour() == gaNoResize){
    if(mom.length() != dad.length()){
      GAErr(c.className(), "GA1DArray1PtCrossover", gaErrParentSizeMismatch);
      return;
    }
    if(child.length() != mom.length()){
      GAErr(c.className(), "GA1DArray1PtCrossover", gaErrChildSizeMismatch);
      return;
    }
    if(xsite.x == gaNoSite) xsite.pick(child);
    xsite.x1 = xsite.x2 = xsite.x;
    x2 = child.length() - xsite.x;
  }
  else{
    if(xsite.x1 == gaNoSite || xsite.x2 == gaNoSite) 
      xsite.pick(child, mom, dad);
    xsite.x = xsite.x1;
    x2 = dad.length() - xsite.x2;
    child.resize(xsite.x+x2);
  }
  child.copy(mom, 0, 0, xsite.x1);
  child.copy(dad, xsite.x1, xsite.x2, x2);
}
#endif


// Two point crossover for the 1D array genome.
#if USE_2PT_CROSSOVER == 1
void
GA1DArray2PtCrossover(GAGenome & c,
		      const GAGenome & c1, const GAGenome & c2)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  GA1DArrayGenome<ARRAY_TYPE> &mom=(GA1DArrayGenome<ARRAY_TYPE> &)c1;
  GA1DArrayGenome<ARRAY_TYPE> &dad=(GA1DArrayGenome<ARRAY_TYPE> &)c2;
  GA1DArray2PtCrossSite<ARRAY_TYPE> & xsite = 
    (GA1DArray2PtCrossSite<ARRAY_TYPE> &)child.crossoverSite();

  unsigned int dadlength, momlength;
  if(child.resizeBehaviour() == gaNoResize){
    if(mom.length() != dad.length()){
      GAErr(c.className(), "GA1DArray2PtCrossover", gaErrParentSizeMismatch);
      return;
    }
    if(child.length() != mom.length()){
      GAErr(c.className(), "GA1DArray2PtCrossover", gaErrChildSizeMismatch);
      return;
    }
    if(xsite.x[0] == gaNoSite || xsite.x[1] == gaNoSite) xsite.pick(child);
    dadlength = xsite.x[1] - xsite.x[0];
    momlength = mom.length() - xsite.x[1];
    xsite.x1[0] = xsite.x2[0] = xsite.x[0];
    xsite.x1[1] = xsite.x2[1] = xsite.x[1];
  }
  else{
    if(xsite.x1[0] == gaNoSite || xsite.x2[0] == gaNoSite ||
       xsite.x1[1] == gaNoSite || xsite.x2[1] == gaNoSite)
      xsite.pick(child, mom, dad);
    dadlength = xsite.x2[1] - xsite.x2[0];
    momlength = mom.length() - xsite.x1[1];
    xsite.x[0] = xsite.x1[0];
    xsite.x[1] = xsite.x[0] + dadlength;
    child.resize(xsite.x[0]+dadlength+momlength);
  }
  child.copy(mom, 0, 0, xsite.x[0]);
  child.copy(dad, xsite.x[0], xsite.x2[0], dadlength);
  child.copy(mom, xsite.x[1], xsite.x1[1], momlength);
}
#endif


// Even and odd crossover for the array works just like it does for the 
// binary strings.  For even crossover we take the 0th element and every other
// one after that from the mother.  The 1st and every other come from the
// father.  For odd crossover, we do just the opposite.
//   Notice that this crossover method does not make use of a crossover site.
#if USE_EVEN_CROSSOVER == 1
void
GA1DArrayEvenCrossover(GAGenome & c,
			const GAGenome & c1, const GAGenome & c2)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  GA1DArrayGenome<ARRAY_TYPE> &mom=(GA1DArrayGenome<ARRAY_TYPE> &)c1;
  GA1DArrayGenome<ARRAY_TYPE> &dad=(GA1DArrayGenome<ARRAY_TYPE> &)c2;

  if(child.length() > mom.length() || child.length() > dad.length()){
    GAErr(c.className(), "GA1DArrayEvenCrossover", gaErrChildSizeMismatch);
    return;
  }
  for(int i=0; i<child.length(); i++)
    child.gene(i, ((i%2 == 0) ? mom.gene(i) : dad.gene(i)));
}
#endif

#if USE_ODD_CROSSOVER == 1
void
GA1DArrayOddCrossover(GAGenome & c,
		       const GAGenome & c1, const GAGenome & c2)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  GA1DArrayGenome<ARRAY_TYPE> &mom=(GA1DArrayGenome<ARRAY_TYPE> &)c1;
  GA1DArrayGenome<ARRAY_TYPE> &dad=(GA1DArrayGenome<ARRAY_TYPE> &)c2;

  if(child.length() > mom.length() || child.length() > dad.length()){
    GAErr(c.className(), "GA1DArrayOddCrossover", gaErrChildSizeMismatch);
    return;
  }
  for(int i=0; i<child.length(); i++)
    child.gene(i, ((i%2 == 0) ? dad.gene(i) : mom.gene(i)));
}
#endif


// Partial match crossover for the 1D array genome.  This uses the partial
// matching algorithm described in Goldberg's book.
//   We do not check for size of parents and child, but they must be the same
// for this crossover operator.  If they are not, things could get really ugly.
#if USE_PARTIAL_MATCH_CROSSOVER == 1
void
GA1DArrayPartialMatchCrossover(GAGenome & c,
			       const GAGenome &c1, const GAGenome &c2)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  GA1DArrayGenome<ARRAY_TYPE> &mom=(GA1DArrayGenome<ARRAY_TYPE> &)c1;
  GA1DArrayGenome<ARRAY_TYPE> &dad=(GA1DArrayGenome<ARRAY_TYPE> &)c2;
  GA1DArrayOrderedCrossSite<ARRAY_TYPE> & xsite = 
    (GA1DArrayOrderedCrossSite<ARRAY_TYPE> &)child.crossoverSite();

  if(xsite.a == gaNoSite || xsite.b == gaNoSite) xsite.pick(c);

  child.GAArray<ARRAY_TYPE>::copy(mom);
  for(int i=xsite.a, index=xsite.a; i<xsite.b; i++, index++){
    int j;
    for(j=0; j<child.length()-1 && child[j] != dad[index]; j++);
    child.swap(i,j);
  }
}
#endif


// Order crossover for the 1D array genome.  This uses the order crossover
// described in Goldberg's book.
//   We assume that b will be greater than a.
//   We do not check for size of parents and child, but they must be the same
// for this crossover operator.  If they are not, things could get really ugly.
//   This implementation isn't terribly smart.  For example, I do a linear
// search rather than caching and doing binary search or smarter hash tables.
#if USE_ORDER_CROSSOVER == 1
int
GA1DArrayIsHole(GA1DArrayGenome<ARRAY_TYPE> &c, 
		const GA1DArrayGenome<ARRAY_TYPE> &dad,
		int index, int a, int b){
  for(int i=a; i<b; i++)
    if(c[index] == dad[i]) return 1;
  return 0;
}

void
GA1DArrayOrderCrossover(GAGenome & c,
			  const GAGenome & c1, const GAGenome & c2)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  GA1DArrayGenome<ARRAY_TYPE> &mom=(GA1DArrayGenome<ARRAY_TYPE> &)c1;
  GA1DArrayGenome<ARRAY_TYPE> &dad=(GA1DArrayGenome<ARRAY_TYPE> &)c2;
  GA1DArrayOrderedCrossSite<ARRAY_TYPE> & xsite = 
    (GA1DArrayOrderedCrossSite<ARRAY_TYPE> &)child.crossoverSite();

  if(xsite.a == gaNoSite || xsite.b == gaNoSite) xsite.pick(c);

  child.GAArray<ARRAY_TYPE>::copy(mom);

// Move all the 'holes' into the crossover section and maintain the ordering of
// the non-hole elements.
  int i, index;
  for(i=0, index=xsite.b; i<child.size(); i++, index++){
    if(index >= child.size()) index=0;
    if(GA1DArrayIsHole(child,dad,index,xsite.a,xsite.b)) break;
  }

  for(; i<child.size()-xsite.b+xsite.a; i++, index++){
    if(index >= child.size()) index=0;
    int j=index;
    do{
      j++;
      if(j >= child.size()) j=0;
    } while(GA1DArrayIsHole(child,dad,j,xsite.a,xsite.b));
    child.swap(index,j);
  }

// Now put the 'holes' in the proper order within the crossover section.
  for(i=xsite.a; i<xsite.b; i++){
    if(child[i] != dad[i]){
      for(int j=i+1; j<xsite.b; j++)
	if(child[j] == dad[i]) child.swap(i,j);
    }
  }
}
#endif


// Cycle crossover for the 1D array genome.  This is implemented as described 
// in goldberg's book.  The first is picked from mom, then cycle using dad.
// Finally, fill in the gaps with the elements from dad.
//   This crossover method does not need a crossover site.
//   We allocate space for a temporary array in this routine.  It never frees
// the memory that it uses, so you might want to re-think this if you're really
// memory-constrained.
#if USE_CYCLE_CROSSOVER == 1
void
GA1DArrayCycleCrossover(GAGenome & c,
			const GAGenome & c1, const GAGenome & c2)
{
  GA1DArrayGenome<ARRAY_TYPE> &child=(GA1DArrayGenome<ARRAY_TYPE> &)c;
  GA1DArrayGenome<ARRAY_TYPE> &mom=(GA1DArrayGenome<ARRAY_TYPE> &)c1;
  GA1DArrayGenome<ARRAY_TYPE> &dad=(GA1DArrayGenome<ARRAY_TYPE> &)c2;

  static short *flag=(short *)0;
  static int N=0;

// Allocate space for an array of flags.  We use this to keep track of whether
// the child's contents came from the mother or the father.  We don't free the
// space here, but it is not a memory leak.
  if(N < child.size()){
    delete [] flag;
    N = child.size();
    flag = new short [N];
  }

  memset(flag, 0, N*sizeof(short));

// Cycle through mom & dad to get the cyclic part of the crossover.
  int current=0;
  child[0] = mom[0];
  flag[0] = 1;
  while(dad[current] != mom[0]){
    for(int i=0; i<child.size(); i++){
      if(mom[i] == dad[current]){
	child[i] = mom[i];
	flag[i] = 1;
	current = i;
	break;
      }
    }
  }

// Now fill in the rest of the child with dad's contents that we didn't use
// in the cycle.
  for(int i=0; i<child.size(); i++)
    if(flag[i] == 0) child[i] = dad[i];
}
#endif
