// $Header: /usr/people/mbwall/src/ga/ga/RCS/array2.ch.C,v 1.2 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  array2.ch.C
  mbwall 25feb95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Source file for the 2D array genome.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <string.h>
#include <ga/array2.ch.h>


/* ----------------------------------------------------------------------------
2DArrayGenome
---------------------------------------------------------------------------- */
template <class T> 
GA2DArrayGenome<T>::
GA2DArrayGenome(const unsigned int width, const unsigned int height, 
		    GAAlleleSet<T> & s,
		    GAObjectiveFunction f,
		    void * u,
		    GA * g) :
GAGenome(GANoInitializer, GANoMutator, GANoCrossover, (GACrossSite *)0),
GAArray<T>(width*height){
  of=f; ud=u; ga=g; as=&s;
  nx=minX=maxX=width; ny=minY=maxY=height;
}


template <class T> 
GA2DArrayGenome<T>::
GA2DArrayGenome(const GA2DArrayGenome<T> & orig) : GAArray<T>(orig.sz){
  GA2DArrayGenome<T>::copy(orig);
}


template <class T> void
GA2DArrayGenome<T>::copy(const GAGenome & orig){
  if(&orig == this) return;
  if(!sameClass(orig)){
    GAErr(className(), "copy", gaErrObjectTypeMismatch);
    return;
  }
  GA2DArrayGenome<T> & c = (GA2DArrayGenome<T> &)orig;
  GAGenome::copy(c);
  GAArray<T>::copy(c);
  as = c.as;
  nx = c.nx; ny = c.ny; 
  minX = c.minX; minY = c.minY; 
  maxX = c.maxX; maxY = c.maxY;
} 


template <class T> GAGenome *
GA2DArrayGenome<T>::clone(const GACloneMethod flag) const {
  GA2DArrayGenome<T> *cpy = new GA2DArrayGenome<T>(nx,ny,*as);
  if(flag == gaCloneContents){
    cpy->copy(*this);
  }
  else{
    cpy->GAGenome::copy(*this);
    cpy->as = as;
    cpy->minX = minX; cpy->minY = minY; 
    cpy->maxX = maxX; cpy->maxY = maxY; 
  }
  return cpy;
}


template <class T> int
GA2DArrayGenome<T>::resize(const int w, const int h){
  int width=w, height=h;
  if(width == nx && height == ny) return sz;

  if(minX == maxX){		// resizing is not allowed for this one
    if(width < 0) return nx;	// do nothing
    minX = maxX = width;	// resize to specified width
  }
  else if(width >=0){
    if(width < minX) minX = width;
    if(width > maxX) maxX = width;
  }
  else width = GARandomInt(minX, maxX);

  if(minY == maxY){		// resizing is not allowed for this one
    if(height < 0) return ny;	// do nothing
    minY = maxY = height;	// resize to specified height
  }
  else if(height >=0){
    if(height < minY) minY = height;
    if(height > maxY) maxY = height;
  }
  else height = GARandomInt(minY, maxY);

  if(width < nx){
    int y=Min(ny,height);
    for(int j=0; j<y; j++)
      move(j*width,j*nx,width);
  }
  GAArray<T>::size(width*height);
  if(width > nx){		// adjust the existing chunks of bits
    int y=Min(ny,height);
    for(int j=y-1; j>=0; j--){
      GAArray<T>::move(j*width,j*nx,nx);
      for(int i=nx; i<width; i++)
	a[j*width+i] = as->allele();
    }
  }
  if(height > ny){		// change in height is always new bits
    for(int i=width*ny; i<width*height; i++)
      a[i] = as->allele();
  }

  nx = width; ny = height;
  _evaluated = gaFalse; _nmut = 0;
  return sz;
}


template <class T> void
GA2DArrayGenome<T>::_read(istream & is)
{
  int i=0, j=0;
  do{
    is >> gene(i, j);
    if(++i >= nx){		// ready for next row
      i=0;
      j++;
    }
  } while(is && is.peek() != EOF && j < ny);

  if(is.peek() == EOF && 
     ((j < ny) ||	     // didn't get some lines
      (i < nx && i != 0)))   // stopped early on a row
    GAErr(className(), "_read", gaErrUnexpectedEOF);
}


template <class T> void
GA2DArrayGenome<T>::_write(ostream & os) const 
{
  for(int j=0; j<ny; j++){
    for(int i=0; i<nx; i++){
      os << gene(i,j);
    }
    os << "\n";
  }
}


template <class T> int
GA2DArrayGenome<T>::
resizeBehaviour(const GADimension which,
		const unsigned int lower, const unsigned int upper)
{
  if(upper < lower){
    GAErr(className(), "resizeBehaviour", gaErrBadResizeBehaviour);
    return resizeBehaviour(which);
  }

  switch(which){
  case gaWidth:
    minX = lower; maxX = upper;
    if(nx > upper) GA2DArrayGenome<T>::resize(upper,ny);
    if(nx < lower) GA2DArrayGenome<T>::resize(lower,ny);
    break;

  case gaHeight:
    minY = lower; maxY = upper;
    if(ny > upper) GA2DArrayGenome<T>::resize(nx,upper);
    if(ny < lower) GA2DArrayGenome<T>::resize(nx,lower);
    break;

  default:
    break;
  }

  return resizeBehaviour(which);
}


template <class T> void
GA2DArrayGenome<T>::copy(const GA2DArrayGenome<T> & orig,
			     const unsigned int r, const unsigned int s,
			     const unsigned int x, const unsigned int y,
			     const unsigned int width,
			     const unsigned int height)
{
  int w=width, h=height;
  if(w == 0 || x >= orig.nx || r >= nx ||
     h == 0 || y >= orig.ny || s >= ny) return;
  if(x + w > orig.nx) w = orig.nx - x;
  if(y + h > orig.ny) h = orig.ny - y;
  if(r + w > nx) w = nx - r;
  if(s + h > ny) h = ny - s;

  for(int j=0; j<h; j++)
    GAArray<T>::copy(orig, (s+j)*nx+r, (y+j)*orig.nx+x, w);
}



template <class T> int 
GA2DArrayGenome<T>::_equal(const GAGenome & c) const
{
  if(this == &c) return 1;
  GA2DArrayGenome<T> & b = (GA2DArrayGenome<T> &)c;
  if(nx != b.nx || ny != b.ny) return 0;
  int val=0;
  for(int j=0; j<ny && val==0; j++)
    val = GAArray<T>::equal(b,j*nx,j*nx,nx) ? 0 : 1;
  return(val ? 0 : 1);
}


template <class T> int 
GA2DArrayGenome<T>::_notequal(const GAGenome & c) const
{
  if(this == &c) return 0;
  GA2DArrayGenome<T> & b = (GA2DArrayGenome<T> &)c;
  if(nx != b.nx || ny != b.ny) return 1;
  int val=0;
  for(int j=0; j<ny && val==0; j++)
    val = GAArray<T>::equal(b,j*nx,j*nx,nx) ? 0 : 1;
  return(val ? 1 : 0);
}
