// $Header: /usr/people/mbwall/src/ga/ga/RCS/binstr1.ch.C,v 1.2 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  binstr1.ch.C
  mbwall 19apr95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Source file for the 1D binary string genome.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <ctype.h>
#include <ga/config.h>
#include <ga/error.h>
#include <ga/random.h>
#include <ga/binstr1.ch.h>
#include <ga/binstr1.xs.h>
#include <ga/binstr1.op.h>


/* ----------------------------------------------------------------------------
1DBinaryStringGenome
---------------------------------------------------------------------------- */
static DEFAULT_1DBINSTR_CROSSSITE ga1DBinStrDefCrossSite;

// Set all the initial values to NULL or zero, then allocate the space we'll
// need (using the resize method).  We do NOT call the initialize method at
// this point - initialization must be done explicitly by the user of the
// genome (eg when the population is created or reset).  If we called the
// initializer routine here then we could end up with multiple initializations
// and/or calls to dummy initializers (for example when the genome is 
// created with a dummy initializer and the initializer is assigned later on).
GA1DBinaryStringGenome::
GA1DBinaryStringGenome(const unsigned int length, 
		       GAObjectiveFunction f,
		       void * u,
		       GA * g) :
GABinaryStringGenome(DEFAULT_1DBINSTR_INITIALIZER,
		     DEFAULT_1DBINSTR_MUTATOR,
		     DEFAULT_1DBINSTR_CROSSOVER,
		     &ga1DBinStrDefCrossSite)
{
  of=f; ud=u; ga=g;
  nx=minX=maxX=0;
  resize(length);
}


// This is the copy initializer.  We set everything to the default values, then
// copy the original.  The BinaryStringGenome creator takes care of zeroing
// the data and sz members.
GA1DBinaryStringGenome::
GA1DBinaryStringGenome(const GA1DBinaryStringGenome & orig)
{
  nx=minX=maxX=0;
  copy(orig);
}


// This is the class-specific copy method.  It will get called by the super
// class since the superclass operator= is set up to call ccopy (and that is
// what we define here - a virtual function).  We should check to be sure that
// both genomes are the same class and same dimension.  This function tries
// to be smart about they way it copies.  If we already have data, then we do
// a memcpy of the one we're supposed to copy.  If we don't or we're not the 
// same size as the one we're supposed to copy, then we adjust ourselves.
//   The BinaryStringGenome takes care of the resize in its copy method.
// It also copies the bitstring for us.
void
GA1DBinaryStringGenome::copy(const GAGenome & orig)
{
  if(&orig == this) return;
  if(!sameClass(orig)){
    GAErr(className(), "copy", gaErrObjectTypeMismatch);
    return;
  }
  GABinaryStringGenome::copy(orig);
  GA1DBinaryStringGenome & c = (GA1DBinaryStringGenome &)orig;
  nx = c.nx; minX = c.minX; maxX = c.maxX;
}


//   Resize the genome.  This function only makes the genome bigger (if
// needed).  If we have enough space already, then it just changes nx, not Nx.
// This is a memory-hungry method of doing things, but its also faster.  If you
// are cramped for memory, then you'll have to run a memory compactor that
// forces a resize to smaller dimensions.  When we allocate more space, we also
// copy the contents of the genome into that new space.  Also set the new
// (or unused, as the case may be) memory to zero (should be to random values,
// but that takes too long).
//   A negative value for the length means that we should randomly set the
// length of the genome (if the resize behaviour is resizeable).  If
// someone tries to randomly set the length and the resize behaviour is fixed
// length, then we don't do anything.
//   We pay attention to the values of minX and maxX - they determine what kind
// of resizing we are allowed to do.  If a resize is requested with a length
// less than the min length specified by the behaviour, we set the length to
// the minimum value.  If the length is longer than the max length specified by
// the behaviour, we set the length to the max value (if the genome is
// fixed length, we allow any length for the resize).
//   We return the total size (in bits) of the genome after resize.  If we
// weren't able to resize to the requested size, the return value will tell you
// what size we were able to resize to.
//   If we resize to a larger size, then we set the new bits to random values.
// If we shrink, we do nothing special.  In both cases we expect the resize to
// copy the original bits (the super class takes care of this).
int
GA1DBinaryStringGenome::resize(const int l)
{
  int length=l;
  if(length == nx) return nx;

  if(minX == maxX){		// resizing is not allowed for this one
    if(length < 0) return nx;	// do nothing
    minX = maxX = length;	// resize to specified length
  }
  else if(length >=0){
    if(length < minX) minX = length;
    if(length > maxX) maxX = length;
  }
  else length = GARandomInt(minX, maxX);

  GABinaryStringGenome::resize(length);
  if(length > nx)
    for(int i=nx; i<length; i++)
      bit(i, GARandomBit());
  nx = length;
  return sz;
}


// We read data from a stream as a series of 1's and 0's.  We want a continuous
// stream (ie no spaces) but if there are spaces then we can handle them.  We
// ignore all whitespace.  We also ignore anything that is not a digit.
void
GA1DBinaryStringGenome::_read(istream & is)
{
  static char c;
  int i=0;
  do{
    is >> c;
    if(isdigit(c)) gene(i++, ((c == '0') ? 0 : 1));
  } while(is && is.peek() != EOF && i < nx);

  if(is.peek() == EOF && i < nx)
    GAErr(className(), "_read", gaErrUnexpectedEOF);
}


// When we write the data to a stream we do it without any spaces.  Also, there
// is no newline at the end of the stream of digits.
void
GA1DBinaryStringGenome::_write(ostream & os) const
{
  for(int i=0; i<nx; i++)
    os << gene(i);
}


//   Set the resize behaviour of the genome.  A genome can be fixed
// length, resizeable with a max and min limit, or resizeable with no limits
// (other than an implicit one that we use internally).
//   A value of 0 means no resize, a value less than zero mean unlimited 
// resize, and a positive value means resize with that value as the limit.
//   We return the upper limit of the genome's size.
int
GA1DBinaryStringGenome::
resizeBehaviour(const unsigned int lower, const unsigned int upper)
{
  if(upper < lower){
    GAErr(className(), "resizeBehaviour", gaErrBadResizeBehaviour);
    return resizeBehaviour();
  }
  minX = lower; maxX = upper;
  if(nx > upper) resize(upper);
  if(nx < lower) resize(lower);
  return resizeBehaviour();
}
