// $Header: /usr/people/mbwall/src/ga/ga/RCS/binstr1.op.C,v 1.2 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  binstr1.op.C
  mbwall 19apr95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Source file for the 1D binary string genome built-in operators.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <string.h>
#include <ga/config.h>
#include <ga/error.h>
#include <ga/random.h>
#include <ga/binstr1.ch.h>
#include <ga/binstr1.xs.h>
#include <ga/binstr1.op.h>


//   Set the bits of the genome to random values.  We use the library's
// random bit function so we don't have to worry about machine-specific stuff.
//   We also do a resize so the genome can resize itself (randomly) if it
// is a resizeable genome.
#if USE_RANDOM_INITIALIZER == 1
void 
GA1DBinStrRandomInitializer(GAGenome & c)
{
  GA1DBinaryStringGenome &child=(GA1DBinaryStringGenome &)c;
  child.resize(gaResizeAny);			// let chrom resize if it can
  for(int i=child.length()-1; i>=0; i--)
    child.gene(i, GARandomBit()); // initial values are all random
}
#endif


//   Unset all of the bits in the genome.
#if USE_UNSET_INITIALIZER == 1
void 
GA1DBinStrUnsetInitializer(GAGenome & c)
{
  GA1DBinaryStringGenome &child=(GA1DBinaryStringGenome &)c;
  child.resize(gaResizeAny);			// let chrom resize if it can
  child.unset(0, child.length());
}
#endif


//   Set all of the bits in the genome.
#if USE_SET_INITIALIZER == 1
void 
GA1DBinStrSetInitializer(GAGenome & c)
{
  GA1DBinaryStringGenome &child=(GA1DBinaryStringGenome &)c;
  child.resize(gaResizeAny);			// let chrom resize if it can
  child.set(0, child.length());
}
#endif


// This function gets called a lot (especially for the simple ga) so it must
// be as streamlined as possible.  If the mutation probability is small, then
// we must call random on each bit in the string.  Otherwise, can can simply
// mutate a known number of bits (based on the mutation rate).  We don't check
// to see how many bits we flip, nor do we keep track of which ones got flipped
// so this can result in an actual mutation that is lower than that specified,
// but the bigger the genome and the smaller the mutation probability, the
// better the chance that it will match the desired mutation rate.
//   If nMut is greater than 1, then we round up, so a mutation of 2.2 would
// be 3 mutations, and 2.9 would be 3 as well.  nMut of 3 would be 3 mutations.
#if USE_FLIP_MUTATOR == 1
int 
GA1DBinStrFlipMutator(GAGenome & c, float pmut)
{
  GA1DBinaryStringGenome &child=(GA1DBinaryStringGenome &)c;
  register int n, i;
  if(pmut <= 0.0) return(0);

  float nMut = pmut * (float)(child.length());
  if(nMut < 1.0){		// we have to do a flip test on each bit
    nMut = 0;
    for(i=child.length()-1; i>=0; i--){
      if(GAFlipCoin(pmut)){
	child.gene(i, ((child.gene(i) == 0) ? 1 : 0));
	nMut++;
      }
    }
  }
  else{				// only flip the number of bits we need to flip
    for(n=0; n<nMut; n++){
      i = GARandomInt(0, child.length()-1); // the index of the bit to flip
      child.gene(i, ((child.gene(i) == 0) ? 1 : 0));
    }
  }
  return((int)nMut);
}
#endif



// Randomly take bits from each parent.  For each bit we flip a coin to see if
// that bit should come from the mother or the father.  This will not work with
// resizable genomes!  If the genome is not the same size as the parents then
// we post an error message.  The crossover method does not need a crossover
// site.
#if USE_UNIFORM_CROSSOVER == 1
void
GA1DBinStrUniformCrossover(GAGenome &c, const GAGenome &c1, const GAGenome &c2)
{
  GA1DBinaryStringGenome &child=(GA1DBinaryStringGenome &)c;
  GA1DBinaryStringGenome &mom=(GA1DBinaryStringGenome &)c1;
  GA1DBinaryStringGenome &dad=(GA1DBinaryStringGenome &)c2;

  if(child.resizeBehaviour() == gaNoResize){
    if(mom.length() != dad.length()){
      GAErr(c.className(), "GA1DBinStrUniformCrossover", 
	    gaErrParentSizeMismatch);
      return;
    }
    if(child.length() != mom.length()){
      GAErr(c.className(), "GA1DBinStrUniformCrossover",
	    gaErrChildSizeMismatch);
      return;
    }
  }
  else{
    GAErr(c.className(), "GA1DBinStrUniformCrossover", gaErrFixedOnly);
    return;
  }
  
  for(int i=child.length()-1; i>=0; i--)
    child.gene(i, (GARandomBit() ? mom.gene(i) : dad.gene(i)));
}
#endif



// The default crossover operators are single point, random/specified crossover
// point.  If no crossover point is specified, then a random point is selected.
// If the genome is allowed to resize itself, then the crossover may resize
// it (depending on the dimensions of the parent genomes).  If not, then
// the parents must be the same size as the child or we post an error message.
//   Be sure that you reset the crossover site before you do a crossover.  The
// crossover site is not spring loaded, so if you set it then do another cross
// the second one will be at the same site as the first (even if the parents 
// have changed sizes relative to those in the first cross).
//   We set the x member of the crossover site to the point in the child
// genome where the crossover occurred.  We set the x1 member to the 
// location in mom, the x2 member to the location in dad.  If the child is
// fixed size then x, x1, and x2 are the same.  If the child is not fixed size
// then x, x1, and x2 may or may not be the same.
//   A word about crossover site mapping.  If a genome has width 10, the
// cross site can assume a value of 0 to 10, inclusive.  A site of 0 means
// that all of the material comes from the father.  A site of 10 means that
// all of the material comes from the mother.  A site of 3 means that bits
// 0-2 come from the mother and bits 3-9 come from the father.
#if USE_1PT_CROSSOVER == 1
void
GA1DBinStr1PtCrossover(GAGenome & c, const GAGenome & c1, const GAGenome & c2)
{
  GA1DBinaryStringGenome &child=(GA1DBinaryStringGenome &)c;
  GA1DBinaryStringGenome &mom=(GA1DBinaryStringGenome &)c1;
  GA1DBinaryStringGenome &dad=(GA1DBinaryStringGenome &)c2;
  GA1DBinStr1PtCrossSite & xsite =
    (GA1DBinStr1PtCrossSite &)child.crossoverSite();

  unsigned int x2;
  if(child.resizeBehaviour() == gaNoResize){
    if(mom.length() != dad.length()){
      GAErr(c.className(), "GA1DBinStr1PtCrossover", gaErrParentSizeMismatch);
      return;
    }
    if(child.length() != mom.length()){
      GAErr(c.className(), "GA1DBinStr1PtCrossover", gaErrChildSizeMismatch);
      return;
    }
    if(xsite.x == gaNoSite) xsite.pick(child);
    xsite.x1 = xsite.x2 = xsite.x;
    x2 = child.length() - xsite.x;
  }
  else{
    if(xsite.x1 == gaNoSite || xsite.x2 == gaNoSite)
      xsite.pick(child, mom, dad);
    xsite.x = xsite.x1;
    x2 = dad.length() - xsite.x2;
    child.resize(xsite.x+x2);
  }
  child.copy(mom, 0, 0, xsite.x1);
  child.copy(dad, xsite.x1, xsite.x2, x2);
}
#endif


// Two point crossover for one dimension binary strings.  The first part is
// taken from the mother, the second from the father, and the third from the
// mother.  If the child is resizable then we resize before copying the parts.
//   The rules for doing resizable crossover apply here as well as for the
// single point crossover (see comments for 1Pt for details).
#if USE_2PT_CROSSOVER == 1
void
GA1DBinStr2PtCrossover(GAGenome & c, const GAGenome & c1, const GAGenome & c2)
{
  GA1DBinaryStringGenome &child=(GA1DBinaryStringGenome &)c;
  GA1DBinaryStringGenome &mom=(GA1DBinaryStringGenome &)c1;
  GA1DBinaryStringGenome &dad=(GA1DBinaryStringGenome &)c2;
  GA1DBinStr2PtCrossSite & xsite =
    (GA1DBinStr2PtCrossSite &)child.crossoverSite();

  unsigned int dadlength, momlength;
  if(child.resizeBehaviour() == gaNoResize){
    if(mom.length() != dad.length()){
      GAErr(c.className(), "GA1DBinStr2PtCrossover", gaErrParentSizeMismatch);
      return;
    }
    if(child.length() != mom.length()){
      GAErr(c.className(), "GA1DBinStr2PtCrossover", gaErrChildSizeMismatch);
      return;
    }
    if(xsite.x[0] == gaNoSite || xsite.x[1] == gaNoSite) xsite.pick(child);
    dadlength = xsite.x[1] - xsite.x[0];
    momlength = mom.length() - xsite.x[1];
    xsite.x1[0] = xsite.x2[0] = xsite.x[0];
    xsite.x1[1] = xsite.x2[1] = xsite.x[1];
  }
  else{
    if(xsite.x1[0] == gaNoSite || xsite.x2[0] == gaNoSite ||
       xsite.x1[1] == gaNoSite || xsite.x2[1] == gaNoSite)
      xsite.pick(child, mom, dad);
    dadlength = xsite.x2[1] - xsite.x2[0];
    momlength = mom.length() - xsite.x1[1];
    xsite.x[0] = xsite.x1[0];
    xsite.x[1] = xsite.x[0] + dadlength;
    child.resize(xsite.x[0]+dadlength+momlength);
  }
  child.copy(mom, 0, 0, xsite.x[0]);
  child.copy(dad, xsite.x[0], xsite.x2[0], dadlength);
  child.copy(mom, xsite.x[1], xsite.x1[1], momlength);
}
#endif


// Even and odd crossovers take alternating bits from the mother and father.
// For even crossover, we take every even bit from the mother and every odd bit
// from the father (the first bit is the 0th bit, so it is even).  Odd
// crossover is just the opposite.  
//   We don't do anything special for resizable binary strings.  We do check to
// be sure that both parents are at least as long as the child (so that we 
// don't go looking for bits in the mother/father that do not exist).
//   Notice that this crossover method does not make use of a crossover site.
// *** this could be sped up a lot in a real bitstream implementation
#if USE_EVEN_CROSSOVER == 1
void
GA1DBinStrEvenCrossover(GAGenome & c, const GAGenome & c1, const GAGenome & c2)
{
  GA1DBinaryStringGenome &child=(GA1DBinaryStringGenome &)c;
  GA1DBinaryStringGenome &mom=(GA1DBinaryStringGenome &)c1;
  GA1DBinaryStringGenome &dad=(GA1DBinaryStringGenome &)c2;

  if(child.length() > mom.length() || child.length() > dad.length()){
    GAErr(c.className(), "GA1DBinStrEvenCrossover", gaErrChildSizeMismatch);
    return;
  }
  for(int i=0; i<child.length(); i++)
    child.gene(i, ((i%2 == 0) ? mom.gene(i) : dad.gene(i)));
}
#endif

#if USE_ODD_CROSSOVER == 1
void
GA1DBinStrOddCrossover(GAGenome & c, const GAGenome & c1, const GAGenome & c2)
{
  GA1DBinaryStringGenome &child=(GA1DBinaryStringGenome &)c;
  GA1DBinaryStringGenome &mom=(GA1DBinaryStringGenome &)c1;
  GA1DBinaryStringGenome &dad=(GA1DBinaryStringGenome &)c2;

  if(child.length() > mom.length() || child.length() > dad.length()){
    GAErr(c.className(), "GA1DBinStrOddCrossover", gaErrChildSizeMismatch);
    return;
  }
  for(int i=0; i<child.length(); i++)
    child.gene(i, ((i%2 == 0) ? dad.gene(i) : mom.gene(i)));
}
#endif
