// $Header: /usr/people/mbwall/src/ga/ga/RCS/ga.C,v 1.25 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  ga.C
  mbwall 28jul94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
   Souce file for the GA objects.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <ga/ga.h>
#include <ga/random.h>
#include <ga/version.h>		// gets the RCS string in for ident purposes


/* ----------------------------------------------------------------------------
Completion functions
---------------------------------------------------------------------------- */
GABoolean GATerminateUponGeneration(GA & ga){
  return(ga.generation() < ga.nGenerations() ? gaFalse : gaTrue);
}
GABoolean GATerminateUponConvergence(GA & ga){
  return(ga.convergence() < ga.pConvergence() ? gaTrue : gaFalse);
}
GABoolean GATerminateUponPopConvergence(GA & ga){
  if(1-ga.population().ave()/ga.population().best().score()<ga.pConvergence())
    return gaTrue;
  else
    return gaFalse;
}




/* ----------------------------------------------------------------------------
GA
-------------------------------------------------------------------------------
  This class is not meant to be instantiated.  The primary purpose of this 
class is to set everything to default values and to give you the basic methods
for a broad set of GAs.
  For each of the args, if we get a valid value, we use it.  If we get a NULL
or bogus value, we use the default for that GA component.
  We do not own the memory used by the user data, so we don't have to clean
that up when we're finished with it.
  We own all of the genomes we create, so we set the GA member of all of 
those to the current GA.  Notice that if you derive a GA from the base class
you should do the same to any other genomes (in temporary populations, for
example) that you own.
  Notice that the GA's user data can be different than that of the 
genomes.
---------------------------------------------------------------------------- */
GA::GA(const GAGenome & c, void * u) : stats(c)
{
  pop = new GAPopulation(c, gaDefPopSize);
  pop->geneticAlgorithm(*this);
  for(int i=0; i<pop->size(); i++) pop->individual(i).geneticAlgorithm(*this);
  sf = DEFAULT_SELECTOR;
  cf = DEFAULT_TERMINATOR;
  ud = u;
  mom = (GAGenome *)0;
  dad = (GAGenome *)0;
}

GA::GA(const GAPopulation & p, void * u) : stats(p.individual(0))
{
  pop = p.clone();
  pop->geneticAlgorithm(*this);
  for(int i=0; i<pop->size(); i++) pop->individual(i).geneticAlgorithm(*this);
  sf = DEFAULT_SELECTOR;
  cf = DEFAULT_TERMINATOR;
  ud = u;
  mom = (GAGenome *)0;
  dad = (GAGenome *)0;
}

GA::~GA(){
  delete pop;
}

void
GA::copy(const GA & arg)
{
  pop->copy(*arg.pop);
  pop->geneticAlgorithm(*this);
  for(int i=0; i<pop->size(); i++) pop->individual(i).geneticAlgorithm(*this);
  sf = arg.sf;
  cf = arg.cf;
  ud = arg.ud;
  mom = (GAGenome *)0;
  dad = (GAGenome *)0;
  stats = arg.stats;
}

const GAParameters &
GA::parameters(const GAParameters & p)
{
  populationSize(p.popSize);
  pMutation(p.pMut);
  pCrossover(p.pCross);
  nGenerations(p.nGen);
  pConvergence(p.pConv);
  nConvergence(p.nConv);
  nBestGenomes(p.nBestChrom);
  scoreFrequency(p.scoreFreq);
  return stats.parameters();
}

int
GA::populationSize(const unsigned int ps)
{
  int popsize=ps;
  if(ps < 1) popsize=1;
  pop->resize(popsize);
  stats.populationSize(popsize);
  return popsize;
}

void
GA::scores(const char * filename)
{
  ofstream outFile(filename, ios::out);
  if(!outFile){
    GAErr(className(), "scores", gaErrWriteError, filename);
    return;
  }
  stats.scores(outFile);
  outFile.close();
}







#if USE_SIMPLE_GA == 1
/* ----------------------------------------------------------------------------
 GASimpleGA
---------------------------------------------------------------------------- */
GASimpleGA::GASimpleGA(const GAGenome & c, void * u) : GA(c, u){
  el = gaTrue;

  oldPop = pop->clone();
//  oldPop->geneticAlgorithm(*this);
//  for(int i=0; i<oldPop->size(); i++)
//    oldPop->individual(i).geneticAlgorithm(*this);
}
GASimpleGA::GASimpleGA(const GAPopulation & p, void * u) : GA(p, u){
  el = gaTrue;

  oldPop = pop->clone();
//  oldPop->geneticAlgorithm(*this);
//  for(int i=0; i<oldPop->size(); i++)
//    oldPop->individual(i).geneticAlgorithm(*this);
}
GASimpleGA::GASimpleGA(const GASimpleGA & arg) : GA(*arg.pop, arg.ud){
  oldPop = (GAPopulation *)0;
  copy(arg);
}
GASimpleGA::~GASimpleGA(){
  delete oldPop;
}
void 
GASimpleGA::copy(const GASimpleGA & arg){
  GA::copy(arg);
  el = arg.el;

  if(oldPop) oldPop->copy(*arg.oldPop);
  else oldPop = arg.oldPop->clone();
  oldPop->geneticAlgorithm(*this);
  for(int i=0; i<oldPop->size(); i++)
    oldPop->individual(i).geneticAlgorithm(*this);
}


// We have a second population, so we have to update it if there is any change
// to the regular population.
const GAParameters &
GASimpleGA::parameters(const GAParameters & p)
{
  GASimpleGA::populationSize(p.popSize);
  return GA::parameters(p);
}

// This is basically the same as the member for the parent class, but we have
// to change the population size of the oldpop here.
int
GASimpleGA::populationSize(const unsigned int ps)
{
  return oldPop->resize(GA::populationSize(ps));
}

void 
GASimpleGA::objectiveFunction(GAObjectiveFunction f){
  for(int i=0; i<pop->size(); i++){
    pop->individual(i).objectiveFunction(f);
    oldPop->individual(i).objectiveFunction(f);
  }
}

void 
GASimpleGA::objectiveVector(const GAObjectiveVector & v){
  for(int i=0; i<pop->size(); i++){
    pop->individual(i).objectiveVector(v);
    oldPop->individual(i).objectiveVector(v);
  }
}



// Initialize the population, set the random seed as needed, do a few stupidity
// checks, reset the stats.  We must initialize the old pop because there is no
// guarantee that each individual will get initialized during the course of our
// operator++ operations.  We do not evaluate the old pop because that will 
// happen as-needed later on.
void
GASimpleGA::initialize(const unsigned int seed)
{
  if(seed) GARandomSeed(seed);
  else     GARandomSeed();

  pop->initialize();
  pop->evaluate(gaTrue);	// the old pop will get it when the pops switch
  oldPop->initialize();

  stats.reset(*pop);
}


//   Evolve a new generation of genomes.  When we start this routine, pop
// contains the current generation.  When we finish, pop contains the new 
// generation and oldPop contains the (no longer) current generation.  The 
// previous old generation is lost.  We don't deallocate any memory, we just
// reset the contents of the genomes.
//   The selection routine must return a pointer to a genome from the old
// population.
GASimpleGA &
GASimpleGA::operator++()
{
  int i;

  GAPopulation *tmppop;		// Swap the old population with the new pop.
  tmppop = oldPop;		// When we finish the ++ we want the newly 
  oldPop = pop;			// generated population to be current (for
  pop = tmppop;			// references to it from member functions).

// Generate the individuals in the temporary population from individuals in 
// the main population.

  for(i=0; i<pop->size()-1; i+=2){	// takes care of odd population
    mom = &(sf(*oldPop));  
    dad = &(sf(*oldPop));
    stats.ns += 2;		// keep track of number of selections
    if(GAFlipCoin(pCrossover())){
      pop->individual( i ).pickCrossSite(*mom, *dad);
      pop->individual(i+1).mirrorCrossSite(pop->individual(i));
      pop->individual( i ).cross(*mom, *dad);
      pop->individual(i+1).cross(*dad, *mom);
      stats.nc += 2;	// keep track of the number of crossovers
    }
    stats.nm += pop->individual( i ).mutate(pMutation());
    stats.nm += pop->individual(i+1).mutate(pMutation());
  }
  if(pop->size() % 2 != 0){	// do the remaining population member
    mom = &(sf(*oldPop));  
    dad = &(sf(*oldPop));
    stats.ns += 2;		// keep track of number of selections
    if(GAFlipCoin(pCrossover())){
      if(GARandomBit()){
	pop->individual( i ).pickCrossSite(*mom, *dad);
	pop->individual( i ).cross(*mom, *dad);
      }
      else{
	pop->individual( i ).pickCrossSite(*dad, *mom);
	pop->individual( i ).cross(*dad, *mom);
      }
      stats.nc += 1;	// keep track of the number of crossovers
    }
    stats.nm += pop->individual( i ).mutate(pMutation());
  }

  pop->touch();			// tell the pop that we changed individuals

// If we are supposed to be elitist, carry the best individual from the old
// population into the current population.
  if(el && oldPop->best().score() > pop->best().score())
    oldPop->replace(pop->replace(&(oldPop->best()), gaPopReplaceWorst), 
		    gaPopReplaceBest);

  pop->evaluate();		// get info about current pop for next time

  stats.update(*pop);		// update the statistics by one generation

  return(*this);
}

#endif





#if USE_STEADY_STATE_GA == 1
/* ----------------------------------------------------------------------------
 GASteadyStateGA
---------------------------------------------------------------------------- */
GASteadyStateGA::GASteadyStateGA(const GAGenome & c, void * u) : 
GA(c, u){
  stats.scoreFrequency(gaDefScoreFrequency2);
  pRepl = gaDefPReplacement;
  rs = gaReplaceWorst;
  el = gaTrue;

  float n = pRepl * pop->size();
  if(n < 1) n = 1;
  tmpPop = new GAPopulation(pop->individual(0), (unsigned int)n);
  tmpPop->geneticAlgorithm(*this);
  for(int i=0; i<tmpPop->size(); i++)
    tmpPop->individual(i).geneticAlgorithm(*this);
}
GASteadyStateGA::GASteadyStateGA(const GAPopulation & p, void * u) : 
GA(p, u){
  stats.scoreFrequency(gaDefScoreFrequency2);
  pRepl = gaDefPReplacement;
  rs = gaReplaceWorst;
  el = gaTrue;

  float n = pRepl * pop->size();
  if(n < 1) n = 1;
  tmpPop = new GAPopulation(pop->individual(0), (unsigned int)n);
  tmpPop->geneticAlgorithm(*this);
  for(int i=0; i<tmpPop->size(); i++)
    tmpPop->individual(i).geneticAlgorithm(*this);
}
GASteadyStateGA::GASteadyStateGA(const GASteadyStateGA & arg) : 
GA(*arg.pop, arg.ud){
  tmpPop = (GAPopulation *)0;
  copy(arg);
}
GASteadyStateGA::~GASteadyStateGA(){
  delete tmpPop;
}
void
GASteadyStateGA::copy(const GASteadyStateGA & arg){
  GA::copy(arg);
  pRepl = arg.pRepl;
  rs = arg.rs;
  el = arg.el;

  if(tmpPop) tmpPop->copy(*arg.tmpPop);
  else tmpPop = arg.tmpPop->clone();
  tmpPop->geneticAlgorithm(*this);
  for(int i=0; i<tmpPop->size(); i++)
    tmpPop->individual(i).geneticAlgorithm(*this);
}


// We have a second population, so we have to update it if there is any change
// to the regular population.
const GAParameters &
GASteadyStateGA::parameters(const GAParameters & p){
  GASteadyStateGA::populationSize(p.popSize);
  return GA::parameters(p);
}

// This is basically the same as the member for the parent class, but we have
// to change the population size of the tmppop here.  Make sure that tmp pop
// will always have a size of 1 or greater.
int
GASteadyStateGA::populationSize(const unsigned int ps)
{
  float n = pRepl * ps;
  if(n < 1) n = 1;
  tmpPop->resize((unsigned int)n);
  return GA::populationSize(ps);
}


// These are provided as convenience routines.  The only reason we do them is
// that the architecture of the library is not perfect - technically you should
// not be able to get to the the genomes if the GA owns them, but we don't
// want to make a genome interface at the GA level.  And since the GA owns
// individuals that are not directly accessible (ie the tmpPop) we have to 
// provide a way for users to change characteristics of ALL individuals.  For 
// now we only provide this for the objective stuff.  Everything else must 
// be set at the beginning (ie you cannot change it after the beginning).  Note
// that this does not preclude repair or other individual-based operations.
void 
GASteadyStateGA::objectiveFunction(GAObjectiveFunction f){
  int i;
  for(i=0; i<pop->size(); i++)
    pop->individual(i).objectiveFunction(f);
  for(i=0; i<tmpPop->size(); i++)
    tmpPop->individual(i).objectiveFunction(f);
}

void
GASteadyStateGA::objectiveVector(const GAObjectiveVector & v){
  int i;
  for(i=0; i<pop->size(); i++)
    pop->individual(i).objectiveVector(v);
  for(i=0; i<tmpPop->size(); i++)
    tmpPop->individual(i).objectiveVector(v);
}

float
GASteadyStateGA::pReplacement(const float p){
  if(p < 0 || p > 1){
    GAErr(className(), "pReplacement", gaErrBadProbValue);
    return pRepl;
  }
  float n = p * pop->size();
  if(n < 1) n = 1;
  tmpPop->resize((unsigned int)n);
  return pRepl=p;
}


// Set the replacement scheme for this GA object.  We allow only some of the
// possible replacement schemes, so be picky about what people try to set.
GAReplacementScheme
GASteadyStateGA::
replacementScheme(const GAReplacementScheme n, GAReplacementFunction f){
  switch(n){
  case gaReplaceBest:
  case gaReplaceWorst:
  case gaReplaceRandom:
    rs = n;
    break;
  case gaReplaceParent:
    GAErr(className(), "replacementScheme",
	  "gaReplaceParent cannot be used with a SteadyState GA");
    break;
  case gaReplaceCustom:
    if(f){ rs = n; rf = f; }
    else GAErr(className(), "replacementScheme", gaErrNeedRS);
    break;
  default:
    GAErr(className(), "replacementScheme", gaErrBadRS);
    break;
  }
  return rs;
}


// For initialization we set the random seed, check for stupid errors, init the
// population, reset the statistics, and that's it.
//   If we don't get a seed then we set it ourselves.  If we do get one, then
// we use it as the random seed.
//   We have to initialize our intermediate genomes (in tmpPop) just in 
// case they don't get initialized during a crossover.  We do not need to do
// an explicit evaluation of them, however.  Note that the tmpPop creates them
// using clone attributes, not clone contents.
void
GASteadyStateGA::initialize(const unsigned int seed)
{
  if(seed) GARandomSeed(seed);
  else     GARandomSeed();

  if(tmpPop->size() < 1)
    GAErr(className(), "initialize", gaErrNoIndGenerated);

  pop->initialize();
  pop->evaluate(gaTrue);
  tmpPop->initialize();		// make sure chroms are valid

  stats.reset(*pop);
}


//   Evolve a new generation of genomes.  A steady-state GA has no 'old'
// and 'new' populations - we pick from the current population and replace its
// members with the new ones we create.  We replace the worst members of the
// preceeding population.  If a genome in the tmp population is worse than
// one in the main population, the genome in the main population will be
// replaced regardless of its better score.
//   If you want to maintain the best genome from generation to generation,
// then you should set the elitist flag.
GASteadyStateGA &
GASteadyStateGA::operator++()
{
  int i;

// Generate the individuals in the temporary population from individuals in 
// the main population.

  for(i=0; i<tmpPop->size()-1; i+=2){	// takes care of odd population
    mom = &(sf(*pop));  
    dad = &(sf(*pop));
    stats.ns += 2;		// keep track of number of selections
    if(GAFlipCoin(pCrossover())){
      tmpPop->individual( i ).pickCrossSite(*mom, *dad);
      tmpPop->individual(i+1).mirrorCrossSite(tmpPop->individual(i));
      tmpPop->individual( i ).cross(*mom, *dad);
      tmpPop->individual(i+1).cross(*dad, *mom);
      stats.nc += 2;	// keep track of the number of crossovers
    }
    stats.nm += tmpPop->individual( i ).mutate(pMutation());
    stats.nm += tmpPop->individual(i+1).mutate(pMutation());
  }
  if(tmpPop->size() % 2 != 0){	// do the remaining population member
    mom = &(sf(*pop));  
    dad = &(sf(*pop));
    stats.ns += 2;		// keep track of number of selections
    if(GAFlipCoin(pCrossover())){
      if(GARandomBit()){
	tmpPop->individual( i ).pickCrossSite(*mom, *dad);
	tmpPop->individual( i ).cross(*mom, *dad);
      }
      else{
	tmpPop->individual( i ).pickCrossSite(*dad, *mom);
	tmpPop->individual( i ).cross(*dad, *mom);
      }
      stats.nc += 1;	// keep track of the number of crossovers
    }
    stats.nm += tmpPop->individual( i ).mutate(pMutation());
  }

// Replace the worst genomes in the main population with all of the individuals
// we just created.  The individuals we replace get stuck into the tmp pop and
// used next time around.  Use the replacement scheme that was specified.
  if(rs == gaReplaceCustom)
    for(i=0; i<tmpPop->size(); i++)
      tmpPop->replace(pop->replace(&(tmpPop->individual(i)),
				   &(rf(tmpPop->individual(i), *pop))), i);
  else
    for(i=0; i<tmpPop->size(); i++)
      tmpPop->replace(pop->replace(&(tmpPop->individual(i)), rs), i);
  tmpPop->touch();		// shouldn't have to do this, but until I fix
				// the population object to work in a more
				// object-oriented manner we must...

// We touch the population to tell it that we have modified the genomes.

  pop->touch();			// tell the population that we have changed it

// If we are supposed to be elitist, then we check to be sure that we're 
// keeping the best genome in the main population.  The elitist 
// genome replaces the worst.  This only has to be done if the replacement
// scheme is replace random or a custom scheme.
//   Whenever we carry over an elite it replaces the worst.
  if((rs == gaReplaceRandom || rs == gaReplaceCustom) && el && 
     tmpPop->best().score() > pop->best().score()){
    tmpPop->replace(pop->replace(&(tmpPop->best()), gaPopReplaceWorst),
		    gaPopReplaceBest);
  }

  pop->evaluate();		// get info about current pop for next time

  stats.update(*pop);		// update the statistics by one generation

  return(*this);
}

#endif









#if USE_REPLACEMENT_GA == 1
/* ----------------------------------------------------------------------------
 GAReplacementGA
---------------------------------------------------------------------------- */
GAReplacementGA::GAReplacementGA(const GAGenome & c, void * u) : 
GA(c, u){
  rs = gaReplaceWorst;
  rf = 0;
  noffspr = gaDefNumOffspring;
  stats.scoreFrequency(gaDefScoreFrequency2);
  child1 = pop->individual(0).clone(gaCloneAttributes);
  child2 = pop->individual(0).clone(gaCloneAttributes);
  child1->geneticAlgorithm(*this);
  child2->geneticAlgorithm(*this);
}

GAReplacementGA::GAReplacementGA(const GAPopulation & p, void * u) : 
GA(p, u){
  rs = gaReplaceWorst;
  rf = 0;
  noffspr = gaDefNumOffspring;
  stats.scoreFrequency(gaDefScoreFrequency2);
  child1 = pop->individual(0).clone(gaCloneAttributes);
  child2 = pop->individual(0).clone(gaCloneAttributes);
  child1->geneticAlgorithm(*this);
  child2->geneticAlgorithm(*this);
}
GAReplacementGA::GAReplacementGA(const GAReplacementGA & arg) : 
GA(*arg.pop, arg.ud){
  child1 = (GAGenome *)0;
  child2 = (GAGenome *)0;
  copy(arg);
}
GAReplacementGA::~GAReplacementGA(){
  delete child1; 
  delete child2;
}
void 
GAReplacementGA::copy(const GAReplacementGA & arg){
  GA::copy(arg);
  rs = arg.rs;
  rf = arg.rf;
  noffspr = arg.noffspr;
  if(child1) child1->copy(*arg.child1);
  else child1 = arg.child1->clone();
  if(child2) child2->copy(*arg.child2);
  else child2 = arg.child2->clone();
  child1->geneticAlgorithm(*this);
  child2->geneticAlgorithm(*this);
}


void 
GAReplacementGA::objectiveFunction(GAObjectiveFunction f){
  for(int i=0; i<pop->size(); i++)
    pop->individual(i).objectiveFunction(f);
  child1->objectiveFunction(f);
  child2->objectiveFunction(f);
}


void 
GAReplacementGA::objectiveVector(const GAObjectiveVector & v){
  for(int i=0; i<pop->size(); i++)
    pop->individual(i).objectiveVector(v);
  child1->objectiveVector(v);
  child2->objectiveVector(v);
}



GAReplacementScheme
GAReplacementGA::
replacementScheme(const GAReplacementScheme n, GAReplacementFunction f){
  switch(n){
  case gaReplaceBest:
  case gaReplaceWorst:
  case gaReplaceRandom:
  case gaReplaceParent:
    rs = n;
    break;
  case gaReplaceCustom:
    if(f){ rs = n; rf = f; }
    else GAErr(className(), "replacementScheme", gaErrNeedRS);
    break;
  default:
    GAErr(className(), "replacementScheme", gaErrBadRS);
    break;
  }
  return rs;
}


int 
GAReplacementGA::nOffspring(const unsigned int n){
  if(n < 1 || 2 < n){
    GAErr(className(), "numCrossStrategy", gaErrBadCS);
    noffspr = 1;
  }
  else{
    noffspr = n;
  }
  return noffspr;
}


// Do some basic stupidity checks then initialize the population.  We must 
// also initialize our temporary genomes, but we don't have to evaluate
// them.  Finally, reset the statistics.
void
GAReplacementGA::initialize(const unsigned int seed)
{
  if(seed)
    GARandomSeed(seed);
  else
    GARandomSeed();

  if(rs == gaReplaceCustom && rf == 0)
    GAErr(className(), "initialize", gaErrWarnNoReplacement);

  pop->initialize();
  pop->evaluate(gaTrue);

  child1->initialize();
  child2->initialize();

  stats.reset(*pop);
}


//   Evolve a new generation of genomes.  A steady-state GA has no 'old'
// and 'new' populations - we pick from the current population and replace its
// members with the new ones we create.  We generate either one or two children
// each 'generation'.  The replacement strategy is set by the GA.
GAReplacementGA &
GAReplacementGA::operator++()
{
  mom = &(sf(*pop));  
  dad = &(sf(*pop));
  stats.ns += 2;		// keep track of the number of selections

  if(GAFlipCoin(pCrossover())){
    child1->pickCrossSite(*mom, *dad);
    child1->cross(*mom, *dad);
    stats.nc += 1;		// keep track of the number of crossovers
  }
  else{
    child1->copy(*mom);
  }
  stats.nm += child1->mutate(pMutation());

  if(rs == gaReplaceParent)
    child1 = pop->replace(child1, mom);
  else if(rs == gaReplaceCustom)
    child1 = pop->replace(child1, &(rf(*child1, *pop)));
  else
    child1 = pop->replace(child1, rs);

  if(noffspr > 1){		// do a second child
    if(GAFlipCoin(pCrossover())){
      child2->mirrorCrossSite(*child1);
      child2->cross(*dad, *mom);
      stats.nc += 1;
    }
    else{
      child2->copy(*dad);
    }
    stats.nm += child2->mutate(pMutation());

    if(rs == gaReplaceParent)
      if(mom == dad)
	child2 = pop->replace(child2, gaReplaceWorst);
      else
	child2 = pop->replace(child2, dad);
    else if(rs == gaReplaceCustom)
      child2 = pop->replace(child2, &(rf(*child2, *pop)));
    else
      child2 = pop->replace(child2, rs);
  }

  pop->evaluate(gaTrue);        // allow pop-based evaluators to do their thing
  stats.update(*pop);		// update the statistics for this generation

  return(*this);
}

#endif
