// $Header: /usr/people/mbwall/src/ga/ga/RCS/ga.h,v 1.24 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  ga.h
  mbwall 28jul94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Main header for the GAlibrary.  This header declares and defines the GA
class as well as its member objects such as GAStatistics and GAParameters.

-------------------------------------------------------------------------------
Overview of who does what in the GAlibrary
-------------------------------------------------------------------------------
Here are the library capabilities and which parts of the library are involved:

Selection methods and fitness scaling:
    The GASelector functions and the GAFitness classes take care of the 
  fitness scaling and selection strategy.  These two items are often closely 
  coupled.  Note the difference between fitness function and objective 
  function.  In this library, the objective function is user-defined and tells
  (on an arbitrary scale) how well a genome performs relative to other
  genomes.  The fitness function (defined in the GAFitness objects), on the
  other hand, takes objective function scores and processes them to produce a
  number for each genome that represents its fitness for mating/selection.
    The selection strategies include roulette wheel, tournament, linear
  ranking, and stochastic sampling.  See the selector headers for more details
  about each strategy and its implementation.
    Speciation is included as a type of selection.  To use the speciating 
  selector you must define a distance function.  See the selector header for 
  more details.
    Scaling methods are loosely coupled with the selection strategies.  They
  include linear scaling, sigma truncation, and ranking.  See the fitness
  headers for more details about how you can mix and match scaling methods with
  selection strategies.

Convergence and completion:
    The GA defines a 'done' member that calls a user-defineable completion
  function to determine whether or not the GA is finished.  Built-in completion
  routines include percent-converged and number-of-generations.

Replacement strategies:
    The replacement strategy defines how a child is inserted into a population.
    Replacement strategies include child-replaces-parent, child-replaces-worst,
  and child-replaces-random.  These make sense only for GAs with overlapping
  populations.

Elitism:
    Elitism is optional.  If you want the best of each generation to be carried
  over to the next generation, turn on the elitism.  If not, turn it off.  This
  only works with non-overlapping populations.

Crossover Strategy:
    The crossover strategy defines how many children should be produced from a
  crossover.  If you only want one child from each crossover, use 1.  If you 
  want both children from the crossover, use 2.

Genome types:
    All of the genomes are derived from a base genome class.  They each
  define a set of methods that let any GA use any genome interchangeably.
  Genomes include binary and non-binary strings in 1, 2, and 3 dimensions,
  binary-to-decimal strings, and generic trees and lists.

Initialization:
    Each genome has its own initialization operator.  When a GA is
  initialized, it initializes its populations, which in turn tells each 
  chromomosome to initialize itself.  You can use the initialization method to
  bias an initial population.
    Genome initializers include random initialization for all non-template
  genomes and user-defined initialization for all (template and non-
  template) genomes.  See the genome headers for more details.

Mutation:
    Each genome has its own mutation operator.  

Crossover:
    Each genome has its own crossover operator.

-------------------------------------------------------------------------------
Basic usage

  You create a GA by first instantiating the non-default parts that you want to
use.  For example, if you want a binary string genome with roulette wheel
selection, you would define an objective function, instantiate a genome,
then instantiate the GA.  You need only instantiate the parts that you want to 
customize.
  In every case you MUST define an objective function and you MUST instantiate
one copy of the genome type you want to use.

-------------------------------------------------------------------------------
 MODIFICATIONS:
---------------------------------------------------------------------------- */
#ifndef _ga_ga_h_
#define _ga_ga_h_

#include <stdio.h>
#include <string.h>
#include <ga/config.h>
#include <ga/id.h>
#include <ga/parameters.h>
#include <ga/statistics.h>
#include <ga/genome.h>
#include <ga/population.h>
#include <ga/selector.h>

// Here we include the headers for all of the various genome types.  
#include <ga/binstr1.h>
#include <ga/binstr2.h>
#include <ga/binstr3.h>
#include <ga/bin2dec.ch.h>
#ifndef NO_TEMPLATES
#include <ga/array1.h>
#include <ga/array2.h>
#include <ga/array3.h>
#include <ga/string.ch.h>
#include <ga/string.op.h>
#include <ga/list.h>
#include <ga/tree.h>
#endif


// The replacement strategy defines how the new children will be stuck into the
// population.  If you want the new child to replace one of its parents, use
// the Parent strategy.  If you want the child to replace a random population
// member, use the Random strategy.  If you want the child to replace the worst
// population member, use the Worst strategy.  These are meaningful only for
// overlapping populations.  To do DeJong-style speciation (crowding), use the
// Crowding strategy.  You must also specify a crowding function as the 
// replacement function if you choose this strategy.  If you use Custom as the
// replacement strategy you must also specify a replacement function.
//   Note that not every replacement scheme can be used with every type of
// genetic algorithm.  If a GA supports replacement schemes, it will specify
// which schemes are valid and which are not.
enum GAReplacementScheme {
  gaReplaceRandom=gaPopReplaceRandom,
  gaReplaceBest=gaPopReplaceBest,
  gaReplaceWorst=gaPopReplaceWorst,
  gaReplaceCustom= -30,
  gaReplaceCrowding= -30,
  gaReplaceParent= -10
};


//   The replacement function is required for crowding and custom replacement
// strategies.  This function is used to pick which genome will be 
// replaced.  The first argument passed to the replacement function is the
// individual that is supposed to go into the population.  The second argument
// is the population into which the individual is supposed to go.
// The replacement function should return a reference to the genome that the
// individual should replace.  If no replacement should take place, the
// replacement function should return a reference to the individual.
typedef GAGenome & (*GAReplacementFunction)(GAGenome&, GAPopulation&);


//   The completion function tells us when the GA should stop evolving.
typedef GABoolean (*GACompletionFunction)(GA & ga);


// CompletionFunction routines.  These functions can be plugged in to a GA as a
// measure of whether or not the GA is 'done'.  They are defined in the source.
// We can't inline them because they are used as function pointers.
GABoolean GATerminateUponGeneration(GA &);
GABoolean GATerminateUponConvergence(GA &);
GABoolean GATerminateUponPopConvergence(GA &);



/* ----------------------------------------------------------------------------
GA class

  This class should never be instantiated by itself.  It defines many of the
data structures and methods common to all GAs, but by itself cannot do 
anything.  We can't make it an abstract class because it must be instantiated.

initialize
  Undefined for the base class.
done
  Calls the completion measure routine to tell whether or not the GA is done.
evolve
  This method is provided as a convenience so that you don't have to increment
the GA generation-by-generation by hand.  If you do decide to do it by hand,
be sure that you initialize before you start evolving!
---------------------------------------------------------------------------- */
class GA : public GAID {
public:
  GA(const GAGenome & c, void * u=(void *)0);
  GA(const GAPopulation & p, void * u=(void *)0);
  virtual ~GA();

  void * userData() const {return ud;}
  void * userData(void * d)
    {return ud=d;}
  GASelectionFunction selectionFunction() const {return sf;}
  GASelectionFunction selectionFunction(const GASelectionFunction f)
    {return sf=f;}
  GACompletionFunction completionFunction() const {return cf;}
  GACompletionFunction completionFunction(GACompletionFunction f)
    {return cf=f;}
  const GAScalingScheme & scalingScheme(){return pop->scaling();}
  const GAScalingScheme & scalingScheme(const GAScalingScheme & fn)
    {return pop->scaling(fn);}

  GAPopulation & population() const {return *pop;}
  GAGenome & bestOfPop(const unsigned int n=0) const {return pop->best(n);}

  const GAStatistics & statistics() const {return stats;}
  const GAParameters & parameters() const {return stats.parameters();}
  const GAParameters & parameters(const GAParameters & p);
  int   nGenerations(const unsigned int ng){return stats.nGenerations(ng);}
  int   nGenerations() const {return stats.nGenerations();}
  float pMutation(const float pm){return stats.pMutation(pm);}
  float pMutation() const {return stats.pMutation();}
  float pCrossover(const float pc){return stats.pCrossover(pc);}
  float pCrossover() const {return stats.pCrossover();}
  int   populationSize(const unsigned int ps);
  int   populationSize() const {return pop->size();}
  float pConvergence(const float pc){return stats.pConvergence(pc);}
  float pConvergence() const {return stats.pConvergence();}
  int   nConvergence(const unsigned int nc){return stats.nConvergence(nc);}
  int   nConvergence() const {return stats.nConvergence();}

  float convergence() const {return stats.convergence();}
  int   generation() const {return stats.generation();}

  GAGenome & bestOfAll(const unsigned int i=0) const
    {return stats.bestPopulation().best(i);}
  int   nBestGenomes(const unsigned int n=gaDefNumBestChrom)
    {return stats.nBestGenomes(n);}
  int   scoreFrequency(const unsigned int freq=1)
    {return stats.scoreFrequency(freq);}
  void  scoreDestination(const char * fname){stats.scoreDestination(fname);}
  void  scores(ostream & os){stats.scores(os);}
  void  scores(const char * filename);

  virtual void initialize(const unsigned int seed=0) =0;
  virtual void evolve(){initialize(); while(!done()){step();}}
  virtual void step() =0;
  GABoolean done(){return cf(*this);}

protected:
  GAStatistics stats;
  GAPopulation *pop;		// current pop, best of all populations
  GAGenome *mom, *dad;		// tmp holders for selected genomes
  GASelectionFunction sf;	// this determines our selection method
  GACompletionFunction cf;	// function for determining done-ness
  void * ud;			// pointer to user data structure

  virtual void _write(ostream & os) const {os << className() << "\n" << stats;}
  virtual void _read(istream &){}	// nothing for now
  void copy(const GA &);

private:
// don't even try to use one of these methods!
//  GA(){}
//  GA(const GA &){}
  GA & operator=(const GA &){return(*this);}
};




/* ----------------------------------------------------------------------------
 SimpleGA class
---------------------------------------------------------------------------- */
#if USE_SIMPLE_GA == 1
class GASimpleGA : public GA {
  GADefineIdentity("GASimpleGA", 2):

public:
  GASimpleGA(const GAGenome & individual, void * u=(void *)0);
  GASimpleGA(const GAPopulation & population, void * u=(void *)0);
  GASimpleGA(const GASimpleGA & arg);
  GASimpleGA & operator=(const GASimpleGA & arg){copy(arg); return(*this);}
  virtual ~GASimpleGA();

  friend ostream & operator<<(ostream & os, GASimpleGA & arg)
    {arg._write(os); return(os);}
  friend istream & operator>>(istream & is, GASimpleGA & arg)
    {arg._read(is); return(is);}

// overrides of parent class
  void initialize(const unsigned int seed=0);
  void step(){++(*this);}
  const GAParameters & parameters(const GAParameters & p);
  int populationSize(const unsigned int ps);
  const GAScalingScheme & scalingScheme(const GAScalingScheme & s)
    {oldPop->scaling(s); return pop->scaling(s);}

// Convenience functions for setting the evaluation/objective function
  void objectiveFunction(GAObjectiveFunction f);
  void objectiveVector(const GAObjectiveVector & v);

// new for this class
  GASimpleGA & operator++();
  GABoolean elitist(const GABoolean n){return el=n;}
  GABoolean elitist() const {return el;}

protected:
  GAPopulation *oldPop;		// old population 
  GABoolean el;			// are we elitist?
  void copy(const GASimpleGA & arg);
};
#endif









/* ----------------------------------------------------------------------------
 SteadyStateGA class
---------------------------------------------------------------------------- */
// The default replacement percentage is 10%
const float gaDefPReplacement     = 0.10;


#if USE_STEADY_STATE_GA == 1
class GASteadyStateGA : public GA {
  GADefineIdentity("GASteadyStateGA", 3):

public:
  GASteadyStateGA(const GAGenome & c, void * u=(void *)0);
  GASteadyStateGA(const GAPopulation & p, void * u=(void *)0);
  GASteadyStateGA(const GASteadyStateGA & arg);
  GASteadyStateGA & operator=(const GASteadyStateGA & arg)
    {copy(arg); return(*this);}
  virtual ~GASteadyStateGA();

  friend ostream & operator<<(ostream & os, GASteadyStateGA & arg)
    {arg._write(os); return(os);}
  friend istream & operator>>(istream & is, GASteadyStateGA & arg)
    {arg._read(is); return(is);}

// overrides of parent class
  void initialize(const unsigned int seed=0);
  void step(){++(*this);}
  const GAParameters & parameters(const GAParameters & p);
  int populationSize(const unsigned int ps);

// convenience routines for accessing all individuals
  void objectiveFunction(GAObjectiveFunction f);
  void objectiveVector(const GAObjectiveVector & v);

// new for this class
  GASteadyStateGA & operator++();
  GABoolean elitist(const GABoolean n){return el=n;}
  GABoolean elitist() const {return el;}
  float pReplacement() const {return pRepl;}
  float pReplacement(const float prepl);
  GAReplacementScheme replacementScheme() const {return rs;}
  GAReplacementScheme replacementScheme(const GAReplacementScheme n,
					GAReplacementFunction f=0);

protected:
  GABoolean el;			// are we elitist?
  GAPopulation *tmpPop;		// temporary population for replacements
  float pRepl;			// percentage of population to replace each gen
  GAReplacementScheme rs;	// replacement strategy
  GAReplacementFunction rf;	// (optional) replacement function
  void copy(const GASteadyStateGA &);
};
#endif










/* ----------------------------------------------------------------------------
ReplacementGA class
---------------------------------------------------------------------------- */
// By default we generate two offspring in each generation.
const int   gaDefNumOffspring     = 2;


#if USE_REPLACEMENT_GA == 1
class GAReplacementGA : public GA {
  GADefineIdentity("GAReplacementGA", 4):

public:
  GAReplacementGA(const GAGenome & c, void * u=(void *)0);
  GAReplacementGA(const GAPopulation & p, void * u=(void *)0);
  GAReplacementGA(const GAReplacementGA & arg);
  GAReplacementGA & operator=(const GAReplacementGA & arg)
    {copy(arg); return(*this);}
  virtual ~GAReplacementGA();

  friend ostream & operator<<(ostream & os, GAReplacementGA & arg)
    {arg._write(os); return(os);}
  friend istream & operator>>(istream & is, GAReplacementGA & arg)
    {arg._read(is); return(is);}

// overrides of parent class
  void initialize(const unsigned int seed=0);
  void step(){++(*this);}

// convenience routines for accessing all individuals
  void objectiveFunction(GAObjectiveFunction f);
  void objectiveVector(const GAObjectiveVector & v);

// new for this class
  GAReplacementGA & operator++();
  int nOffspring() const {return noffspr;}
  int nOffspring(const unsigned int n);
  GAReplacementScheme replacementScheme() const {return rs;}
  GAReplacementScheme replacementScheme(const GAReplacementScheme n,
					GAReplacementFunction f=0);

protected:
  GAReplacementScheme rs;	// replacement strategy
  GAReplacementFunction rf;	// (optional) replacement function
  GAGenome *child1, *child2;
  unsigned int noffspr;		// number of children to generate in crossover
  void copy(const GAReplacementGA &);
};
#endif


#endif
