// $Header: /usr/people/mbwall/src/ga/ga/RCS/list.op.C,v 1.5 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  list.op.C
  mbwall 25feb95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Outline code for doing built-in operators for the list genome.

  See the comments in the array operators file for instructions about how to 
use this code.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <ga/config.h>
#include <ga/random.h>
#include <ga/list.ch.h>
#include <ga/list.xs.h>

#if USE_ORDER_CROSSOVER == 1
int GAListIsHole(GAListGenome<LIST_TYPE> &, 
		 const GAListGenome<LIST_TYPE> &, int, int, int);
#endif


// Mutate a list by nuking nodes.  Any node has a pmut chance of getting nuked.
// This is actually kind of bogus for the second part of the if clause (if nMut
// is greater than or equal to 1).  Nodes end up having more than pmut 
// probability of getting nuked.
//   After the mutation the iterator is left at the head of the list.
#if USE_DESTRUCTIVE_MUTATOR == 1
int
GAListDestructiveMutator(GAGenome & c, float pmut)
{
  GAListGenome<LIST_TYPE> &child=(GAListGenome<LIST_TYPE> &)c;
  register int n, i;
  if(pmut <= 0.0) return 0;

  n = child.size();
  float nMut = pmut * (float)n;
  if(nMut < 1.0){		// we have to do a flip test for each node
    nMut = 0;
    for(i=0; i<n; i++){
      if(GAFlipCoin(pmut) && child.warp(i)){
	child.destroy();
	nMut++;
      }
    }
  }
  else{				// only nuke the number of nodes we need to
    for(i=0; i<nMut; i++){
      if(child.warp(GARandomInt(0, n-1)))
	child.destroy();
    }
  }
  child.head();		// set iterator to root node

  return((int)nMut);
}
#endif


// Mutate a list by swapping two nodes.  Any node has a pmut chance of getting
// swapped, and the swap could happen to any other node.  And in the case of
// nMut < 1, the swap may generate a swap partner that is the same node, in 
// which case no swap occurs (we don't check).
//   After the mutation the iterator is left at the head of the list.
#if USE_SWAP_MUTATOR == 1
int
GAListSwapMutator(GAGenome & c, float pmut)
{
  GAListGenome<LIST_TYPE> &child=(GAListGenome<LIST_TYPE> &)c;
  register int n, i;
  if(pmut <= 0.0) return 0;

  n = child.size();
  float nMut = pmut * (float)n;
  nMut *= 0.5;			// swapping one node swaps another as well
  if(nMut < 1.0){		// we have to do a flip test for each node
    nMut = 0;
    for(i=0; i<n; i++){
      if(GAFlipCoin(pmut)){
	child.swap(i,GARandomInt(0,n-1));
	nMut++;
      }
    }
  }
  else{				// only nuke the number of nodes we need to
    for(i=0; i<nMut; i++)
      child.swap(GARandomInt(0,n-1),GARandomInt(0,n-1));
  }
  child.head();		// set iterator to root node

  return((int)nMut*2);
}
#endif


// This crossover picks a site between nodes in each parent.  It is the same 
// as single point crossover on a resizeable binary string genome.
//   When we pick a crossover site, it is between nodes of the list (otherwise
// we won't be able to do NULL lists or get an empty list from one parent or
// the other).  Beware that a list is numbered from 0 to size-1, inclusive,
// whereas the cross site possibilities are numbered from 0 to size, inclusive.
// This means we have to map the site to the list to determine whether an
// insertion should occur before or after a node.
//   We first copy the mother into the child (this deletes whatever contents 
// were in the child originally).  Then we clone the father from the cross site
// to the end of the list.  Then we delete the tail of the child from the
// mother's cross site to the end of the list.  Finally, we insert the clone
// at the end of the child.
//   The last thing we do is delete the clone (the contents of the clone are
// now owned by the child, but the clone itself uses memory that we must free).
//   This implementation isn't particularly efficient.  For example, we copy
// the mother then proceed to destroy much of the copy we just made.  We could
// do better by copying only what we need of the mother.
#if USE_1PT_CROSSOVER == 1
void
GAList1PtCrossover(GAGenome & c, const GAGenome & a, const GAGenome & b)
{
  GAListGenome<LIST_TYPE> &child=(GAListGenome<LIST_TYPE> &)c;
  GAListGenome<LIST_TYPE> &mom=(GAListGenome<LIST_TYPE> &)a;
  GAListGenome<LIST_TYPE> &dad=(GAListGenome<LIST_TYPE> &)b;
  GAList1PtCrossSite<LIST_TYPE> & xsite = 
    (GAList1PtCrossSite<LIST_TYPE> &)child.crossoverSite();

  if(xsite.a == gaNoSite || xsite.b == gaNoSite) xsite.pick(c,a,b);
  if(xsite.a > mom.size()) xsite.a = mom.size();
  if(xsite.b > dad.size()) xsite.b = dad.size();

  child.GAList<LIST_TYPE>::copy(mom);
  GAList<LIST_TYPE> * list = dad.GAList<LIST_TYPE>::clone(xsite.b);
  if(xsite.a < mom.size()){
    LIST_TYPE *site = child.warp(xsite.a);
    while(child.tail() != site)
      child.destroy();		// delete the tail node
    child.destroy();		// trash the trailing node (list[xsite.a])
  }
  else{
    child.tail();		// move to the end of the list
  }
  child.insert(list);		// stick the clone onto the end
  delete list;
  child.warp(0);		// set iterator to head of list
}
#endif


// Partial match crossover for list genomes.  We need two versions of this
// routine:  one for lists whose nodes are pointers to objects (each genome
// points to the same objects as all of the other genomes) and one for
// lists whose nodes contain independent objects (each node has its own copy
// of the object).
//   This version of the partial match crossover uses objects that are multiply
// instantiated - each list genome contains its own objects in its nodes.
// The operator== method must be defined on the object for this implementation
// to work!  In this case, the 'object' is an int, so we're OK.  If you are 
// putting your own objects in the nodes, be sure you have operator== defined
// for your object.  You must also have operator!= defined for your object.  We
// do not do any assignments, so operator= and/or copy is not required.
//   We assume that none of the nodes will return a NULL pointer.  Also assume
// that the cross site has been selected properly.
//   First we make a copy of the mother.  Then we loop through the match 
// section and try to swap each element in the child's match section with its
// partner (as defined by the current node in the father's match section).
//   Mirroring will work the same way - just swap mom & dad and you're all set.
//   The parents should be the same size as the child (and they should contain
// the same nodes, but in any order).  We do not check for this!!
#if USE_PARTIAL_MATCH_CROSSOVER == 1
void
GAListPartialMatchCrossover(GAGenome &c, const GAGenome &a, const GAGenome &b)
{
  GAListGenome<LIST_TYPE> &child=(GAListGenome<LIST_TYPE> &)c;
  GAListGenome<LIST_TYPE> &mom=(GAListGenome<LIST_TYPE> &)a;
  GAListGenome<LIST_TYPE> &dad=(GAListGenome<LIST_TYPE> &)b;
  GAListOrderedCrossSite<LIST_TYPE> & xsite =
    (GAListOrderedCrossSite<LIST_TYPE> &)child.crossoverSite();

  if(xsite.a == gaNoSite || xsite.b == gaNoSite) xsite.pick(c);

  child.GAList<LIST_TYPE>::copy(mom);
  GAListIter<LIST_TYPE> iter(dad);

  LIST_TYPE *index = iter.warp(xsite.a);
  for(int i=xsite.a; i<xsite.b; i++, index=iter.next()){
    if(*child.head() == *index){
      child.swap(i,0);
    }
    else{
      int j;
      for(j=1; (j<child.size()) && (*child.next() != *index); j++);
      child.swap(i,j);  // no op if j>size
    }
  }

  child.head();         // set iterator to head of list
}
#endif


#if USE_ORDER_CROSSOVER == 1
// Order crossover for lists.  As described in Goldberg's book.
//   We assume that we'll never get a NULL pointer while iterating through the
// list.  Also we assume that the lists are the same size and non-NULL.
int
GAListIsHole(GAListGenome<LIST_TYPE> &c, 
	     const GAListGenome<LIST_TYPE> &dad,
	     int index, int a, int b){
  GAListIter<LIST_TYPE> citer(c), diter(dad);
  citer.warp(index);
  diter.warp(a);
  for(int i=a; i<b; i++){
    if(*citer.current() == *diter.current()) return 1;
    diter.next();
  }
  return 0;
}

void
GAListOrderCrossover(GAGenome & c, const GAGenome & a, const GAGenome & b)
{
  GAListGenome<LIST_TYPE> &child=(GAListGenome<LIST_TYPE> &)c;
  GAListGenome<LIST_TYPE> &mom=(GAListGenome<LIST_TYPE> &)a;
  GAListGenome<LIST_TYPE> &dad=(GAListGenome<LIST_TYPE> &)b;
  GAListOrderedCrossSite<LIST_TYPE> & xsite =
    (GAListOrderedCrossSite<LIST_TYPE> &)child.crossoverSite();

  if(xsite.a == gaNoSite || xsite.b == gaNoSite) xsite.pick(c);

  child.GAList<LIST_TYPE>::copy(mom);
  GAListIter<LIST_TYPE> citer(child);
  GAListIter<LIST_TYPE> diter(dad);

// Move all the 'holes' into the crossover section and maintain the ordering of
// the non-hole elements.
  int i, index;
  for(i=0, index=xsite.b; i<child.size(); i++, index++){
    if(index >= child.size()) index=0;
    if(GAListIsHole(child,dad,index,xsite.a,xsite.b)) break;
  }

  for(; i<child.size()-xsite.b+xsite.a; i++, index++){
    if(index >= child.size()) index=0;
    int j=index;
    do{
      j++;
      if(j >= child.size()) j=0;
    } while(GAListIsHole(child,dad,j,xsite.a,xsite.b));
    child.swap(index,j);
  }

// Now put the 'holes' in the proper order within the crossover section.
  
  for(i=xsite.a, child.warp(xsite.a), diter.warp(xsite.a);
      i<xsite.b; i++, child.next(), diter.next()){
    if(*child.current() != *diter.current()){
      citer.warp(i);
      for(int j=i+1; j<xsite.b; j++)
	if(*citer.next() == *diter.current()){
	  child.swap(i,j);
	  child.warp(citer);	// move iterator back to previous location
	  break;
	}
    }
  }

  child.head();		// set iterator to head of list
}
#endif


#if USE_CYCLE_CROSSOVER == 1
void
GAListCycleCrossover(GAGenome & c, const GAGenome & a, const GAGenome & b)
{
  GAListGenome<LIST_TYPE> &child=(GAListGenome<LIST_TYPE> &)c;
  GAListGenome<LIST_TYPE> &mom=(GAListGenome<LIST_TYPE> &)a;
  GAListGenome<LIST_TYPE> &dad=(GAListGenome<LIST_TYPE> &)b;

  child.GAList<LIST_TYPE>::copy(mom);
  GAListIter<LIST_TYPE> diter(dad);

  static short *flag=(short *)0;
  static int N=0;

// Allocate space for an array of flags.  We use this to keep track of whether
// the child's contents came from the mother or the father.  We don't free the
// space here, but it is not a memory leak.
  if(N < child.size()){
    delete [] flag;
    N = child.size();
    flag = new short [N];
  }

  memset(flag, 0, N*sizeof(short));

// Cycle through mom & dad to get the cyclic part of the crossover.
  flag[0] = 1;
  diter.head();
  while(*diter.current() != *child.head()){
    for(int i=0; i<child.size(); i++, child.next()){
      if(*child.current() == *diter.current()){
	flag[i] = 1;
	diter.warp(i);
	break;
      }
    }
  }

// Now fill in the rest of the child with dad's contents that we didn't use
// in the cycle.
  child.head();
  diter.head();
  for(int i=0; i<child.size(); i++){
    if(flag[i] == 0) *child.current() = *diter.current();
    child.next(); diter.next();
  }

  child.head();		// set iterator to head of list
}
#endif
