// $Header: /usr/people/mbwall/src/ga/ga/RCS/objective.h,v 1.5 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  objective.h
  mbwall 27apr95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  This header defines the interface for the objective object.  This is 
basically a resizable array.  We don't do any caching, however, so resizing
frequently will be expensive in terms of performance.
  This is supposed to be a lightweight object so that it can be contained in
every genome.  Try to keep it that way.
  We define a couple of built-in converters to go from the multi-objective
objects to a single score.  You can use any of the built-in methods or define
your own.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#ifndef _ga_objective_h_
#define _ga_objective_h_

#include <iostream.h>
#include <string.h>
#include <ga/types.h>

#ifdef BOGUS_CONST
#define CONST 
#else
#define CONST const
#endif

class GAGenome;
class GAPopulation;
class GAObjectiveVector;


// This is the signature for objective functions.  Every objective function
// must have this signature!  This is the type of function used by the GA to
// evaluate each member of the population.
//  The first one is the objective function that is used on genomes.
//  The second one it the objective function that is used on populations.
typedef float (*GAObjectiveFunction)(GAGenome &);
typedef float (*GAGenomeObjectiveFunction)(GAGenome &);
typedef float (*GAPopulationObjectiveFunction)(GAPopulation &);


// This is the signature for the evaluation function for the objective vector
// objects.  We define a couple of evaluators and make it possible to override
// them with custom evaluators.  This type of function is used only by the
// objective vector object for doing its evaluation.
typedef float (*GAVectorObjectiveFunction)(const GAObjectiveVector &);


// These are the evaluators that are pre-defined for the library.  They 
// determine how the vector scores get transformed into a single, scalar score.
float GAObjectiveProduct(const GAObjectiveVector &);
float GAObjectiveSummation(const GAObjectiveVector &);


// The objective vector keeps track of the components of a multi-criteria
// objective.  It also contains a function that knows how evaluate the
// set of objectives.  It is initialized to zero.
class GAObjectiveVector {
public:
  GAObjectiveVector(const unsigned int n,
		    GAVectorObjectiveFunction f=GAObjectiveProduct) : 
  ef(f), score(0), evaluated(gaFalse), sz(n), a(n?new float[n]:(float *)0)
    {if(a) memset(a, 0, n*sizeof(float));}
  GAObjectiveVector(const GAObjectiveVector & orig) : a(0) {copy(orig);}
  GAObjectiveVector & operator=(const GAObjectiveVector & orig)
    {copy(orig); return *this;}
  virtual ~GAObjectiveVector(){delete [] a;}
  GAObjectiveVector * clone() const {return new GAObjectiveVector(*this);}
  void copy(const GAObjectiveVector & orig){
    sz = orig.sz;
    delete [] a;			// assumes NULL if no contents
    a = new float[sz];		// no check for failed allocation
    memcpy(a, orig.a, sz*sizeof(float));
    evaluated = orig.evaluated;
    score = orig.score;
    ef = orig.ef;
  }

  int size() const {return sz;}
  int resize(const unsigned int i){
    float * tmp = a;
    a = new float[i];		// no check for failed allocation or i bounds
    memcpy(a, tmp, (i<sz?i:sz)*sizeof(float));
    delete [] tmp;
    evaluated = gaFalse;
    score = 0;
    return sz=i;
  }
  GAVectorObjectiveFunction evaluator() const {return ef;}
  GAVectorObjectiveFunction evaluator(GAVectorObjectiveFunction f)
    {evaluated=gaFalse; return ef=f;}
  float component(const unsigned int i) const {return a[i];} // no checks
  float component(const unsigned int i, const float f)
    {evaluated=gaFalse; return a[i]=f;}
  const float & operator[](const unsigned int i) const {return a[i];}
  float & operator[](const unsigned int i){evaluated=gaFalse; return a[i];}
  float evaluate() CONST {
    GAObjectiveVector * This = (GAObjectiveVector *)this;
    if(!This->evaluated){This->score = ef(*this); This->evaluated=gaTrue;}
    return score;
  }


// These behave as if the objective object were a vector.  We can add, 
// subtract, multiply, divide...  In each case, we do the operation to each
// element of the array.
  GAObjectiveVector & operator=(const float value){
    for(unsigned int i=0; i<sz; i++) a[i] = value;
    evaluated = gaFalse;
    return *this;
  }
  GAObjectiveVector & operator+=(const float value){
    for(unsigned int i=0; i<sz; i++) a[i] += value;
    evaluated = gaFalse;
    return *this;
  }
  GAObjectiveVector & operator-=(const float value){
    for(unsigned int i=0; i<sz; i++) a[i] -= value;
    evaluated = gaFalse;
    return *this;
  }
  GAObjectiveVector & operator*=(const float value){
    for(unsigned int i=0; i<sz; i++) a[i] *= value;
    evaluated = gaFalse;
    return *this;
  }
  GAObjectiveVector & operator/=(const float value){
    for(unsigned int i=0; i<sz; i++) a[i] /= value;
    evaluated = gaFalse;
    return *this;
  }


// These behave as if the objective object were a vector.  We can add, 
// subtract, multiply, divide...
  GAObjectiveVector & operator+=(const GAObjectiveVector & orig){
    if(sz != orig.sz) return *this; // no error message!
    for(unsigned int i=0; i<sz; i++) a[i] += orig.a[i];
    evaluated = gaFalse;
    return *this;
  }
  GAObjectiveVector & operator-=(const GAObjectiveVector & orig){
    if(sz != orig.sz) return *this; // no error message!
    for(unsigned int i=0; i<sz; i++) a[i] -= orig.a[i];
    evaluated = gaFalse;
    return *this;
  }
  GAObjectiveVector & operator*=(const GAObjectiveVector & orig){
    if(sz != orig.sz) return *this; // no error message!
    for(unsigned int i=0; i<sz; i++) a[i] *= orig.a[i];
    evaluated = gaFalse;
    return *this;
  }
  GAObjectiveVector & operator/=(const GAObjectiveVector & orig){
    if(sz != orig.sz) return *this; // no error message!
    for(unsigned int i=0; i<sz; i++) a[i] /= orig.a[i];
    evaluated = gaFalse;
    return *this;
  }

// *** need to define + - * /


// The comparison operators do their thing based on the way the 
// multi-objective-to-scalar converter does its thing.
  friend int operator==(CONST GAObjectiveVector &a,CONST GAObjectiveVector &b){
    return a.evaluate() == b.evaluate();
  }
  friend int operator!=(CONST GAObjectiveVector &a,CONST GAObjectiveVector &b){
    return a.evaluate() != b.evaluate();
  }
  friend int operator<(CONST GAObjectiveVector &a,CONST GAObjectiveVector &b){
    return a.evaluate() < b.evaluate();
  }
  friend int operator>(CONST GAObjectiveVector &a,CONST GAObjectiveVector &b){
    return a.evaluate() > b.evaluate();
  }
  friend int operator<=(CONST GAObjectiveVector &a,CONST GAObjectiveVector &b){
    return a.evaluate() <= b.evaluate();
  }
  friend int operator>=(CONST GAObjectiveVector &a,CONST GAObjectiveVector &b){
    return a.evaluate() >= b.evaluate();
  }

// Input and output friends.  Just read the values in (separated by whitespace)
// or write them out (separated by single space character, no newline at end).
  friend ostream & operator<<(ostream & os, const GAObjectiveVector & arg){
    for(unsigned int i=0; i<arg.sz; i++) os << arg.a[i] << " ";
    return(os);
  }
  friend istream & operator>>(istream & is, GAObjectiveVector & arg){
    for(unsigned int i=0; i<arg.sz; i++) is >> arg.a[i];
    return(is);
  }

protected:
  unsigned int sz;		// number of components, number allocated
  float *a;			// the components of the objective
  GABoolean evaluated;		// is the score valid?
  float score;			// scalar score based on the components
  GAVectorObjectiveFunction ef;	// the evaluation function
};

#endif
