// $Header: /usr/people/mbwall/src/ga/ga/RCS/population.C,v 1.18 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  population.C
  mbwall 11aug94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  
 TO DO:

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <string.h>
#include <ga/config.h>
#include <ga/population.h>
#include <ga/selector.h>
#include <ga/random.h>
#include <ga/sort.h>

#ifndef NDEBUG
#define NDEBUG
#endif

#define GA_POP_CHUNKSIZE 10	// allocate chrom ptrs in chunks of this many

const DEFAULT_SCALING gaDefPopScaling; // Use linear scaling as the default


// This is the default population initializer.  It simply calls the initializer
// for each member of the population.  Then we touch the population to tell it
// that it needs to update stats and/or sort (but we don't actually force 
// either one to occur.
//   The population object takes care of setting/unsetting the status flags.
void GADefaultPopulationInitializer(GAPopulation & p){
  for(int i=0; i<p.size(); i++)
    p.individual(i).initialize();
}

// Sort from lowest to highest using the quicksort method.  If you write your
// own sorter be sure to sort from lowest to highest.  
//   The population object takes care of setting/unsetting the status flags.
void GADefaultPopulationSorter(GAPopulation & p){
#ifndef NDEBUG
  cerr << "GADefPopSorter: calling QuickSort\n";
#endif

  GAQuickSort(p.ind, 0, p.size()-1);
}

//   The population object takes care of setting/unsetting the status flags.
void GADefaultPopulationEvaluator(GAPopulation & p){
#ifndef NDEBUG
  static int count=0;
  count++;
  cerr << "GADefPopEval: count is " << count << "\n";
#endif

  for(int i=0; i<p.size(); i++)
    p.individual(i).score(); // do the calculation of each individual score
}




/* ----------------------------------------------------------------------------
 Population

  The population class is basically just a holder for the genomes.  We also
keep track of statistics about the fitness of our genomes.  We don't care
what kind of genomes we get.  To create the population we call the clone 
method of the genome we're given.
  The population must always contain at least one genome (otherwise we
don't know how to resize the population to a bigger size).
---------------------------------------------------------------------------- */
GAPopulation::GAPopulation(const GAGenome & c, const unsigned int popsize)
{
  csz = GA_POP_CHUNKSIZE;
  N = 0; n = (popsize < 1 ? 1 : popsize);
  while(N < n) N += csz;

  ind = new GAGenome * [N];
  for(int i=0; i<n; i++)
    ind[i] = c.clone(gaCloneAttributes);

  rawSum = rawAve = rawDev = rawVar = rawMax = rawMin = 0.0;
  statted = sorted = evaluated = gaFalse;
  im = GADefaultPopulationInitializer;
  ef = GADefaultPopulationEvaluator;
  sf = GADefaultPopulationSorter;
  sm = gaDefPopScaling.clone();
  sm->resize(n);
  ov = (GAObjectiveVector *)0;
}

GAPopulation::GAPopulation(const GAGenome & c, const GAScalingScheme & s,
			   const unsigned int popsize)
{
  csz = GA_POP_CHUNKSIZE;
  N = 0; n = (popsize < 1 ? 1 : popsize);
  while(N < n) N += csz;

  ind = new GAGenome * [N];
  for(int i=0; i<n; i++)
    ind[i] = c.clone(gaCloneAttributes);

  rawSum = rawAve = rawDev = rawVar = rawMax = rawMin = 0.0;
  statted = sorted = evaluated = gaFalse;
  im = GADefaultPopulationInitializer;
  ef = GADefaultPopulationEvaluator;
  sf = GADefaultPopulationSorter;
  sm = s.clone();
  sm->resize(n);
  ov = (GAObjectiveVector *)0;
}

GAPopulation::GAPopulation(const GAPopulation & orig){
  n=0; N=0; ind=(GAGenome **)0;
  sm=(GAScalingScheme *)0; ov=(GAObjectiveVector *)0;
  copy(orig);
}

GAPopulation::~GAPopulation()
{
  for(int i=0; i<n; i++)
    delete ind[i];
  delete [] ind;
  delete sm;
  delete ov;
}


// Make a complete copy of the original population.  We have to check to make
// sure that the genomes are the same in each population if we start out
// with some genomes.  If the population is empty, we just clone the 
// original.  If the population is full and the genomes do not match, we
// trash the old and deep copy the original.  If the genomes are the same
// type then we let the genomes copy themselves.

// For now we just destroy any existing population and deep copy the original.
void
GAPopulation::copy(const GAPopulation & arg)
{
  int i;
  for(i=0; i<n; i++)
    delete ind[i];
  delete [] ind;
  delete sm;
  delete ov;

  csz = arg.csz; N = arg.N; n = arg.n;
  ind = new GAGenome * [N];
  for(i=0; i<n; i++)
    ind[i] = arg.ind[0]->clone();

  if(arg.sm) sm = arg.sm->clone();
  else sm = (GAScalingScheme *)0;
  if(arg.ov) ov = arg.ov->clone();
  else ov = (GAObjectiveVector *)0;

  rawSum = arg.rawSum; rawAve = arg.rawAve; 
  rawMax = arg.rawMax; rawMin = arg.rawMin;
  rawVar = arg.rawVar; rawDev = arg.rawDev;
  sorted = arg.sorted;
  statted = arg.statted;
  evaluated = arg.evaluated;

  im = arg.im;
  ef = arg.ef;
  sf = arg.sf;
  ud = arg.ud;
  ga = arg.ga;
}


// Resize the population.  If we shrink, we delete the extra genomes.  If
// we grow, we clone new ones (and we DO initialize them!!!).  When we
// trash the genomes, we delete the worst of the population!  We do not
// free up the space used by the array of pointers, but we do free up the
// space used by the genomes.
//   We can never shrink to less than one genome.  If we did then we would
// not be able to resize to a larger size later on.
int
GAPopulation::resize(const unsigned int popsize)
{
  unsigned int pops = popsize;
  if(pops < 1) pops=1;
  if(pops == n){
    if(sm->size() != n) sm->resize(n); // just in case
    return n;
  }

  if(pops > n){
    GAGenome ** tmp = ind;
    while(N < pops) N += csz;
    ind = new GAGenome * [N];              // make more space
    memcpy(ind, tmp, n*sizeof(GAGenome*)); // copy the ones we still need
    delete [] tmp;			       // trash the old array
    for(int i=n; i<pops; i++){
      ind[i] = ind[0]->clone(gaCloneAttributes);
      ind[i]->initialize();
    }
  }
  else{
    for(int i=0; i<n-pops; i++)	// trash the worst ones
      delete ind[i];
    memmove(&(ind[0]), &(ind[n-pops]), pops*sizeof(GAGenome*));
  }

  sm->resize(pops);
  statted = sorted = evaluated = gaFalse;
  n = pops;

  return n;
}


// Get rid of 'extra' memory that we have allocated.
void
GAPopulation::compact()
{
  if(n == N) return;
  N = n;

  GAGenome ** tmp = ind;
  ind = new GAGenome * [N];
  memcpy(ind, tmp, n*sizeof(GAGenome *));
  delete [] tmp;
}


// Evaluate each member of the population and store basic population statistics
// in the member variables.  It is OK to run this on a const object - it 
// changes to physical state of the population, but not the logical state.
void
GAPopulation::statistics(const GABoolean flag) const
{
  if(statted && flag != gaTrue) return;
  GAPopulation * This = (GAPopulation *)this;
  This->rawMin = This->rawMax = This->rawSum = ind[0]->score();
  int i;
  for(i=1; i<n; i++){
    This->rawSum += ind[i]->score();
    This->rawMax = Max(rawMax, ind[i]->score());
    This->rawMin = Min(rawMin, ind[i]->score());
  }
  This->rawAve = rawSum/n;

  if(n > 1){
    float s;
    This->rawVar = 0;
    for(i=0; i<n; i++){
      s = ind[i]->score() - This->rawAve;
      s *= s;
      This->rawVar += s;
    }
    This->rawVar /= n-1;
  }
  else{
    This->rawVar = 0;
  }
  This->rawDev = sqrt(This->rawVar);

  This->statted = gaTrue;
}


// Return a reference to the scaling object.  We return it as a const so that
// the caller cannot change it.
const GAScalingScheme &
GAPopulation::scaling(const GAScalingScheme & s)
{
  if(s.sameClass(*sm)) return *sm;
  delete sm;
  sm = s.clone();
  sm->resize(n);
  return *sm;
}


// Replace the specified genome with the one that is passed to us then 
// return the one that got replaced.  Use the replacement flags to determine
// which genome will be replaced.  If we get a genome as the second
// argument, then replace that one.  If we get a NULL genome, then we
// return a NULL and don't do anything.
//   If the population is sorted, then we maintain the sort by doing a smart
// replacement.
//   If the population is not sorted, then we just do the replacement without
// worrying about the sort.  Replace best and worst both require that we know
// which chromsomes are which, so we do a sort before we do the replacement,
// then we do a smart replacement.
//   In both cases we flag the stats as out-of-date, but we do not update the
// stats.  Let that happen when it needs to happen.
//   If which is < 0 then it is a flag that tells us to do a certain kind of
// replacement.  Anything non-negative is assumed to be an index to a 
// genome in the population.
//   This does not affect the state of the evaluated member - it assumes that
// the individual genome has a valid number for its score.
GAGenome *
GAPopulation::replace(GAGenome * repl, const int which)
{
  register int lower, upper;
  int i=-1, j;
  GAGenome * orig=(GAGenome *)0;
  if(repl == (GAGenome *)0) return (GAGenome *)0;

  switch(which){
  case gaPopReplaceBest:
    if(sorted){
      i = n-1;			// assumes that best is element n-1
    }
    else{
      j=0;
      for(i=1; i<n; i++)
	if(ind[j]->score() < ind[i]->score()) j=i;
      i=j;
    }
    break;
      
  case gaPopReplaceWorst:
    if(sorted){
      i = 0;			// assumes that worst is element 0
    }
    else{
      j=0;
      for(i=1; i<n; i++)
	if(ind[j]->score() > ind[i]->score()) j=i;
      i=j;
    }
    break;

  case gaPopReplaceRandom:
    i = GARandomInt(0, n-1);
    break;

  default:
    if(0 <= which && which < n)
      i = which;
    break;
  }

  if(i >= 0){
    orig = ind[i];		// keep the original to return at the end

    if(sorted){			// put the replacement in the right place
      memmove(&(ind[i]), &(ind[i+1]), (n-i-1)*sizeof(GAGenome *));
      lower = 0; upper = n-2;   // this is our bin search for insert pt
      while(upper >= lower){	// we do the insertion later on.
	i = lower + (upper-lower)/2;
	if(ind[i]->score() > repl->score()) upper = i-1;
	else lower = i+1;
      }
      lower = Min(n-1, lower);
      lower = Max(0, lower);
      memmove(&(ind[lower+1]),&(ind[lower]),(n-lower-1)*sizeof(GAGenome*));
      ind[lower] = repl;	// Now do the insertion.
    }
    else{			// put the replacement where original was
      ind[i] = repl;
    }

    if(statted){		// incrementally update the pop statistics
      if(sorted)
	rawMin = ind[0]->score();
      else if(orig->score() == rawMin){
	rawMin = ind[0]->score();
	for(i=1; i<n; i++)
	  rawMin = Min(rawMin, ind[i]->score());
      }
      else
	rawMin = Min(rawMin, repl->score());

      if(sorted)
	rawMax = ind[n-1]->score();
      else if(orig->score() == rawMax){
	rawMax = ind[0]->score();
	for(i=1; i<n; i++)
	  rawMax = Max(rawMax, ind[i]->score());
      }
      else
	rawMax = Max(rawMax, repl->score());

      float s;
      rawVar *= n-1;
      s = orig->score() - rawAve;
      s *= s;
      rawVar -= s;

      rawSum += repl->score() - orig->score();
      rawAve = rawSum/n;

      s = repl->score() - rawAve;
      s *= s;
      rawVar += s;
      rawVar /= n-1;

#ifdef DEBUG
      if(rawVar < 0) cerr << "rawVar: " << rawVar << "\n";
#endif

      if(rawVar < 0) rawVar = 0;

// *** We have to do this check, but I don't know if truncating to 0 is the
// best thing to do.  We should never get a negative number for rawVar, but
// they do pop up - probably due to roundoff errors.

      rawDev = sqrt(rawVar);

      sm->touch();
// This is kind of ugly - it won't let you do any batch replacements.  As is,
// the scaling gets evaluated on each replacement rather than at the end of
// a sequence of replacements...
//      sm->evaluate(*this);
    }

// if not statted, don't stat it

  }

  return orig;
}


// Replace the genome o in the population with the genome r.  Return a
// pointer to the original genome, o.  This assumes that o exists in the
// population.   If it does not, we return a NULL.  If the genomes are the
// same, do nothing and return a pointer to the genome.
GAGenome *
GAPopulation::replace(GAGenome * r, GAGenome * o)
{
  GAGenome * orig=(GAGenome *)0;
  if(r == (GAGenome *)0 || o == (GAGenome *)0) return (GAGenome *)0;
  if(r == o) return r;
  int i;
  for(i=0; i<n && ind[i] != o; i++);
  if(i < n) orig = replace(r, i);
  return orig;
}


// Remove the xth genome from the population.  If x is out of bounds, we
// return NULL.  Otherwise we return a pointer to the genome that was 
// removed.  The population is now no longer responsible for freeing the
// memory used by that genome.  
// *** make this smarter
//   We could be smarted here about how we do the statistics and sorting, but
// for now we'll just reset everything.
GAGenome *
GAPopulation::remove(const unsigned int i)
{
  GAGenome * removed=(GAGenome *)0;
  if(i >= n) return (GAGenome *)0;
  removed = ind[i];
  memmove(&(ind[i]), &(ind[i+1]), (n-i-1)*sizeof(GAGenome *));
  n--;
  sm->resize(n);
  statted = sorted = evaluated = gaFalse;
  return removed;
}


// Remove the specified genome from the population.  If the genome is
// not in the population, we return NULL.
GAGenome *
GAPopulation::remove(GAGenome * r)
{
  GAGenome * removed=(GAGenome *)0;
  if(r == (GAGenome *)0) return (GAGenome *)0;
  int i;
  for(i=0; i<n && ind[i] != r; i++);
  if(i < n) removed = remove(i);
  return removed;
}


// Insert the specified individual into the population.  We don't do any 
// error checking here other than for limits on the specified location.
// If the location is bogus, return NULL.  Otherwise return pointer to the
// genome that was inserted.
//   Allocate more space if needed (but don't fill it with genomes).
// *** we should be smarter here about stats and sort
GAGenome * 
GAPopulation::insert(GAGenome * c, const unsigned int i)
{
  if(c == (GAGenome *)0 || i > n) return (GAGenome *)0;
  if(n+1 > N){
    GAGenome ** tmp = ind;
    while(N < n+1) N += csz;
    ind = new GAGenome * [N];
    memcpy(ind, tmp, n*sizeof(GAGenome *));
    delete [] tmp;
  }
  memmove(&(ind[i+1]), &(ind[i]), (n-i)*sizeof(GAGenome *));
  ind[i] = c;
  n++;
  sm->resize(n);
  statted = sorted = evaluated = gaFalse;
  return c;
}














/* linear search for best and worst population members
GAGenome &
GAPopulation::best(int n)
{
  int index=0;
  for(int i=1; i<n; i++)
    if(ind[i]->evaluate() > ind[index]->evaluate())
      index = i;
  return(*(ind[index]));
}

GAGenome &
GAPopulation::worst(int n)
{
  int index=0;
  for(int i=1; i<n; i++)
    if(ind[i]->evaluate() < ind[index]->evaluate())
      index = i;
  return(*(ind[index]));
}
*/
