// $Header: /usr/people/mbwall/src/ga/ga/RCS/population.h,v 1.19 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  population.h
  mbwall 3aug94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  The population holds an array of pointers to genomes.  It also keeps
track of the fitness statistics for the genomes in the population.

 TO DO:

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#ifndef _ga_population_h_
#define _ga_population_h_

#include <ga/id.h>
#include <ga/scaling.h>
#include <ga/objective.h>
#include <ga/genome.h>

const int gaPopReplaceBest = -1;
const int gaPopReplaceWorst = -2;
const int gaPopReplaceRandom = -3;


// The population initializer is called when you call the initialize member of
// a population object.  The default initializer simply calls the initializers
// of all of the population members.
//   Your sort routine MUST sort from lowest to highest or else everything will
// be screwed up.
typedef void (*GAPopulationInitializer)(GAPopulation &);
typedef void (*GAPopulationEvaluator)(GAPopulation &);
typedef void (*GAPopulationSorter)(GAPopulation &);


// These are the function declarations for the default population operators.
void GADefaultPopulationInitializer(GAPopulation &);
void GADefaultPopulationSorter(GAPopulation &);
void GADefaultPopulationEvaluator(GAPopulation &);



/* ----------------------------------------------------------------------------
Population

  The sort member is defined so that it can work on a const population.  It
does not change the logical state of the population, but it does change its
physical state.
---------------------------------------------------------------------------- */
class GAPopulation : public GAID {
  GADefineIdentity("GAPopulation", 10):

public:
  GAPopulation(const GAGenome & c, const unsigned int psize=1);
  GAPopulation(const GAGenome & c, const GAScalingScheme & scale,
	       const unsigned int popsize=1);
  GAPopulation(const GAPopulation & arg);
  GAPopulation & operator=(const GAPopulation & arg){copy(arg); return(*this);}
  virtual ~GAPopulation();
  void copy(const GAPopulation & arg);
  GAPopulation * clone() const {return new GAPopulation(*this);}

  int resize(const unsigned int popsize);
  int size() const {return n;}
  int chunksize() const {return csz;}
  int chunksize(const unsigned int csize){return csz=csize;}
  void compact();
  void touch(){statted = sorted = evaluated = gaFalse; sm->touch();}
  void statistics(const GABoolean flag=gaFalse) const;
  float sum() const {if(!statted) statistics(); return rawSum;}
  float ave() const {if(!statted) statistics(); return rawAve;}
  float var() const {if(!statted) statistics(); return rawVar;}
  float dev() const {if(!statted) statistics(); return rawDev;}
  float max() const {if(!statted) statistics(); return rawMax;}
  float min() const {if(!statted) statistics(); return rawMin;}
  float fitsum() const {return sm->sum();}
  float fitave() const {return sm->ave();}
  float fitmax() const {return sm->max();}
  float fitmin() const {return sm->min();}

  void evaluate(const GABoolean flag=gaFalse) {
    if(evaluated == gaFalse || flag == gaTrue){
      ef(*this); statted = gaFalse;
      sm->evaluate(*this);
    }
    evaluated = gaTrue;
  }
  GAPopulationEvaluator evaluator() const {return ef;}
  GAPopulationEvaluator evaluator(GAPopulationEvaluator e){return ef=e;}
  void initialize(){im(*this); touch();}
  GAPopulationInitializer initializer() const {return im;}
  GAPopulationInitializer initializer(GAPopulationInitializer i){return im=i;}
  void sort(const GABoolean flag=gaFalse) {
    if(sorted == gaFalse || flag == gaTrue) sf(*this);
    sorted = gaTrue;
  }
  GAPopulationSorter sorter() const {return sf;}
  GAPopulationSorter sorter(GAPopulationSorter s){return sf=s;}

  GA * geneticAlgorithm() const {return ga;}
  GA * geneticAlgorithm(GA & g){return(ga = &g);}
  void * userData() const {return ud;}
  void * userData(void * u){return(ud=u);}
  GAObjectiveVector * objectiveVector() const {return ov;}
  GAObjectiveVector * objectiveVector(const GAObjectiveVector & o)
    {delete ov; ov = o.clone(); return ov;}

  GAGenome & best(const unsigned int i=0) const {
    GAPopulation * This = (GAPopulation *)this;
    This->sort();
    return *(ind[n-1-i]);	// assumes that we sort from lowest to highest
  }
  GAGenome & worst(const unsigned int i=0) const {
    GAPopulation * This = (GAPopulation *)this;
    This->sort();
    return *(ind[i]);		// assumes that we sort from lowest to highest
  }
  GAGenome & operator[](const unsigned int x) const {return *(ind[x]);}
  GAGenome & individual(const unsigned int x) const {return *(ind[x]);}
  GAGenome * insert(GAGenome *, const unsigned int i);
  GAGenome * remove(const unsigned int i);
  GAGenome * remove(GAGenome *);
  GAGenome * replace(GAGenome *, const int which=gaPopReplaceRandom);
  GAGenome * replace(GAGenome *, GAGenome *);
  const GAScalingScheme & scaling() const {return *sm;}
  const GAScalingScheme & scaling(const GAScalingScheme & s);

  friend ostream & operator<<(ostream & os, const GAPopulation & arg)
    {arg._write(os); return os;}
  friend istream & operator>>(istream & is, GAPopulation & arg)
    {arg._read(is); return is;}

protected:
  unsigned int csz;		// how big are chunks we allocate?
  unsigned int n, N;		// how many are in the population, allocated
  GABoolean sorted;		// are the individuals sorted? (low to high)
  GABoolean statted;		// are the stats valid?
  GABoolean evaluated;		// has the population been evaluated?
  float rawSum, rawAve;		// sum, ave of the population's objectives
  float rawMax, rawMin;		// max, min of the population's objectives
  float rawVar, rawDev;		// variance, standard deviation
  GA * ga;			// the ga that is using this population
  void * ud;		        // pointer to user data
  GAScalingScheme * sm;		// scaling method
  GAPopulationInitializer im;	// initialization method
  GAPopulationSorter sf;	// sorting method
  GAPopulationEvaluator ef;	// population evaluation method
  GAObjectiveVector * ov;	// objective vector

  virtual void _read(istream &){}
  virtual void _write (ostream & os=cout) const {
    for(int i=0; i<n; i++)
      os << ind[i]->score() << "\t" << *ind[i] << "\n";
    os << "\n";
  }

public:				// only till I figure out a better way...
  GAGenome **ind;		// the individuals of the population
};

#endif
