// $Header: /usr/people/mbwall/src/ga/ga/RCS/random.h,v 1.3 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  random.h
  mbwall 29jun95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Formerly pieces of utilities.h, this file contains the random number 
generators needed in the GA library.

  Initialize the random function by using the current time as the seed.
  We use the pid as well as the time for the random seed in an attempt to get
rid of some of the periodicity from the low bits when using only the time as 
the random seed.
  These routines use the random/srandom routines to generate random numbers.
The documentation says that rand/srand is about 30% faster than random/srandom,
but rand/srand results that are less random.

GARandomInt
  Return a number selected at random within the bounds low and high, inclusive.
Don't forget the 'inclusive' part!  If you're using this to get the index of
an element in an array, be sure to use the last element of the array, not the
number of elements in the array as the arguement for 'high'.
  If you screw up and pass a value for low that is less than the value for
high, this routine does not complain.  It will merrily return a possibly 
negative value.  We need speed here, so there's no error checking.

GARandomFloat
  Same thing as random int but returns a float.  
GARandomBit
  Logically AND the random function with a single bit.  We cache the value we
  get from the random function then pop down the bits in order to reduce the
  number of calls to random.
GAFlipCoin
  Simulate a coin toss.  Use specified probability to bias toss.

  I should probably implement the random stuff as a class by itself.  For now
we'll just use functions (less overhead, and not much demand for custom random
functions at this point).
  We should be able to speed up this random number generator.  This is just 
a hack to get something working.  Also, I'd like to do something that is more
random - the built-in random functions are pretty good, but not truly random.
You can always pass an argument to GARandomSeed if you have a better seed
generator (maybe I'll get around to the network random number generator and
roll that into the code...)

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#ifndef _ga_random_h_
#define _ga_random_h_

#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <limits.h>
#include <ga/types.h>
#include <ga/config.h>

// If the machine has multiple processes, use the PID to help make the random
// number generator seed more random.
#ifdef NO_PID
#define _GA_PID
#else
#include <unistd.h>
#define _GA_PID * (int)getpid()
#endif




// Here we determine which random number generator will be used.  The critical
// parts here are the name of the random number generator (e.g. rand or random)
// the name of the seed routine (e.g. srand or srandom), the maximum value of
// the number returned by the generator (e.g. RAND_MAX or LONG_MAX), and the
// C++ type of the number returned (e.g. int or long int).
//   If you want to use your own random function, this is where to declare it.
// Just set the _GA_RND macros to the appropriate values for your random number
// generator.

#ifdef NO_RANDOM

#define _GA_RND             rand
#define _GA_RND_SEED        srand
#define _GA_RND_MAX         RAND_MAX
#define _GA_RND_TYPE        int

#else

// Some machines don't have random and srandom defined in their headers (at
// least not a c++-safe version of the headers...) when they do show up in the
// math library.
#ifdef NO_CPLUSPLUS_MATH
extern "C" {
  extern long random(void);
  extern int srandom(int seed);
  extern char *initstate(unsigned int seed, char *state, int n);
  extern char *setstate(char *state);
}
#endif

#define _GA_RND             random
#define _GA_RND_SEED        srandom
#define _GA_RND_MAX         LONG_MAX
#define _GA_RND_TYPE        long

#endif



inline void GARandomSeed(){_GA_RND_SEED ((int)time(NULL) _GA_PID);}
inline void GARandomSeed(const unsigned s){_GA_RND_SEED (s);}

inline int GARandomInt(){
  unsigned long val=2*_GA_RND(); val/=(unsigned long)_GA_RND_MAX+1;
  return (int)val;
}
inline int GARandomInt(const int low, const int high){ 
  float val=high-low+1; val*=_GA_RND(); val/=(float)_GA_RND_MAX+1; val+=low;
  return (int)val;
}

inline double GARandomDouble(){
  double val=_GA_RND(); val/=_GA_RND_MAX;
  return val;
}
inline double GARandomDouble(const double low, const double high){
  double val=high-low; val*=_GA_RND(); val/=_GA_RND_MAX; val+=low;
  return val;
}

inline float GARandomFloat(){
  float val=_GA_RND(); val/=_GA_RND_MAX;
  return val;
}
inline float GARandomFloat(const float low, const float high){
  float val=high-low; val*=_GA_RND(); val/=_GA_RND_MAX; val+=low;
  return val;
}
inline GABoolean GAFlipCoin(float p){
  return((p == 1.0) ? gaTrue : 
	 (p == 0.0) ? gaFalse :
	 ((GARandomFloat() <= p) ? gaTrue : gaFalse));
}

int GARandomBit();

#endif
