// $Header: /usr/people/mbwall/src/ga/ga/RCS/scaling.C,v 1.10 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  fitness.C
  mbwall 10aug94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Class definitions for the scaling objects.  These objects translate (scale)
the raw objective scores of a population into 'fitness' scores that are used to
determine which genomes are fit for mating/selection.
  In all of these routines we do a check to be sure that we have enough space
to do our evaluations.  We don't need to do this so long as the population 
objects that call us do the test themselves.  I'll leave the redundancy for
now.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <math.h>
#include <ga/error.h>
#include <ga/scaling.h>
#include <ga/genome.h>
#include <ga/population.h>
#include <ga/sort.h>

#ifndef NDEBUG
#define NDEBUG
#endif

#define GA_SCALING_CHUNKSIZE 10


/* ----------------------------------------------------------------------------
Scaling base class methods
---------------------------------------------------------------------------- */
GAScalingScheme::GAScalingScheme(const unsigned int size)
{
  csz = GA_SCALING_CHUNKSIZE;
  N = 0; n = size;
  evaluated = gaFalse;
  while(N < n) N += csz;
  fit = new float[N];
  psum = new float[N];
  ind = new GAGenome * [N];
  fitSum = fitAve = fitMax = fitMin = fitVar = fitDev = 0.0;
}

GAScalingScheme::GAScalingScheme(const GAPopulation & p)
{
  csz = p.chunksize();
  N = 0; n = p.size();
  evaluated = gaFalse;
  while(N < n) N += csz;
  fit = new float[N];
  psum = new float[N];
  ind = new GAGenome * [N];
  fitSum = fitAve = fitMax = fitMin = fitVar = fitDev = 0.0;
}

GAScalingScheme::~GAScalingScheme()
{
  delete [] fit;
  delete [] psum;
  delete [] ind;
}

void
GAScalingScheme::copy(const GAScalingScheme & s)
{
  if(&s == this) return;
  csz = s.csz;
  evaluated = s.evaluated;
  resize(s.n);
  memcpy(fit, s.fit, n*sizeof(float));
  memcpy(psum, s.psum, n*sizeof(float));
  memcpy(ind, s.ind, n*sizeof(GAGenome *));
  fitSum = s.fitSum; fitAve = s.fitAve;
  fitMax = s.fitMax; fitMin = s.fitMin;
  fitVar = s.fitVar; fitDev = s.fitDev;
}


// This is how to resize a scaling object.  Call this method to initialize the
// scaling and any time your population changes size.  If the population is
// smaller than the array of fits that we have, then we don't do any memory
// reallocation.  In any case, we set all of the fits to 0.
//   We do not do any evaluate or recalculation of statistics here.
int
GAScalingScheme::resize(const unsigned int popsize)
{
  if(N < popsize){
    while(N < popsize) N += csz;
    delete [] fit;
    delete [] psum;
    delete [] ind;
    fit = new float[N];
    psum = new float[N];
    ind = new GAGenome * [N];
  }
  memset(fit, 0, N*sizeof(float));
  memset(psum, 0, N*sizeof(float));
  memset(ind, 0, N*sizeof(GAGenome *));
  fitSum = fitAve = fitMax = fitMin = fitVar = fitDev = 0.0;
  return(n = popsize);
}


void
GAScalingScheme::compact()
{
  if(n == N) return;
  N = n;

  float * tmpfit = fit;
  fit = new float[N];
  memcpy(fit, tmpfit, n*sizeof(float));
  delete [] tmpfit;

  float * tmppsum = psum;
  psum = new float[N];
  memcpy(psum, tmppsum, n*sizeof(float));
  delete [] tmppsum;

  GAGenome ** tmpind = ind;
  ind = new GAGenome * [N];
  memcpy(ind, tmpind, n*sizeof(float));
  delete [] tmpind;
}
  




/* ----------------------------------------------------------------------------
NoScaling
---------------------------------------------------------------------------- */
// Dummy scaling function.  Just copies the objective scores and calculates the
// partial sums.  
void 
GANoScaling::evaluate(const GAPopulation & p)
{
  if(evaluated == gaTrue) return;

  memcpy(ind, p.ind, n*sizeof(GAGenome*));
  int i;
  for(i=0; i<n; i++)
    fit[i] = ind[i]->score();

#ifndef NDEBUG
  cerr << "GANoScaling::evaluate(): calling QuickSort\n";
#endif

  GAQuickSort(ind, fit, 0, n-1);

  fitMin = fitMax = fitSum = psum[0] = fit[0];
  for(i=1; i<n; i++){
    fitSum += fit[i];
    psum[i] = fitSum;
    if(fit[i] > fitMax) fitMax = fit[i];
    if(fit[i] < fitMin) fitMin = fit[i];
  }
  fitAve = fitSum/n;

  evaluated = gaTrue;
}



#if USE_LINEAR_SCALING == 1
/* ----------------------------------------------------------------------------
LinearScaling
---------------------------------------------------------------------------- */
// Scale the objective scores in the population.  We assume that the raw
// evaluation has already taken place (the calculation of the objective scores
// of the genomes in the population).  We force a sort to be sure that the
// genomes are in order from smallest to largest (in terms of objective
// score).

// Is this sort necessary???????
// check for negative scaled values, not negative raw scores?

void 
GALinearScaling::evaluate(const GAPopulation & p)
{
  if(evaluated == gaTrue) return;

// Here we calculate the slope and intercept using the multiplier and objective
// score ranges...

  float delta, a, b;
  if(p.ave() == p.max()){	// no scaling - population is all the same
    a = 1; 
    b = 0;
  }
  else if(p.min() > (c * p.ave() - p.max())/(c - 1.0)){
    delta = p.max() - p.ave();
    a = (c - 1.0) * p.ave() / delta;
    b = p.ave() * (p.max() - c * p.ave()) / delta;
  }
  else{				// stretch to make min be 0
    delta = p.ave() - p.min();
    a = p.ave() / delta;
    b = -p.min() * p.ave() / delta;
  }

// and now we calculate the scaled scaling values and partial sums (we use the
// partial sums for the roulette wheel selection).  Negative scaling is not
// allowed with this kind of scaling, so check for negative values.  If we get
// a negative value, dump an error message then set all of the scores to 0.

  memcpy(ind, p.ind, n*sizeof(GAGenome*));
  float f;
  int i;
  for(i=0; i<n; i++){
    if((f = ind[i]->score()) < 0.0){
      GAErr(className(), "evaluate", gaErrNegFitness);
      memset(fit, 0, n*sizeof(float));
      memset(psum, 0, n*sizeof(float));
      return;
    }
    fit[i] = f * a + b;
    if(fit[i] < 0) fit[i] = 0;
  }

#ifndef NDEBUG
  cerr << "GALinearScaling::evaluate(): calling QuickSort\n";
#endif

  GAQuickSort(ind, fit, 0, n-1);

  fitMin = fitMax = fitSum = psum[0] = fit[0];
  for(i=1; i<n; i++){
    fitSum += fit[i];
    psum[i] = fitSum;
    if(fit[i] > fitMax) fitMax = fit[i];
    if(fit[i] < fitMin) fitMin = fit[i];
  }
  fitAve = fitSum/n;

  evaluated = gaTrue;
}


// Set the multiplier for this selection type.  The fmultiplier must be greater
// than 1.0 or else we'll get a divide by zero error in our scaling operations.
float
GALinearScaling::multiplier(const float fm)
{
  if(fm <= 1.0){
    GAErr(className(), "multiplier", gaErrBadLinearScalingMult);
    return c;
  }
  return(c = fm);
}

#endif



#if USE_SIGMA_TRUNC_SCALING == 1
/* ----------------------------------------------------------------------------
SigmaTruncationScaling
---------------------------------------------------------------------------- */
// This is an implementation of the sigma truncation scaling method descibed in
// goldberg p 124.  If the scaled fitness is less than zero, we arbitrarily set
// it to zero (thus the truncation part of 'sigma truncation').
void 
GASigmaTruncationScaling::evaluate(const GAPopulation & p)
{
  if(evaluated == gaTrue) return;

  memcpy(ind, p.ind, n*sizeof(GAGenome*));
  int i;
  for(i=0; i<n; i++){
    fit[i] = ind[i]->score() - p.ave() + c * p.dev();
    if(fit[i] < 0) fit[i] = 0;
  }

#ifndef NDEBUG
  cerr << "GASigmaTruncScaling::evaluate(): calling QuickSort\n";
#endif

  GAQuickSort(ind, fit, 0, n-1);

  fitMin = fitMax = fitSum = psum[0] = fit[0];
  for(i=1; i<n; i++){
    fitSum += fit[i];
    psum[i] = fitSum;
    if(fit[i] > fitMax) fitMax = fit[i];
    if(fit[i] < fitMin) fitMin = fit[i];
  }
  fitAve = fitSum/n;

  evaluated = gaTrue;
}


// Set the multiplier for this selection type.  The fmultiplier must be greater
// than 1.0 or else we'll get a divide by zero error in our scaling operations.
float
GASigmaTruncationScaling::multiplier(const float fm)
{
  if(fm <= 0.0){
    GAErr(className(), "multiplier", gaErrBadSigmaTruncationMult);
    return c;
  }
  return(c = fm);
}

#endif



#if USE_POWER_LAW_SCALING == 1
/* ----------------------------------------------------------------------------
PowerLawScaling
---------------------------------------------------------------------------- */
// This is an implementation of the most basic form of power scaling, where the
// fitness is a function of the objective score raised to some power.  Negative
// objective scores are not allowed.  If we get one, we post an error and set
// all of the fitness scores to zero.
void 
GAPowerLawScaling::evaluate(const GAPopulation & p)
{
  if(evaluated == gaTrue) return;

  memcpy(ind, p.ind, n*sizeof(GAGenome*));
  float f;
  int i;
  for(i=0; i<n; i++){
    if((f = ind[i]->score()) < 0.0){
      GAErr(className(), "evaluate", gaErrPowerNegFitness);
      memset(fit, 0, n*sizeof(float));
      memset(psum, 0, n*sizeof(float));
      return;
    }
    fit[i] = (float)pow((double)f,(double)k);
  }
  
#ifndef NDEBUG
  cerr << "GAPowerLawScaling::evaluate(): calling QuickSort\n";
#endif

  GAQuickSort(ind, fit, 0, n-1);

  fitMin = fitMax = fitSum = psum[0] = fit[0];
  for(i=1; i<n; i++){
    fitSum += fit[i];
    psum[i] = fitSum;
    if(fit[i] > fitMax) fitMax = fit[i];
    if(fit[i] < fitMin) fitMin = fit[i];
  }
  fitAve = fitSum/n;

  evaluated = gaTrue;
}
#endif



#if USE_SHARING == 1
/* ----------------------------------------------------------------------------
Sharing
---------------------------------------------------------------------------- */
// This is an implementation of speciation using the sharing method described
// by goldberg in his book.  This requires a user-defined distance function in
// order to work.  The distance function returns a value between
// 0 and 1 inclusive to tell us how similar two genomes are to each other.
// A value of 1 means that the two genomes are identical to each other, a
// value of 0 means they are completely different.
//   A single genome is identical to itself, so d(i,i) is one. 
//   This is an implementation of the triangular sharing function similar to
// that described in goldberg's book.  The only difference is that here our 
// distance function always returns [0,1] and 1 means identical.  Goldberg
// defines distance as [0,sigma] and 0 means identical.
//   We use the distance function to scale the objective scores of the 
// genomes so that if there are many similar genomes in a population 
// their scores will be decreased, thus giving sub-species a greater chance to 
// reproduce.
//   We can cut in half the number of calls to the sharing function by keeping
// one half of the ixj matrix.  This is because d(i,j) is the same as d(j,i).
// We cache the distances in an upper right triangular matrix stored as a 
// series of floats.
// *** This could be done with n*n/2 instead of n*n, we can optimize later...
void 
GASharing::evaluate(const GAPopulation & p)
{
  if(evaluated == gaTrue) return;

  memcpy(ind, p.ind, n*sizeof(GAGenome*));
  int i;
  for(i=0; i<n; i++){		// calculate and cache the distances
    distance[i*n+i] = 1;	// each genome is same as itself
    for(int j=i+1; j<n; j++)
      distance[i*n+j] = distance[j*n+i] = df(*ind[i], *ind[j]);
  }

  float sum;			// now derate the fitness of each genome
  for(i=0; i<n; i++){		
    sum = 0.0;
    for(int j=0; j<n; j++)
      sum += distance[i*n+j];
    fit[i] = ind[i]->score() / sum;
  }

#ifndef NDEBUG
  cerr << "GASharing::evaluate(): calling QuickSort\n";
#endif

  GAQuickSort(ind, fit, 0, n-1);

  fitMin = fitMax = fitSum = psum[0] = fit[0];
  for(i=1; i<n; i++){
    fitSum += fit[i];
    psum[i] = fitSum;
    if(fit[i] > fitMax) fitMax = fit[i];
    if(fit[i] < fitMin) fitMin = fit[i];
  }
  fitAve = fitSum/n;

  evaluated = gaTrue;
}
#endif
