// $Header: /usr/people/mbwall/src/ga/ga/RCS/scaling.h,v 1.10 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  scaling.h
  mbwall 10aug94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  This is the GAScaling class.  It is used to do scaling on a population.  When
a genome is evaluated, the user's objective function provides an overall
rating for each genome.  The GA is concerned with fitness, not objective
score (unless you do no scaling).  So this object is the container for the
scaled values.
  Examples of scaling include linear scaling, sigma truncation, and power law.
Goldberg's sharing function is also a type of scaling, and it is implemented 
here as a unique class.
  The scaling class is designed to be used with a population.  It is stupid -
it does know how to update itself, but it must be told when.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#ifndef _ga_scaling_h_
#define _ga_scaling_h_

#include <ga/config.h>
#include <ga/id.h>
#include <ga/genome.h>

class GAPopulation;

// A distance function is used with the sharing object.  It is a type
// of speciation (similar in functionality to DeJong crowding, but this uses
// fitness scaling rather than replacement strategy to affect the speciation).
// If the genomes are identical, return a value of 1.0, if completely 
// different then return a value of 0.0
typedef float (*GADistanceFunction)(GAGenome &, GAGenome &);


/* ----------------------------------------------------------------------------
 Default values
---------------------------------------------------------------------------- */
const int   gaDefSize                      = 30;
const float gaDefLinearScalingMultiplier   = 1.2;
const float gaDefSigmaTruncationMultiplier = 2.0;
const float gaDefPowerScalingFactor        = 1.0005;



/* ----------------------------------------------------------------------------
Scaling

  The scaling object is used to scale the objective scores of a population to
avoid clustering and premature convergence (among other things).  See golberg
for more about the theory.  This class basically just contains an array of
values that map directly to the genomes in the population that is being
scaled.  The derived classes contain additional information that they need to
do their scaling operations.
  This is a pretty stupid object.  It doesn't keep track of whether its 
information is up to date, nor does it care which population it does its stats
on.  It is designed primarily as a container.
  We keep our own array of the population members so that they don't get 
confounded with the ordering in the population object.  We don't do any 
allocation here - the pointers in the array of genomes are all allocated
somewhere else.
---------------------------------------------------------------------------- */
class GAScalingScheme : public GAID {
  GADefineIdentity("GAScalingScheme", 15):

public:
  GAScalingScheme(const unsigned int n=0);
  GAScalingScheme(const GAPopulation & p);
  GAScalingScheme(const GAScalingScheme & s){copy(s);}
  virtual ~GAScalingScheme();
  int size() const {return n;}
  virtual void copy(const GAScalingScheme & arg);
  virtual int resize(const unsigned int popsize);
  virtual GAScalingScheme * clone() const=0;
  virtual void evaluate(const GAPopulation & p)=0;

  void touch(){evaluated = gaFalse;}
  int chunksize() const {return csz;}
  int chunksize(const unsigned int csize){return csz=csize;}
  void compact();

  float sum() const {return fitSum;}
  float ave() const {return fitAve;}
  float max() const {return fitMax;}
  float min() const {return fitMin;}
  float fitness(const unsigned int i) const {return fit[i];}
  float partialSum(const unsigned int i) const {return psum[i];}

  GAGenome & best(const unsigned int i=0) const {return *(ind[n-1-i]);}
  GAGenome & worst(const unsigned int i=0) const {return *(ind[i]);}
  GAGenome & individual(const unsigned int x) const {return *(ind[x]);}

protected:
  unsigned int n, N;		// how many in the array?  how much allocated?
  unsigned int csz;		// how many to allocate in each chunk?
  GABoolean evaluated;		// has the population been evaluated?
  float *fit;			// fitness (not objective) of each genome
  float *psum;			// partial sum of fitnesses
  float fitSum, fitAve;		// sum, ave of the population's fitnesses
  float fitMax, fitMin;		// max, min of the population's fitnesses
  float fitVar, fitDev;		// variance, standard deviation
  GAGenome **ind;		// the individuals of the population
};


/* ----------------------------------------------------------------------------
NoScaling

  This scaling object does no scaling.  It simply copies the raw objective 
scores into its array of fitness values.
---------------------------------------------------------------------------- */
class GANoScaling : public GAScalingScheme {
  GADefineIdentity("GANoScaling", 16):

public:
  GANoScaling(const unsigned int n=gaDefSize) : GAScalingScheme(n){}
  GANoScaling(const GAPopulation & p) : GAScalingScheme(p){}
  GANoScaling(const GANoScaling & arg) : GAScalingScheme(arg.n){}
  GANoScaling & operator=(const GAScalingScheme & arg)
    {copy(arg); return(*this);}
  GANoScaling & operator=(const GANoScaling & arg){copy(arg); return(*this);}
  virtual ~GANoScaling(){}
  GAScalingScheme * clone() const {return new GANoScaling(*this);}
  void evaluate(const GAPopulation & p);
};


/* ----------------------------------------------------------------------------
LinearScaling

  This scaling object does linear scaling as described in goldberg pp 122-124.
---------------------------------------------------------------------------- */
#if USE_LINEAR_SCALING == 1
class GALinearScaling : public GAScalingScheme {
  GADefineIdentity("GALinearScaling", 17):

public:
  GALinearScaling(const float fm=gaDefLinearScalingMultiplier) :
    GAScalingScheme(gaDefSize) {multiplier(fm);}
  GALinearScaling(const GAPopulation & p,
		  const float fm=gaDefLinearScalingMultiplier) :
    GAScalingScheme(p) {multiplier(fm);}
  GALinearScaling(const GALinearScaling & arg) : GAScalingScheme(arg.n)
    {copy(arg);}
  GALinearScaling & operator=(const GAScalingScheme & arg)
    {copy(arg); return(*this);}
  GALinearScaling & operator=(const GALinearScaling & arg)
    {copy(arg); return(*this);}
  virtual ~GALinearScaling(){}
  void copy(const GAScalingScheme & arg){
    if(&arg != this && sameClass(arg))
      GAScalingScheme::copy(arg); c = ((GALinearScaling&)arg).c;
  }
  GAScalingScheme * clone() const {return new GALinearScaling(*this);}
  void evaluate(const GAPopulation & p);

  float multiplier(const float fm);
  float multiplier() const {return c;}

protected:
  float c;			// linear scaling multiplier
};
#endif



/* ----------------------------------------------------------------------------
SigmaTruncationScaling

  This scaling object does sigma truncation as defined in goldberg p124.
---------------------------------------------------------------------------- */
#if USE_SIGMA_TRUNC_SCALING == 1
class GASigmaTruncationScaling : public GAScalingScheme {
  GADefineIdentity("GASigmaTruncationScaling", 18):

public:
  GASigmaTruncationScaling(const float m=gaDefSigmaTruncationMultiplier) :
    GAScalingScheme(gaDefSize){multiplier(m);}
  GASigmaTruncationScaling(const GAPopulation & p, 
			   const float m=gaDefSigmaTruncationMultiplier) :
    GAScalingScheme(p){multiplier(m);}
  GASigmaTruncationScaling(const GASigmaTruncationScaling & arg) : 
    GAScalingScheme(arg.n){copy(arg);}
  GASigmaTruncationScaling & operator=(const GAScalingScheme & arg)
    {copy(arg); return(*this);}
  GASigmaTruncationScaling & operator=(const GASigmaTruncationScaling & arg)
    {copy(arg); return(*this);}
  virtual ~GASigmaTruncationScaling(){}
  void copy(const GAScalingScheme & arg){
    if(&arg != this && sameClass(arg))
      GAScalingScheme::copy(arg); c=((GASigmaTruncationScaling&)arg).c;
  }
  GAScalingScheme * clone() const {return new GASigmaTruncationScaling(*this);}
  void evaluate(const GAPopulation & p);

  float multiplier(const float fm);
  float multiplier() const {return c;}

protected:
  float c;			// std deviation multiplier
};
#endif



/* ----------------------------------------------------------------------------
PowerLawScaling

  This scaling object does power law scaling as defined in goldberg p124.
---------------------------------------------------------------------------- */
#if USE_POWER_LAW_SCALING == 1
class GAPowerLawScaling : public GAScalingScheme {
  GADefineIdentity("GAPowerLawScaling", 19):

public:
  GAPowerLawScaling(const float f=gaDefPowerScalingFactor) :
    GAScalingScheme(gaDefSize){k = f;}
  GAPowerLawScaling(const GAPopulation & p, 
		    const float f=gaDefPowerScalingFactor) :
    GAScalingScheme(p){k = f;}
  GAPowerLawScaling(const GAPowerLawScaling & arg) : GAScalingScheme(arg.n)
    {copy(arg);}
  GAPowerLawScaling & operator=(const GAScalingScheme & arg)
    {copy(arg); return(*this);}
  GAPowerLawScaling & operator=(const GAPowerLawScaling & arg)
    {copy(arg); return(*this);}
  virtual ~GAPowerLawScaling(){}
  void copy(const GAScalingScheme & arg){
    if(&arg != this && sameClass(arg))
      GAScalingScheme::copy(arg); k = ((GAPowerLawScaling&)arg).k;
  }
  GAScalingScheme * clone() const {return new GAPowerLawScaling(*this);}
  void evaluate(const GAPopulation & p);

  float power(const float p){return(k=p);}
  float power() const {return k;}

protected:
  float k;			// power scaling factor
};
#endif



/* ----------------------------------------------------------------------------
Sharing

  This scaling object does sharing as described in goldberg p 192.  This 
implementation does triangular sharing.  If you want a different type of 
sharing function, then derive from this class and define a new (virtual)
evaluate method and add whatever member data you need to specify the shape
of your sharing function.
  Distance functions are independent of the sharing functions (as defined in
Goldberg, that is).  This implementation is functionally equivalent to Goldberg
but the implementation is slightly different than what he describes.
*** This should be called TriangularSharing rather than simply Sharing.
---------------------------------------------------------------------------- */
#if USE_SHARING == 1
class GASharing : public GAScalingScheme {
  GADefineIdentity("GASharing", 20):

public:
  GASharing(GADistanceFunction f)
    {distance = new float[n*n]; df=f;}
  GASharing(const GAPopulation & p, GADistanceFunction f) : GAScalingScheme(p)
    {distance = new float[n*n]; df=f;}
  GASharing(const GASharing & arg) : GAScalingScheme(arg.n)
    {distance=(float*)0; copy(arg);}
  GASharing & operator=(const GAScalingScheme & arg){copy(arg); return(*this);}
  GASharing & operator=(const GASharing & arg){copy(arg); return(*this);}
  virtual ~GASharing(){ delete [] distance;}
  void copy(const GAScalingScheme & arg){
    if(&arg != this && sameClass(arg)){
      GAScalingScheme::copy(arg);
      df=((GASharing&)arg).df;
      delete [] distance;
      distance = new float[n*n];
      memcpy(distance, ((GASharing&)arg).distance, n*n*sizeof(float));
    }
  }
  int resize(const unsigned int popsize){
    GAScalingScheme::resize(popsize);
    delete [] distance;
    distance = new float[n*n];
    return popsize;
  }
  GAScalingScheme * clone() const {return new GASharing(*this);}
  void evaluate(const GAPopulation & p);

  GADistanceFunction distanceFunction(GADistanceFunction f){return(df = f);}
  GADistanceFunction distanceFunction() const {return df;}

protected:
  GADistanceFunction df;		// the user-defined distance function
  float *distance;			// cache the distances
};
#endif


#endif
