// $Header: /usr/people/mbwall/src/ga/ga/RCS/selector.C,v 1.19 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  selector.C
  mbwall 10aug94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  
 TO DO:

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <ga/config.h>
#include <ga/selector.h>
#include <ga/random.h>
#include <ga/sort.h>

#ifndef NDEBUG
#define NDEBUG
#endif


/* ----------------------------------------------------------------------------
RankSelector
---------------------------------------------------------------------------- */
GAGenome &
GARankSelector(const GAPopulation & p)
{
  return(p.scaling().best());
}



/* ----------------------------------------------------------------------------
RouletteWheelSelector
---------------------------------------------------------------------------- */
// This selection routine is straight out of Goldberg's Genetic Algorithms 
// book.  We look through the members of the population using a weighted 
// roulette wheel.

// This is the binary search method (using cached partial sums).  It assumes 
// that the genomes are in order from lowest to highest.
#if USE_ROULETTE_SELECTOR == 1
GAGenome &
GARouletteWheelSelector(const GAPopulation & p)
{
  static float cutoff;
  static int i, upper, lower;

  cutoff = GARandomFloat() * p.scaling().sum();
  lower = 0; upper = p.size()-1;
  while(upper >= lower){
    i = lower + (upper-lower)/2;
    if(p.scaling().partialSum(i) > cutoff)
      upper = i-1;
    else
      lower = i+1;
  }
  lower = Min(p.size()-1, lower);
  lower = Max(0, lower);
  return(p.scaling().individual(lower));
}

#endif

// Here's the linear search version of the selector (using cached partial sums)
/*
GAGenome &
GARouletteWheelSelector(const GAPopulation & p)
{
  float cutoff = GARandomFloat() * p.scaling().sum();
  int i = 0;
  while(p.scaling().partialSum(i) <= cutoff && i < p.size()) i++;
  return(p.scaling().individual(i));
}
*/


/* ----------------------------------------------------------------------------
TournamentSelector

  Pick two individuals from the population using the RouletteWheel selection
method.  Then return the better of the two individuals.
---------------------------------------------------------------------------- */
#if USE_TOURNAMENT_SELECTOR == 1
GAGenome &
GATournamentSelector(const GAPopulation & p)
{
  static int picked;
  static float cutoff;
  static int i, upper, lower;

  cutoff = GARandomFloat() * p.scaling().sum();
  lower = 0; upper = p.size()-1;
  while(upper >= lower){
    i = lower + (upper-lower)/2;
    if(p.scaling().partialSum(i) > cutoff)
      upper = i-1;
    else
      lower = i+1;
  }
  lower = Min(p.size()-1, lower);
  lower = Max(0, lower);
  picked = lower;

  cutoff = GARandomFloat() * p.scaling().sum();
  lower = 0; upper = p.size()-1;
  while(upper >= lower){
    i = lower + (upper-lower)/2;
    if(p.scaling().partialSum(i) > cutoff)
      upper = i-1;
    else
      lower = i+1;
  }
  lower = Min(p.size()-1, lower);
  lower = Max(0, lower);

  if(p.scaling().individual(lower).score() >
     p.scaling().individual(picked).score())
    picked = lower;

  return(p.scaling().individual(picked));
}
#endif

/* ----------------------------------------------------------------------------
SRSSelector - stochastic remainder sampling

  The selection happens in two stages.  First we generate an array using the
integer and remainder parts of the expected number of individuals.  Then we
pick an individual from the population by randomly picking from this array.
  This is implemented just as in Goldberg's book.  Not very efficient...  In
Goldberg's implementation he uses a variable called 'nremain' so that multiple
calls to the selection routine can be dependent upon previous calls.  We don't
have that option with this architecture; we would need to make selection an
object coupled closely with the population to make that work.
---------------------------------------------------------------------------- */
#if USE_SRS_SELECTOR == 1
GAGenome &
GASRSSelector(const GAPopulation & p)
{
  static float *fraction=(float *)0;
  static unsigned int *choices=(unsigned int *)0;
  static int N=0;
  float expected;
  int n, k=0;

// Make sure we have enough memory to work with.  Maybe this should be an
// object?  But we don't want all of that overhead...

  if(N < p.size()){
    delete [] fraction;
    delete [] choices;
    N = p.size();
    fraction = new float[N];
    choices = new unsigned int[N];
  }

// This is the preselection part.  Figure out how many we should expect of 
// each individual.  An individual with e of 4.3 will get 4 places and have a
// 30% chance of a 5th place.

  int i;
  for(i=0; i<p.size(); i++){
    expected = p.scaling().fitness(i)/p.scaling().ave();
    n = (int)expected;
    fraction[i] = expected - n;
    while(n > 0){
      choices[k] = i;
      k++; n--;
    }
  }
  i=0;
  while(k < p.size()){
    if(fraction[i] > 0.0 && GAFlipCoin(fraction[i])){
      choices[k] = i;
      fraction[i] -= 1.0;
      k++;
    }
    i++;
    if(i >= p.size()) i=0;
  }

// Now just pick randomly from the list we generated in the preselection step.

  return(p.scaling().individual(choices[GARandomInt(0, p.size()-1)]));
}

#endif


/* ----------------------------------------------------------------------------
SUSSelector - stochastic uniform sampling

  Randomly select an individual from the population.  Any individual has as 
much chance of getting selected as any other.
---------------------------------------------------------------------------- */
#if USE_SUS_SELECTOR == 1
GAGenome &
GASUSSelector(const GAPopulation & p)
{
  return(p.individual(GARandomInt(0,p.size()-1)));
}
#endif


/* ----------------------------------------------------------------------------
DSSelector - deterministic sampling

  This is implemented as described in Goldberg's book.
---------------------------------------------------------------------------- */
#if USE_DS_SELECTOR == 1
GAGenome &
GADSSelector(const GAPopulation & p)
{
  static float *fraction=(float *)0;
  static unsigned int *choices=(unsigned int *)0;
  static unsigned int *idx=(unsigned int *)0;
  static int N=0;
  float expected;
  int n, k=0;

// Make sure we have enough memory to work with.  Maybe this should be an
// object?  But we don't want all of that overhead...

  if(N < p.size()){
    delete [] fraction;
    delete [] choices;
    delete [] idx;
    N = p.size();
    fraction = new float[N];
    choices = new unsigned int[N];
    idx = new unsigned int[N];
  }

// This is the preselection part.  Figure out how many we should expect of 
// each individual.  An individual with e of 4.3 will get 4 places, an 
// individual with e of 5.9 will get 5 places.  If there are any spaces
// remaining then we fill them with the individuals with highest fractional
// values (don't as *ME* why Goldberg does it this way...)

  int i;
  for(i=0; i<p.size(); i++){
    idx[i] = i;
    expected = p.scaling().fitness(i)/p.scaling().ave();
    n = (int)expected;
    fraction[i] = expected - n;

    while(n > 0){
      choices[k] = i;
      k++; n--;
    }
  }

#ifndef NDEBUG
  cerr << "GADSSelector: calling QuickSort\n";
#endif

  GAQuickSort(idx, fraction, 0, N-1);
  for(i=p.size()-1; k<p.size(); k++, i--)
    choices[k] = idx[i];

// Now just pick randomly from the list we generated in the preselection step.

  return(p.scaling().individual(choices[GARandomInt(0, p.size()-1)]));
}

#endif
