// $Header: /usr/people/mbwall/src/ga/ga/RCS/statistics.C,v 1.1 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  statistics.C
  mbwall 28jul94
  Copyright 1995 Massachusetts Institute of Technology - all rights reserved

 DESCRIPTION:
  Definition of the statistics object.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <fstream.h>
#include <ga/statistics.h>


GAStatistics::GAStatistics() : params() {
  cg=ns=nc=nm=0;
  bi=bc=be=0.0;
  boa = (GAPopulation *)0; params.nBestChrom=0;
  nconv=0; Nconv = params.nConv;
  cscrs = new float[Nconv]; memset(cscrs, 0, Nconv*sizeof(float));
  nscrs=0; Nscrs = 10;
  scrs = new float[Nscrs]; memset(scrs, 0, Nscrs*sizeof(float));
  scorefile = (char *)0;
}
GAStatistics::GAStatistics(const GAGenome & c) : params() {
  cg=ns=nc=nm=0;
  bi=bc=be=0.0;
  boa = new GAPopulation(c, params.nBestChrom);
  nconv=0; Nconv = params.nConv;
  cscrs = new float[Nconv]; memset(cscrs, 0, Nconv*sizeof(float));
  nscrs=0; Nscrs = 10;
  scrs = new float[Nscrs]; memset(scrs, 0, Nscrs*sizeof(float));
  scorefile = (char *)0;
}
GAStatistics::GAStatistics(const GAStatistics & orig){
  cscrs=(float *)0;
  scrs=(float *)0;
  boa=(GAPopulation *)0;
  scorefile=(char *)0;
  copy(orig);
}
GAStatistics::~GAStatistics(){
  delete [] cscrs;
  delete [] scrs;
  delete boa;
  delete [] scorefile;
}
void 
GAStatistics::copy(const GAStatistics & orig){
  params = orig.params;
  cg=orig.cg;
  bi=orig.bi; bc=orig.bc; be=orig.be;
  on=orig.on; off=orig.off;
  delete [] cscrs;
  nconv=orig.nconv; Nconv=orig.Nconv;
  cscrs = new float [Nconv]; memcpy(cscrs, orig.cscrs, Nconv*sizeof(float));
  delete [] scrs;
  nscrs=orig.nscrs; Nscrs=orig.Nscrs;
  scrs = new float [Nscrs]; memcpy(scrs, orig.scrs, Nscrs*sizeof(float));
  delete boa;
  boa = orig.boa->clone();
  ns=orig.ns; nc=orig.nc; nm=orig.nm;
  delete [] scorefile;
  scorefile = new char [strlen(orig.scorefile)+1];
  sprintf(scorefile, orig.scorefile);
}


// Update the genomes in the 'best of all' population to reflect any 
// changes made to the current population.  We just grab the genomes with
// the highest scores from the current population, and if they are higher than
// those of the genomes in the boa population, they get copied.  Note that
// the bigger the boa array, the bigger your running performance hit because
// we have to look through all of the boa to figure out which are better than
// those in the population.  The fastest way to use the boa is to keep only 
// one genome in the boa population.  A flag of 'True' will reset the boa
// population so that it is filled with the best of the current population.
//   Unfortunately it could take a long time to update the boa array using the
// copy method.  We'd like to simply keep pointers to the best genomes, but
// the genomes change from generation to generation, so we can't depend on
// that.
//   Notice that keeping boa is useful even for overlapping populations.  The
// boa keeps individuals that are different from each other - the overlapping
// population may not.  However, keeping boa is most useful for populations
// with little overlap.
void
GAStatistics::
updateBestIndividual(const GAPopulation & pop, const GABoolean flag){
  if(boa == (GAPopulation *)0 || boa->size() == 0) return; // do nothing

  if(flag == gaTrue){		// reset the BOA array
    int j=0;
    for(int i=0; i<boa->size(); i++){
      boa->best(i).copy(pop.best(j));
      if(j < pop.size()-1) j++;
    }
    return;
  }

  if(boa->size() == 1){		// there's only one boa so replace it with bop
    if(pop.best().score() > boa->best().score())
      boa->best().copy(pop.best());
  }
  else{
    int i=0, j, k;
    while(i < pop.size() && pop.best(i).score() > boa->worst().score()){
      for(k=0;
	  pop.best(i).score()<boa->best(k).score() && k<boa->size();
	  k++);
      for(j=k; j<boa->size(); j++){
	if(pop.best(i) == boa->best(j)) break;
	if(pop.best(i).score() > boa->best(j).score()){
	  boa->worst().copy(pop.best(i));
	  boa->sort(gaTrue);
	  break;
	}
      }
      i++;
    }
  }
  return;
}


// Use this method to update the statistics to account for the current
// population.  This routine increments the generation counter and assumes that
// the population that gets passed is the current population.
void
GAStatistics::update(const GAPopulation & pop){
  ++cg;
  if(params.scoreFreq > 0 && (cg % params.scoreFreq == 0))
    setScore(pop.best().score());
  bc = pop.best().score();
  be = (bc > be) ? bc : be;
  on = (on*(cg-1)+pop.ave())/cg;
  off = (off*(cg-1)+bc)/cg;
  setConvScore(bc);
  updateBestIndividual(pop);
}


// Reset the GA's statistics based on the population.  To do this right you
// should initialize the population before you pass it to this routine.  If you
// don't, the stats will be bogus.
void
GAStatistics::reset(const GAPopulation & pop){
  cg=0;

  memset(scrs, 0, Nscrs*sizeof(float));
  scrs[0] = pop.best().score();
  nscrs = 1;

  memset(cscrs, 0, Nconv*sizeof(float));
  cscrs[0] = scrs[0];
  nconv = 1;

  updateBestIndividual(pop, gaTrue);
  bc=be=bi=pop.best().score();
  on=pop.ave();
  off=bc;
  ns=nc=nm=0;
}


ostream & 
GAStatistics::scores(ostream & os){
  if(params.scoreFreq > 0){
    for(int i=0; i<nscrs; i++)
      os << (cg-(nscrs-i-1)*params.scoreFreq) << "\t" << scrs[i] << "\n";
  }
  return os;
}


void
GAStatistics::dumpScores(){
  ofstream outfile(scorefile, ios::app);
  if(!outfile){
    GAErr("GAStatistics", "dumpScores", gaErrWriteError, scorefile);
    nscrs=0;
    return;
  }
  scores(outfile);
  outfile.close();
  nscrs=0;
}


ostream & operator<<(ostream & os, const GAStatistics & arg){
  os << arg.params << "\n";
  os << arg.cg << "\t# current generation\n";
  os << arg.ns << "\t# number of selections since initialization\n";
  os << arg.nc << "\t# number of crossovers since initialization\n";
  os << arg.nm << "\t# number of mutations since initialization\n";
  os << arg.bi << "\t# best score in initial population\n";
  os << arg.bc << "\t# best score in current population\n";
  os << arg.be << "\t# best score ever\n";
  os << arg.on << "\t# average of all scores ('on-line' performance)\n";
  os << arg.off << "\t# average of best scores ('off-line' performance)\n";
  return(os);
}

