// $Header: /usr/people/mbwall/src/ga/ga/RCS/statistics.h,v 1.1 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  statistics.h
  mbwall 14jul95
  Copyright 1995 Massachusetts Institute of Technology - all rights reserved

 DESCRIPTION:
  Header for the statistics object used by the GA objects.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#ifndef _ga_statistics_h_
#define _ga_statistics_h_

#include <stdio.h>
#include <math.h>
#include <ga/types.h>
#include <ga/genome.h>
#include <ga/population.h>
#include <ga/parameters.h>


/* ----------------------------------------------------------------------------
Statistics class
  We define this class as a storage object for the current state of the GA.
Whereas the parameters object keeps track of the user-definable settings for 
the GA, the statistics object keeps track of the data that the GA generates 
along the way.
---------------------------------------------------------------------------- */
class GAStatistics {
public:
  GAStatistics();
  GAStatistics(const GAGenome &);
  GAStatistics & operator=(const GAStatistics & orig){copy(orig);return *this;}
  GAStatistics(const GAStatistics & orig);
  virtual ~GAStatistics();
  void copy(const GAStatistics &);

  const GAParameters & parameters() const {return params;}

  unsigned int populationSize(const unsigned int x){return params.popSize=x;}
  unsigned int populationSize() const {return params.popSize;}
  float pCrossover(const float x){return params.pCross=x;}
  float pCrossover() const {return params.pCross;}
  float pMutation(const float x){return params.pMut=x;}
  float pMutation() const {return params.pMut;}
  unsigned int nGenerations(const unsigned int x){return params.nGen=x;}
  unsigned int nGenerations() const {return params.nGen;}
  float pConvergence(const float x){return params.pConv=x;}
  float pConvergence() const {return params.pConv;}

  unsigned int nConvergence(const unsigned int x)
    {return params.nConv = nConvScores(x);}
  unsigned int nConvergence() const {return params.nConv;}
  unsigned int nBestGenomes(const unsigned int x)
    {return params.nBestChrom = (boa ? boa->resize(x) : 0);}
  unsigned int nBestGenomes() const {return(boa ? boa->size() : 0);}
  int scoreFrequency(const unsigned int x){return(params.scoreFreq = x);}
  int scoreFrequency() const {return params.scoreFreq;}

  int generation() const {return cg;}
  int selections() const {return ns;}
  int crossovers() const {return nc;}
  int mutations() const {return nm;}
  float online() const {return on;}
  float offline() const {return off;}
  float bestInitial() const {return bi;}
  float bestCurrent() const {return bc;}
  float bestEver() const {return be;}
  float bestNth(const unsigned int n=0) const { // best score in nth prev gen
    return((n<nconv) ? cscrs[n] : 0);
  }
  float convergence() const {	// return current convergence percentage
    return((cg >= nconv) ? 
	   (cscrs[0]-bi ? fabs((cscrs[0]-cscrs[nconv-1])/(cscrs[0]-bi)) : 0.0):
	   1);
  }
  float score(const unsigned int n) const {return((n<nscrs) ? scrs[n] : 0);}
  GAPopulation & bestPopulation() const {return *boa;}
  GAGenome & bestIndividual(const unsigned int n=0) const{return boa->best(n);}
  void update(const GAPopulation & pop);
  void reset(const GAPopulation & pop);
  ostream & scores(ostream & os);
  void scoreDestination(const char *filename){
    delete [] scorefile;
    scorefile = new char[strlen(filename)+1];
    sprintf(scorefile, filename);
  }

  friend ostream & operator<<(ostream & os, const GAStatistics & arg);

protected:
  GAParameters params;		// our own copy of the GA parameters
  unsigned int cg;		// current generation number
  float bi;		// best score in initial population
  float bc;		// current best of population
  float be;		// best score since initialization
  float on;		// "on-line" performance (ave of all scores)
  float off;		// "off-line" performance (ave of best)

  unsigned int nconv, Nconv;	// how many last best we're keeping track of
  float * cscrs;		// best score of last n generations
  unsigned int nscrs, Nscrs;	// how many scores do we have?
  float * scrs;			// score of best of each Nth generation
  GAPopulation * boa;		// keep a copy of the best genomes

  char *scorefile;		// name of file to which scores get written

  float setConvScore(const float s){	// set convergence score and increment
    if(nconv < Nconv) nconv++;
    if(nconv>0) memmove(&(cscrs[1]), &(cscrs[0]), (nconv-1)*sizeof(float));
    return(cscrs[0] = s);
  }
  float setScore(const float s){ // add score to scores 
    if(nscrs+1 >= Nscrs){	 // trigger flush 1 score too soon
      if(scorefile == (char*)0)
	nScores(Nscrs+50);	// grow if too small and no file
      else dumpScores();		       // if file, flush buffer to file
    }
    return(scrs[nscrs++]=s);
  }
  void dumpScores();
  int nScores(const unsigned int n){	// keep this many scores
    float *tmp = scrs;
    scrs = new float[n];
    memcpy(scrs, tmp, (n > Nscrs ? Nscrs : n)*sizeof(float));
    delete [] tmp;
    if(nscrs > n) nscrs = n;
    return(Nscrs=n);
  }
  int nConvScores(const unsigned int n){ // keep this many consecutive scores
    float *tmp = cscrs;
    cscrs = new float[n];
    memcpy(cscrs, tmp, (n > Nconv ? Nconv : n)*sizeof(float));
    delete [] tmp;
    if(nconv > n) nconv = n;
    return(Nconv=n);
  }
  void updateBestIndividual(const GAPopulation&, const GABoolean flag=gaFalse);

public:
// I hate to make these public, but this will do for now.  Do not use these!
// They may not be here in the next release!  These should be accessed only
// by GA object.
//friend GA::nselect();
//friend GA::ncross();
//friend GA::nmutate();

  unsigned int ns;		// number of selections since initialization
  unsigned int nc;		// number of crossovers since initialization
  unsigned int nm;		// number of mutations since initialization
};

#endif
