// $Header: /usr/people/mbwall/src/ga/ga/RCS/tree.op.C,v 1.4 1995/06/30 21:27:55 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  tree.op.C
  mbwall 25feb95
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  Operators for the tree genome.

  See the comments in the array operators file for instructions about how to 
use this code.

*** there are many ways to optimize this code.  we do way too many traversals
    right now, for example.

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#include <ga/config.h>
#include <ga/random.h>
#include <ga/tree.ch.h>
#include <ga/tree.xs.h>


//  This mutation method is destructive.  We randomly pick a node in the tree
// then delete the subtree and node at that point.  Each node in the tree has
// a pmut probability of getting nuked.
//  After the mutation the iterator is left at the root of the tree.
#if USE_DESTRUCTIVE_MUTATOR == 1
int
GATreeDestructiveMutator(GAGenome & c, float pmut)
{
  GATreeGenome<TREE_TYPE> &child=(GATreeGenome<TREE_TYPE> &)c;
  register int n, i;
  if(pmut <= 0.0) return 0;

  n = child.size();
  float nMut = pmut * (float)n;
  if(nMut < 1.0){		// we have to do a flip test for each node
    nMut = 0;
    for(i=0; i<n; i++){
      if(GAFlipCoin(pmut) && child.warp(i)){
	child.destroy();
	nMut++;
      }
    }
  }
  else{				// only nuke the number of nodes we need to
    for(i=0; i<nMut; i++){
      if(child.warp(GARandomInt(0, n-1)))
	child.destroy();
    }
  }
  child.root();		// set iterator to root node

  return((int)nMut);
}
#endif


// This is a rearranging mutation operator.  It randomly picks two nodes in the
// tree and swaps them.  Any node has a pmut chance of getting
// swapped, and the swap could happen to any other node.  And in the case of
// nMut < 1, the swap may generate a swap partner that is the same node, in 
// which case no swap occurs (we don't check).
//   After the mutation the iterator is left at the root of the tree.
#if USE_SWAP_MUTATOR == 1
int
GATreeSwapNodeMutator(GAGenome & c, float pmut)
{
  GATreeGenome<TREE_TYPE> &child=(GATreeGenome<TREE_TYPE> &)c;
  register int n, i;
  if(pmut <= 0.0) return 0;

  n = child.size();
  float nMut = pmut * (float)n;
  nMut *= 0.5;			// swapping one node swaps another as well
  if(nMut < 1.0){		// we have to do a flip test for each node
    nMut = 0;
    for(i=0; i<n; i++){
      if(GAFlipCoin(pmut)){
	child.swap(i,GARandomInt(0,n-1));
	nMut++;
      }
    }
  }
  else{				// only nuke the number of nodes we need to
    for(i=0; i<nMut; i++)
      child.swap(GARandomInt(0,n-1),GARandomInt(0,n-1));
  }
  child.root();		// set iterator to root node

  return((int)nMut*2);
}
#endif



// This is a rearranging mutation operator with subtree swap.  It does the same
// thing as the rearranging mutator above, but swaps subtrees as well as the
// nodes that are selected.
//   After the mutation the iterator is left at the root of the tree.
//   We check to make sure that we don't try to swap ancestral nodes.  If it is
// an ancestral swap, we give up and don't do anything to the tree.  This could
// result in mutation rates that are lower than the specified rate!
// *** mutation rates are not exact!
#if USE_SWAP_SUBTREE_MUTATOR == 1
int
GATreeSwapSubtreeMutator(GAGenome & c, float pmut)
{
  GATreeGenome<TREE_TYPE> &child=(GATreeGenome<TREE_TYPE> &)c;
  register int n, i;
  int a, b;
  if(pmut <= 0.0) return 0;

  n = child.size();
  float nMut = pmut * (float)n;
  nMut *= 0.5;			// swapping one node swaps another as well
  if(nMut < 1.0){		// we have to do a flip test for each node
    nMut = 0;
    for(i=0; i<n; i++){
      if(GAFlipCoin(pmut)){
	b = GARandomInt(0,n-1);
	if(!child.ancestral(i,b)) child.swaptree(i,b);
	nMut++;
      }
    }
  }
  else{				// only nuke the number of nodes we need to
    for(i=0; i<nMut; i++){
      a = GARandomInt(0,n-1);
      b = GARandomInt(0,n-1);
      if(!child.ancestral(a,b)) child.swaptree(a,b);
    }
  }
  child.root();		// set iterator to root node

  return((int)nMut*2);
}
#endif


// The default crossover operator takes a node from parent a (with its entire
// sub-tree) and replaces it with a node from parent b (with its entire sub-
// tree).  If the crossover site is not set, then we pick a random site based
// on the trees in the genomes we're going to cross.  Once we have a valid
// crossover site, we copy the trees from the two genomes.
//   If the crossover site is out of bounds (ie refers to a node not in the
// tree) then we don't do anything to the child.
//   We allow crossover at ANY site in the genomes (including at the root
// node).
// *** we should be able to speed this up.  there is an extra traversal when we
//     do the check to see if the crossover site is valid.
#if USE_1PT_CROSSOVER == 1
void
GATree1PtCrossover(GAGenome & c, const GAGenome & a, const GAGenome & b)
{
  GATreeGenome<TREE_TYPE> &child=(GATreeGenome<TREE_TYPE> &)c;
  GATreeGenome<TREE_TYPE> &mom=(GATreeGenome<TREE_TYPE> &)a;
  GATreeGenome<TREE_TYPE> &dad=(GATreeGenome<TREE_TYPE> &)b;
  GATree1PtCrossSite<TREE_TYPE> & xsite = 
    (GATree1PtCrossSite<TREE_TYPE> &)child.crossoverSite();

  if(xsite.a == gaNoSite || xsite.b == gaNoSite) xsite.pick(c,a,b);

  GATreeIter<TREE_TYPE> momiter(mom), daditer(dad);
  if(momiter.warp(xsite.a) && daditer.warp(xsite.b)){
    child.GATree<TREE_TYPE>::copy(mom);
    GATree<TREE_TYPE> * tree = dad.GATree<TREE_TYPE>::clone(xsite.b);
    child.warp(xsite.a);
    child.swaptree(tree);
    delete tree;
    child.warp(0);
  }
}
#endif
