// $Header: /usr/people/mbwall/src/ga/ga/RCS/genome.h,v 1.3 1995/09/13 22:30:10 mbwall Exp mbwall $
/* ----------------------------------------------------------------------------
  genome.h
  mbwall 28jul94
  Copyright 1995 Massachusetts Institute of Technology

 DESCRIPTION:
  The base genome class is a pure virtual class.
  We use write and read methods for the << and >> operators so that you can
call << or >> on the superclass and still end up getting the right method even
if the pointer you're doing the << or >> on is a sub class.
  When you make a sub-class, be sure to define your own copy (class-copy)
method.  Then for the operator= method you can define a simple call to copy.
This way if someone calls a GAGenome cast of your genome type, the
proper operator= function will get called.

 TO DO:
  We may have to change the return value of the objective function, but for now
it is simply a float (it can be changed to void * eventually so that we can do
vectors and multiple objectives).

 MODIFICATIONS:
---------------------------------------------------------------------------- */
#ifndef _ga_genome_h_
#define _ga_genome_h_

#include <stdlib.h>
#include <iostream.h>
#include <string.h>
#include <ga/id.h>
#include <ga/error.h>
#include <ga/xsite.h>
#include <ga/objective.h>

#ifdef DEBUG
#define DBG(x) x
#else
#define DBG(x)
#endif

class GA;		// tell the compiler about GAs
class GAGenome;


const int gaNoResize        = -1;
const int gaResizeAny       = -10;

enum GADimension {
  gaLength = 0,
  gaWidth = 0,
  gaHeight = 1,
  gaDepth = 2
};

enum GACloneMethod {
  gaCloneContents = 0,
  gaCloneAttributes = 1
};


/* ----------------------------------------------------------------------------
  These are the typedefs for the mutation, crossover, and initialization
methods.  We have to pass a genome since they are pointers to functions
and not member functions.
---------------------------------------------------------------------------- */
typedef void (*GAInitializationOperator)(GAGenome &);
typedef int  (*GAMutationOperator)(GAGenome &, float);
typedef void (*GACrossoverOperator)(GAGenome&,const GAGenome&,const GAGenome&);


/* ----------------------------------------------------------------------------
 These are the default dummy operators for the base genome class or any
derived class that does not define a functional default operator.
---------------------------------------------------------------------------- */
void GANoInitializer(GAGenome &);
int  GANoMutator(GAGenome &, float);
void GANoCrossover(GAGenome &, const GAGenome &, const GAGenome &);



/* ----------------------------------------------------------------------------
Genome
-------------------------------------------------------------------------------
  This should be a pure virtual (abstract) class, except that we need to have
GAGenome pointers running around because we never know what kind of
genome we have until we look at it.  As a result, the base class is not
abstract, but other derived 'base' classes are.
  For any derived class be sure to define the canonical methods:  initializer,
copy initializer, operator=, and destructor.  Make sure that you check for a
self-copy in your copy method (it is possible that a genome will be
selected to cross with itself, and self-copying is not out of the question
either).
  There is always a mutation, crossover, and initialization method defined for
a genome, so we don't have to check for NULL function pointers.  Similarly,
the crossover site is always non-NULL so we don't have to check it either.

initialize
  Use this method to set the initial state of your genomes once they have
  been created.  This initialization is for setting up the genome's state,
  not for doing the basic mechanics of genome class management.  The 
  default behaviour of this method is to change randomly the contents of the 
  genome.  If you want to bias your initial population, this is where to
  make that happen.  (Don't confuse this routine with init().  init() is used
  to set the initial values of the class members, not for settting the state
  of the genome.  initialize() is public, init() is private.)
    
mutate
  Mutate the genome with probability as specified.  What mutation means
  depends upon the type of genome.
 
cross
  Perform a crossover with genetic material from two parents.  Store the
  results in the current genome.  Use information within the child
  genome for where and how the crossover should take place.
    If the genome is crossed with itself, we make a copy and use that as
  the 'parent' or parents.  You don't have to check for this in your own
  crossover methods - the base class takes care of it.
    
pickCrossSite
  Pick the crossover site (based on the genome type and type of crossover
  that you want).  If nothing is specified, generate a random crossover site.
  If you specify a crossover site, that site will be used.  If you specify two
  genomes, then use information in the two genomes to do your cross
  site selection (this is the one that is used by default in the standard GAs
  in this implementation).

mirrorCrossSite
  Make a mirror image of the xsite for the second child.  This only really
  makes sense for crossover methods that are symmetric and that nominally 
  produce two offspring (eg single point crossover).
    
crossoverSite
  Return a pointer to the crossover site information.  You'll have to cast it
  to the appropriate type in order to do anything with the contents.
    
setObjectiveFunction
  Set the genome's objective function.  This also sets the score to 0.

score
  Evaluate the 'performance' of the genome using the objective function.
  The score is kept in the 'score' member.  The 'evaluated' member tells us
  whether or not we can trust the score.  Be sure to set/unset the 'evaluated'
  member as appropriate (eg cross and mutate change the contents of the 
  genome so they unset the 'evaluated' flag).
    If there is no objective function, then simply return the score.  This 
  allows us to use population-based evaluation methods (where the population
  method sets the score of each genome).

clone
  This method allocates space for a new genome and copies the original into
  the new space.  Depending on the argument, it either copies the entier 
  original or just parts of the original.
clone(cont)
  Clone the contents of the genome.  Returns a pointer to a GAGenome
  (which actually points to a genome of the type that was cloned).  Every
  part of the genome is cloned, including the crossover site.
clone(attr)
  Clone the attributes of the genome.  This method does nothing to the
  contents of the genome.  It does NOT call the initialization method.  The
  crossover site is left in its default state.
---------------------------------------------------------------------------- */
class GAGenome : public GAID {
  GADefineIdentity("GAGenome", 50):

public:
  GAGenome(GAInitializationOperator i=GANoInitializer,
	   GAMutationOperator m=GANoMutator,
	   GACrossoverOperator c=GANoCrossover,
	   GACrossSite * s=(GACrossSite *)0,
	   GAObjectiveVector * o=(GAObjectiveVector *)0);
  GAGenome(const GAGenome & orig);
  GAGenome & operator=(const GAGenome & arg){copy(arg); return *this;}
  virtual ~GAGenome(){delete cs; delete ov;}
  virtual GAGenome * clone(const GACloneMethod f=gaCloneContents) const {
    if(f == gaCloneContents)	// just to use flag so no compiler warnings
      GAErr(className(), "clone", gaErrNoCloneContOp);
    else
      GAErr(className(), "clone", gaErrNoCloneAttrOp);
    return (GAGenome *)0;
  }
  virtual void copy(const GAGenome & c);
  float score(const GABoolean flag=gaFalse){
    if(_evaluated == 0 || flag == gaTrue){
      if(of) _score=of(*this);
      _evaluated=gaTrue;
    }
    return(_score);
  }
  float score(const float s){_evaluated=gaTrue; return _score=s;}
  int mutations() const {return _nmut;}
  GA * geneticAlgorithm() const {return ga;}
  GA * geneticAlgorithm(GA & g){return(ga = &g);}
  void * userData() const {return ud;}
  void * userData(void * u){return(ud=u);}
  GAObjectiveVector * objectiveVector() const {return ov;}
  GAObjectiveVector * objectiveVector(const GAObjectiveVector & o)
    {delete ov; ov = o.clone(); return ov;}
  GAObjectiveFunction objectiveFunction() const {return of;}
  GAObjectiveFunction objectiveFunction(GAObjectiveFunction f)
    {_evaluated=gaFalse; return(of=f);}
  void initialize(){_evaluated=gaFalse; _nmut=0; io(*this);}
  GAInitializationOperator initializationOperator() const {return io;}
  GAInitializationOperator initializationOperator(GAInitializationOperator op)
    {return(io=op);}
  int mutate(const float p){_evaluated=gaFalse; return(_nmut=mo(*this,p));}
  GAMutationOperator mutationOperator() const {return mo;}
  GAMutationOperator mutationOperator(GAMutationOperator op)
    {_nmut=0; return(mo=op);}
  void cross(const GAGenome & a, const GAGenome & b){
    if(&a == this || &b == this){
      GAGenome *cpy = clone(gaCloneContents);
      if(&a == this && &b == this) co(*this,*cpy,*cpy);
      else if(&a == this) co(*this,*cpy,b);
      else if(&b == this) co(*this,a,*cpy);
      delete cpy;
    }
    else{ 
      co(*this,a,b);
    }
    _evaluated=gaFalse;
  }
  GACrossSite & crossoverSite() const {return *cs;}
  GACrossSite & crossoverSite(const GACrossSite & s)
    {delete cs; cs=s.clone(); return *cs;}
  GACrossoverOperator crossoverOperator() const {return co;}
  GACrossoverOperator crossoverOperator(GACrossoverOperator op){return(co=op);}
  void crossover(GACrossoverOperator op, const GACrossSite & s)
    {co=op; delete cs; cs=s.clone();}
  void mirrorCrossSite(const GAGenome &a){cs->mirror(*this,a);}
  void pickCrossSite(){cs->pick(*this);}
  void pickCrossSite(const GACrossSite &s){cs->pick(*this,s);}
  void pickCrossSite(const GAGenome &a,const GAGenome &b){cs->pick(*this,a,b);}

  friend ostream & operator<<(ostream & os, const GAGenome & arg)
    {arg._write(os); return(os);}
  friend istream & operator>>(istream & is, GAGenome & arg)
    {arg._read(is); return(is);}
  friend int operator==(const GAGenome & a, const GAGenome & b)
    {return a._equal(b);}
  friend int operator!=(const GAGenome & a, const GAGenome & b)
    {return a._notequal(b);}

protected:
  float _score;			// value returned by the objective function
  GABoolean _evaluated;		// has this genome been evaluated?
  unsigned int _nmut;		// how many mutations occurred?
  GA * ga;			// the ga that is using this genome
  void * ud;		        // pointer to user data
  GAObjectiveFunction of;	// objective function
  GAObjectiveVector * ov;	// vector of objective scores (optional)
  GACrossSite * cs;		// information for use in crossover
  GACrossoverOperator co;
  GAMutationOperator mo;
  GAInitializationOperator io;

  virtual void _read(istream &)
    {GAErr(className(), "read", gaErrNoReadOp);}
  virtual void _write(ostream &) const
    {GAErr(className(), "write", gaErrNoWriteOp);}
  virtual int _equal(const GAGenome &) const
    {GAErr(className(), "equal", gaErrNoEqualOp); return 1;}
  virtual int _notequal(const GAGenome &) const
    {GAErr(className(), "not equal", gaErrNoNotEqualOp); return 0;}
};


typedef GAGenome GAChromosome;

#endif
